#
## $URL: https://repos.ser.asu.edu/svn/lroc/lunaserv/tags/release_20160316/lunaserv/dispatcher.rb $
## 
## Copyright (C) 2013 by Arizona State University and Mark Robinson.
## All rights reserved.
##
## This file has been released under the modified BSD license.
## See COPYING in the distribution package for details.
##
## Author: Nick Estes <nme@ser.asu.edu>
##
## $Author: nme $
##   $Date: 2015-01-30 12:28:47 -0700 (Fri, 30 Jan 2015) $
##    $Rev: 16006 $
##
#
require 'tmpdir'
require 'json'

module Lunaserv
  class Dispatcher
    attr_reader :public

    def self.run
      Lunaserv.dispatcher = self.new
      Lunaserv.dispatcher.run
    end

    def initialize
      Lunaserv.load_or_require 'lunaserv/options_checker.rb'
      Lunaserv.load_or_require 'lunaserv/exception_maker.rb'
      Lunaserv.load_or_require 'lunaserv/wms_error.rb'
      Lunaserv.load_or_require 'lunaserv/projections.rb'
      Lunaserv.load_or_require 'lunaserv/layers.rb'
      Lunaserv.load_or_require 'lunaserv/config.rb'
      Lunaserv.load_or_require 'lunaserv/handler/get_capabilities.rb'
      Lunaserv.load_or_require 'lunaserv/handler/get_map.rb'
      Lunaserv.load_or_require 'lunaserv/handler/get_feature_info.rb'

      self
    end

    def init_params
      unless @params
	@params = {}
	request.params.each do |key, value|
	  @params[key.downcase.to_sym] = value
	end
	Lunaserv.params = @params
      end
    end

    def cors_headers
      Lunaserv.response['Access-Control-Allow-Origin'] = '*'
      Lunaserv.response['Access-Control-Allow-Methods'] = 'POST, GET, OPTIONS'
      Lunaserv.response['Access-Control-Max-Age'] = '1728000'
    end

    def request
      Lunaserv.request
    end

    def response
      Lunaserv.response
    end

    def logger
      Lunaserv.logger
    end

    def params
      Lunaserv.params
    end

    def run
      Dir.mktmpdir("lunaserv", Lunaserv.config.tmpdir) do |tmpdir|
	Lunaserv.tmpdir = tmpdir
	logger.info "Request for: #{Lunaserv.request.path} from #{request.ip} query: #{request.params.inspect}"
	start_time = Time.now
	begin
	  init_params

	  Lunaserv::Projections.init
	  Lunaserv::Layers.init

	  response['Cache-Control'] = 'no-cache'
	  cors_headers

	  @public = true

	  if params[:private] and @public = Rack::Auth::Basic.new(Proc.new {logger.info "Authenticated request from #{Lunaserv.rack_environment['REMOTE_USER']}"; false}, 'Lunaserv') do |username, password|
	    Lunaserv.auth_modules.inject(false) do |acc, auth_module|
	      acc || auth_module.auth(username, password)
	    end
	  end.call(Lunaserv.rack_environment)
	  response.status = @public[0]
	  @public[1].each do |header, value|
	    response[header] = value
	  end
	  response.body = [@public[2].collect(&:to_s).join('')]
	  elsif request.options?
	    response['Access-Control-Allow-Headers'] = 'X-Requested-With'
	    response['Content-Type'] = 'text/plain'
	    response.body = [""]
	    response.status = 200
          elsif request.path == "/lge_json"
            response.status = 200
            response.body = [YAML.load(File.read(File.join(Lunaserv.root, 'public/config/lunaserv.yml'))).to_json]
	  elsif request.path == "/" and not params[:request]
	    response.status = 302
	    response["Location"] = Lunaserv.config.redirect || "lunaserv.html"
	    response.body = ["<html><body><a href=\"lunaserv.html\">Lunaserv</a></body></html>"]
	  elsif params[:request] and (params[:request].downcase == "getcapabilities" or params[:request].downcase == "capabilities")
	    Lunaserv::OptionsChecker.run
	    logger.info "Capabilities request"
	    Lunaserv::Handler::GetCapabilities.run
	  elsif params[:request] and params[:request].downcase =~ /^(get)?map$/
	    Lunaserv::OptionsChecker.run
	    logger.info "Map request"
	    Lunaserv::Handler::GetMap.run
	  elsif params[:request] and params[:request].downcase == 'getfeatureinfo'
	    Lunaserv::OptionsChecker.run
	    logger.info "Feature Info request"
	    Lunaserv::Handler::GetFeatureInfo.run
	  else
	    response.status = 404
	    response["Content-Type"] = "text/html"

	    rhtml = ERB.new(File.read(File.join(Lunaserv.root, "templates/404.erb")))
	    rhtml.filename = "404.erb"
	    response.body = [rhtml.result(binding)]
	  end
	rescue Exception=>e
	  if e.is_a? Lunaserv::WmsError
	    logger.info "A WMS error ocurred: #{e.message}"
	  else
	    logger.error "#{e.class}: #{e.message} - #{e.backtrace.join("\n")}"
	  end
	  Lunaserv::ExceptionMaker.run(e)
	ensure
	  logger.info "Request completed in #{Time.now - start_time} seconds. Status: #{response.status}"
	end
      end
    end
  end
end
