#
# $URL: https://repos.ser.asu.edu/svn/lroc/lunaserv/tags/release_20160316/lunaserv/layers/master_layer.rb $
# 
# Copyright (C) 2013 by Arizona State University and Mark Robinson.
# All rights reserved.
#
# This file has been released under the modified BSD license.
# See COPYING in the distribution package for details.
#
# Author: Nick Estes <nme@ser.asu.edu>
#
# $Author: nme $
#   $Date: 2015-01-30 12:28:47 -0700 (Fri, 30 Jan 2015) $
#    $Rev: 16006 $
#
require 'ostruct'

module Lunaserv
  module Layers
    class MasterLayer
      include Math

      def self.render(options)
	self.new(options).render
      end

      def self.features(params)
	if queryable
	  self.new(params).features
	else
	  raise WmsError.new("Layer #{self} is not queryable.", 'LayerNotQueryable')
	end
      end

      def self.create_layer(defs)
	base_class = check_defs(defs)
	return base_class if base_class.is_a? Array

	layer_name = "#{defs[:body]}_#{defs[:wms_name]}"
	klass = Class.new(base_class)

	klass.const_set "LayerDefs", defs
	klass.const_set "LayerName", layer_name
	klass.class_eval do
	  def self.method_missing(method_sym, *args, &block)
	    if self.const_get('LayerDefs') and self.const_get('LayerDefs').has_key?(method_sym)
	      self.const_get('LayerDefs')[method_sym]
	    else
	      super
	    end
	  end

	  def method_missing(method_sym, *args, &block)
	    if self.class.const_get('LayerDefs') and self.class.const_get('LayerDefs').has_key?(method_sym)
	      self.class.const_get('LayerDefs')[method_sym]
	    else
	      super
	    end
	  end

	  def self.respond_to?(method_sym, include_private = false)
	    if self.const_get('LayerDefs') and self.const_get('LayerDefs').has_key?(method_sym)
	      true
	    else
	      super
	    end
	  end

	  def respond_to?(method_sym, include_private = false)
	    if self.class.const_get('LayerDefs') and self.class.const_get('LayerDefs').has_key?(method_sym)
	      true
	    else
	      super
	    end
	  end
	end

	klass
      end

      def self.check_defs(defs)
	messages = []
	[:type, :wms_name, :title, :abstract, :queryable, :opaque, :minx, :maxx, :miny, :maxy, :keywords].each do |key|
	  messages << "Config is missing base option #{key}." unless defs.has_key? key
	end
	messages << "Wms_name must start with a letter." unless defs[:wms_name] =~ Regexp.new('^[a-z]')
	messages << "Keywords must be an array." unless defs[:keywords].is_a? Array
	messages << "feature_search_radius is required for queryable layers" if defs[:queryable] and not defs[:feature_search_radius]

	begin
	  base_class = Object.module_eval("::Lunaserv::Layers::" + "#{defs[:type]}_base_layer".split('_').collect(&:capitalize).join(''))
	  base_class.check_defs(defs, messages)
	rescue NameError
	  messages << "Layer type #{defs[:type].inspect} not recognized."
	end

	if messages.empty?
	  return base_class
	else
	  return messages
	end
      end

      def self.public?
	respond_to? :public_layer and public_layer == true
      end

      def override_option(option_flag, option_name, default=nil)
	override_value = @options[:extra]["#{self.body}_#{self.wms_name}_#{option_name}".to_sym]
	if override_value
	  [option_flag, override_value]
	elsif self.respond_to? option_name
	  [option_flag, self.send(option_name)]
	elsif default
	  [option_flag, default]
	else
	  []
	end
      end

      def get_map_circle
	options = []
	options += ['-p', @options[:proj]]
	options += ['-x', @options[:bbox][0]]
	options += ['-X', @options[:bbox][2]]
	options += ['-y', @options[:bbox][1]]
	options += ['-Y', @options[:bbox][3]]
	options += ['-W', @options[:width]]
	options += ['-H', @options[:height]]
	options += ['-f']

	Open3.popen3 "map_get_circle", *(options.collect(&:to_s)) do |stdin, stdout, stderr|
	  data = stdout.read
	  return data.strip.split(',').collect(&:to_f)
	end
      end

      def get_pixel_circle(xi, yi)
	options = []
	options += ['-p', @options[:proj]]
	options += ['-x', @options[:bbox][0]]
	options += ['-X', @options[:bbox][2]]
	options += ['-y', @options[:bbox][1]]
	options += ['-Y', @options[:bbox][3]]
	options += ['-W', @options[:width]]
	options += ['-H', @options[:height]]
	options += ['-c', xi]
	options += ['-r', yi]

	Open3.popen3 "map_get_circle", *(options.collect(&:to_s)) do |stdin, stdout, stderr|
	  data = stdout.read
	  return data.strip.split(',').collect(&:to_f)
	end
      end

      def apply_filters(tmp_query)
	if respond_to? :filter_params
	  filter_params.each do |param|
	    if @options[:extra][param.downcase.to_sym]
              if respond_to? "#{param}_CLAUSE".to_sym
                tmp_query.gsub!("\#{#{param}_CLAUSE}", send("#{param}_CLAUSE").gsub("\#{#{param}}", @dbh.escape(@options[:extra][param.downcase.to_sym])))
              elsif respond_to? "#{param}_CODE".to_sym
                vars = OpenStruct.new
                vars.arg = @options[:extra][param.downcase.to_sym]
                vars.layer_name = self.class.const_get('LayerName')
                vars.escape = Proc.new {|str| @dbh.escape(str)}
                rb = File.read(File.join(self.dir, send("#{param}_CODE")))
                clause = vars.instance_eval("binding").eval(rb)
                Lunaserv.logger.debug {"Received this SQL clause: #{clause}"}
                tmp_query.gsub!("\#{#{param}_CLAUSE}", clause)
              end
	    elsif respond_to? "#{param}_CLAUSE_DEFAULT".to_sym
	      tmp_query.gsub!("\#{#{param}_CLAUSE}", send("#{param}_CLAUSE_DEFAULT"))
	    else
	      tmp_query.gsub!("\#{#{param}_CLAUSE}", '')
	    end
	  end
	end

	if tmp_query =~ /#\{CIRCLE_CLAUSE\}/
	  lon, lat, radius = get_map_circle

	  if radius and radius < 10
	    tmp_query.gsub! '#{CIRCLE_CLAUSE}', send("CIRCLE_CLAUSE")
	    tmp_query.gsub! /#\{CIRCLE_LON\}/, lon.to_s
	      tmp_query.gsub! /#\{CIRCLE_LAT\}/, lat.to_s
	      tmp_query.gsub! /#\{CIRCLE_RADIUS\}/, radius.to_s
	  else
	    tmp_query.gsub! '#{CIRCLE_CLAUSE}', ''
	  end
	end

	if tmp_query =~ /#\{BINS\}/
	  options = []
	  options += ['-p', @options[:proj]]
	  options += ['-x', @options[:bbox][0]]
	  options += ['-X', @options[:bbox][2]]
	  options += ['-y', @options[:bbox][1]]
	  options += ['-Y', @options[:bbox][3]]
	  Open3.popen3 "map_get_bins", *(options.collect(&:to_s)) do |bin_stdin, bin_stdout, bin_stderr|
	    tmp_query.gsub!('#{BINS}', bin_stdout.read.strip.gsub(' ', ','))
	  end
	end

        if tmp_query =~ /#\{GEO_SEARCH\}/
          tmp_query.gsub! '#{GEO_SEARCH}', if @options[:extra][:search_bbox] and respond_to? "SEARCH_BBOX_CLAUSE".to_sym
                                             send("SEARCH_BBOX_CLAUSE")
                                           elsif respond_to? "SEARCH_POINT_CLAUSE".to_sym
                                             send("SEARCH_POINT_CLAUSE")
                                           else
                                             ''
                                           end
        end

        if @options[:extra][:search_bbox]
          minx,miny,maxx,maxy = @options[:extra][:search_bbox].split(',')
          options = []
          options += ['-p', @options[:proj]]
          options += ['-x', minx]
          options += ['-y', miny]
          options += ['-X', maxx]
          options += ['-Y', maxy]
          options += ['-n', 10]
          points = Open3.popen3 "map_bbox_to_poly", *(options.collect(&:to_s)) do |bin_stdin, bin_stdout, bin_stderr|
            bin_stdout.read.strip.split(';').collect do |point|
              point.split(',')
            end
          end

          points << points[0] if respond_to?("POINT_REPEAT_LAST".to_sym) and send("POINT_REPEAT_LAST")
          if respond_to?("POINT_TEMPLATE".to_sym)
            point_clause = points.collect do |lon, lat|
              send("POINT_TEMPLATE").gsub('#{LON}',lon).gsub('#{LAT}',lat)
            end.join(send("POINT_SEPARATOR"))
            tmp_query.gsub!('#{POINTS}', point_clause)
          end
        end

        if @options[:x] and @options[:y]
          lon, lat, radius = get_pixel_circle(@options[:x], @options[:y])
          radius *= feature_search_radius
          tmp_query = tmp_query.gsub('#{LON}', lon.to_s).gsub('#{LAT}', lat.to_s).gsub('#{RADIUS}', radius.to_s)
        end

        tmp_query.gsub!('#{FEATURE_LIMIT}', @options[:feature_count].to_s)

        tmp_query
      end
    end
  end
end
