#
## $URL: https://repos.ser.asu.edu/svn/lroc/lunaserv/tags/release_20160316/lunaserv/options_checker.rb $
## 
## Copyright (C) 2013 by Arizona State University and Mark Robinson.
## All rights reserved.
##
## This file has been released under the modified BSD license.
## See COPYING in the distribution package for details.
##
## Author: Nick Estes <nme@ser.asu.edu>
##
## $Author: nme $
##   $Date: 2014-02-20 13:20:19 -0700 (Thu, 20 Feb 2014) $
##    $Rev: 13841 $
##
#
module Lunaserv
  class OptionsChecker
    def self.run
      self.new.run
    end

    def initialize
      @public = Lunaserv.dispatcher.public
      @params = Lunaserv.params
      @config = Lunaserv.config
      @errors = []
    end

    def run
      @params[:extra] = {}
      wms_code = 'WmsError'

      if @params[:wmtver] and @params[:version].nil?
	@params[:version] = @params[:wmtver]
      end

      @params.dup.each do |key, value|
	case key
	when :styles
	  unless value.gsub(',','').gsub(/\s/,'').empty?
	    valueError(key, value)
	    wms_code = "StyleNotDefined"
	  end
	when :extra
	  # skip this to avoid recursive stuff
	when :request
	  valueError(key, value) unless ['getmap', 'getcapabilities', 'getfeatureinfo', 'capabilities', 'map'].include? value.downcase
	  @params[key] = value.downcase
	when :version
	  valueError(key, value) unless ["1.0.0", "1.1.0", "1.1.1"].include? value unless @params[:request].downcase == 'getcapabilities'
	when :body
	  @params[key] = value.downcase
	  @params[key] = 'luna' if @params[key] == 'moon'
	when :layers
	  layers = []
	  value.split(',').each do |layer_name|
	    layer_class = Lunaserv::Layers.get_layer layer_name
	    if layer_class.nil?
	      wms_code = 'LayerNotDefined'
	      @errors << "Layer #{layer_name} is not recognized."
	    elsif @public and not layer_class.public?
	      wms_code = 'LayerNotDefined'
	      @errors << "Layer #{layer_name} is not recognized."
	    else
	      layers << layer_class
	    end
	  end
	  @params[:layers] = layers
	when :srs
	  proj = Lunaserv::Projections.get_srs value
	  if proj
	    @params[:proj] = proj
	  else
	    valueError(key, value, 'not supported.')
	    wms_code = "InvalidSRS"
	  end
	when :bbox
	  minx, miny, maxx, maxy = value.split(',').collect(&:to_f)
	  valueError(key, value) unless minx and miny and maxx and maxy
	  @params[:bbox] = [minx,miny,maxx,maxy]
	when :width
	  valueError(key, value, "must be numeric.") unless value=~/^\d+(\.\d*)?$/
	  valueError(key, value, "must be <= #{@config.maxsize}.") unless value.to_i <= @config.maxsize
	  @params[:width] = value.to_i
	when :height
	  valueError(key, value, "must be numeric.") unless value=~/^\d+(\.\d*)?$/
	  valueError(key, value, "must be <= #{@config.maxsize}.") unless value.to_i <= @config.maxsize
	  @params[:height] = value.to_i
	when :format
	  unless Lunaserv::FORMATS.include? value
	    wms_code = "InvalidFormat"
	    valueError(key, value, "not supported.")
	  end
	when :bgcolor
	  valueError(key, value) unless value =~ /^0x[0-9a-fA-F]{6}$/
	when :transparent
	  valueError(key, value) unless value.upcase =~ /^(TRUE|FALSE)$/
	  @params[:transparent] = value.upcase == 'TRUE'
	when :exceptions
	  value = "application/vnd.ogc.se_blank" if value == "application/vnd.ogc.se_application/vnd.ogc.se_blank"
	  valueError(key, value) unless ['se_xml', 'se_inimage', 'se_blank', 'xml', 'inimage', 'blank', 'application/vnd.ogc.se_xml', 'application/vnd.ogc.se_blank', 'application/vnd.ogc.se_inimage'].include? value.downcase
	  value.sub! /^se_/, ''
	  @params[:exceptions] = (value =~ /application/) ? value : "application/vnd.ogc.se_#{value}"
	when :query_layers
	  layers = []
	  value.split(',').each do |layer_name|
	    layer_class = Lunaserv::Layers.get_layer layer_name
	    if layer_class.nil?
	      wms_code = 'LayerNotDefined'
	      @errors << "Query Layer #{layer_name} is not recognized."
	    elsif @public and not layer_class.public?
	      wms_code = 'LayerNotDefined'
	      @errors << "Query Layer #{layer_name} is not recognized."
	    else
	      layers << layer_class
	    end
	  end
	  @params[:query_layers] = layers
	when :info_format
	  valueError(key, value) unless ['application/vnd.ogc.gml', 'text/yaml', 'text/html'].include? value
	when :feature_count
	  valueError(key, value, "must be a positive integer") unless value =~ /^\d+$/
	    @params[:feature_count] = value.to_i
	when :x
	  valueError(key, value, "must be a positive integer") unless value =~ /^\d+$/
	  @params[:x] = value.to_i
	when :y
	  valueError(key, value, "must be a positive integer") unless value =~ /^\d+$/
	  @params[:y] = value.to_i
	when :service
	  valueError(key, value) unless ['WMS', 'WFS'].include? value.upcase
	  @params[:service] = value.upcase
	else
	  @params[:extra][key] = value
	end
      end

      if @params[:x] and @params[:y] and @params[:width] and @params[:height]
	valueError('X', @params[:x], "must be <= WIDTH") unless @params[:x] >= 0 and @params[:x] <= @params[:width]
	valueError('Y', @params[:y], "must be <= WIDTH") unless @params[:y] >= 0 and @params[:y] <= @params[:height]
      end

      @params[:feature_count] ||= 1
      @params[:info_format] ||= 'text/html'
      @params[:transparent] ||= false
      @params[:bgcolor] ||= "0xffffff"
      @params[:exceptions] ||= 'application/vnd.ogc.se_xml'
      @params[:allow_bad_bbox] ||= false

      unless @params[:bbox].nil?
	unless @params[:bbox][0] < @params[:bbox][2] and @params[:bbox][1] < @params[:bbox][3]
	  @errors << "Your BBOX minx should be less than maxx and miny should be less than maxy.  If you know what you are doing and really want a flipped image, pass a parameter of ALLOW_BAD_BBOX=true"
	end
      end

      case @params[:request]
      when 'getcapabilities'
      when 'getmap'
	musgotta :version, :layers, :srs, :bbox, :width, :height, :format
      when 'getfeatureinfo'
	musgotta :version, :query_layers, :srs, :bbox, :width, :height, :x, :y
      end

      unless @errors.empty?
	raise Lunaserv::WmsError.new @errors.join("\n"), wms_code
      end
    end
    
    def musgotta(*options)
      options.each do |option|
	@errors << "Option #{option} missing from request." unless @params[option]
      end
    end


    def valueError(key, value, message = "was not recognized.")
      @errors << "#{key}=#{value} #{message}"
    end
  end
end
