/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_day_night_layer.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Mapserver Day/Night layer
 */

#include <getopt.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include <gd.h>

#include "map_shared_code.h"

typedef struct {
	int red;
	int green;
	int blue;
	int opacity;
} pixel_value;

int create_day_night_layer(const char* dest_file, const char* projection, BBox bbox, int dest_width, int dest_height, pixel_value color, double sub_solar_longitude, double sub_solar_latitude) {
	projPJ pj_dest, pj_latlon;

	pj_latlon = pj_init_plus("+proj=latlon +a=1");
	pj_dest = pj_init_plus(projection);
	if (!pj_dest) {
		fprintf(stderr, "Error: Unable to generate output projection object.\n");
		exit(1);
	}
	validate_bbox(projection, pj_dest, &bbox);

	double dest_scale_x =  (((double) (bbox.maxx-bbox.minx))/dest_width);
	double dest_scale_y =  (((double) (bbox.miny-bbox.maxy))/dest_height);

	// Image, colors; Note: libgd measures opacity from 0-127, and it wants transparency, not opacity, so we invert our opacity values as well as scaling them down by half
	gdImagePtr output_image = gdImageCreateTrueColor(dest_width, dest_height);
	gdImageSaveAlpha(output_image, 1);
	int transparent = gdTrueColorAlpha(255, 0, 0, 127);
	int shade_color = gdTrueColorAlpha(color.red, color.green, color.blue, (255 - color.opacity) / 2);
	gdImageAlphaBlending(output_image, 0);
	gdImageFilledRectangle(output_image, 0, 0, dest_width, dest_height, transparent);
	gdImageAlphaBlending(output_image, 1);

	for (int y = 0; y < dest_height; y++) {
		for (int x = 0; x < dest_width; x++) {
			double longitude;
			double latitude;
			// Project this point to lat/lon, and night-shade the point if the projection went fine and it's at least 90 degrees away from the sun
			bool good_point = project_image_to_latlon(projection, &longitude, &latitude, pj_latlon, pj_dest, bbox, x, y, dest_scale_x, dest_scale_y);
			if (good_point && short_arc_length(longitude, latitude, sub_solar_longitude, sub_solar_latitude) >= 90)
				gdImageSetPixel(output_image, x, y, shade_color);
		}
	}

	FILE * output_file = fopen(dest_file, "wb");
	gdImagePng(output_image, output_file);
	fclose(output_file);
	gdImageDestroy(output_image);
	pj_free(pj_dest);
	pj_free(pj_latlon);
	return 0;
}

void usage(const char* progname) {
	printf(
			"%s: render a day/night layer to be superimposed onto a map projection.\n\n"
			"Options:\n"
			"  --help, -h: show help\n"
			"  --projection, -p projection: projection to map in (proj4 format)\n"
			"  --longitude, -L n: longitude of sub-Solar point\n"
			"  --latitude, -l n: latitude of sub-Solar point\n"
			"  --minx, -x n;\n"
			"  --maxx, -X n;\n"
			"  --miny, -y n;\n"
			"  --maxy, -Y n: define the bounding box for the output image.\n"
			"  --dest_width, -W size: width of the destination image.\n" 
			"  --dest_height, -H size: height of the destination image.\n"
			"  --output, -o filename: file to render the image into.\n"
			"  --color -C s: 6-digit hexadecimal string specifying RGB for 'night' color\n"
			"  --opacity, -O s: 2-digit hexadecimal string specifying opacity of 'night' color\n"
			"    Note: color/opacity are optional and default to black, 50%% opaque.\n", progname);
}

int main(int argc, char **argv) {
	int c, option_index;
	static struct option long_options[] =
	{
		{"help",                 no_argument,         0, 'h'},
		{"projection",           required_argument,   0, 'p'},
		{"longitude",            required_argument,   0, 'L'},
		{"latitude",             required_argument,   0, 'l'},
		{"minx",                 required_argument,   0, 'x'},
		{"maxx",                 required_argument,   0, 'X'},
		{"miny",                 required_argument,   0, 'y'},
		{"maxy",                 required_argument,   0, 'Y'},
		{"dest_width",           required_argument,   0, 'W'},
		{"dest_height",          required_argument,   0, 'H'},
		{"output",               required_argument,   0, 'o'},
		{"color",                required_argument,   0, 'C'},
		{"opacity",              required_argument,   0, 'O'},
	};

	char *projection = 0, *dest_file = 0;
	int dest_width=0, dest_height=0;
	double sub_solar_longitude = 0.0, sub_solar_latitude = 0.0;
	bool got_sub_solar_longitude = false, got_sub_solar_latitude = false;
	BBox bbox = {0,0,0,0};
	pixel_value color = {0, 0, 0, 128};
	bool gotx=(false), gotX=(false), goty=(false), gotY=(false);

	while (1) { 
		c = getopt_long(argc, argv, "hp:L:l:x:X:y:Y:W:H:o:C:O:", long_options, &option_index);

		if (c == -1) break;

		switch(c) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'p':
				projection = strdup(optarg);
				break;
			case 'L':
				sscanf(optarg, "%lf", &sub_solar_longitude);
				got_sub_solar_longitude = true;
				break;
			case 'l':
				sscanf(optarg, "%lf", &sub_solar_latitude);
				got_sub_solar_latitude = true;
				break;
			case 'x':
				sscanf(optarg, "%lf", &bbox.minx);
				gotx=(true);
				break;
			case 'X':
				sscanf(optarg, "%lf", &bbox.maxx);
				gotX=(true);
				break;
			case 'y':
				sscanf(optarg, "%lf", &bbox.miny);
				goty=(true);
				break;
			case 'Y':
				sscanf(optarg, "%lf", &bbox.maxy);
				gotY=(true);
				break;
			case 'W':
				dest_width = atoi(optarg);
				break;
			case 'H':
				dest_height = atoi(optarg);
				break;
			case 'o':
				dest_file = strdup(optarg);
				break;
			case 'C':
				if (strlen(optarg) != 6) {fprintf(stderr, "Argument to -C must be a six-digit hexidecimal number.\n"); usage(argv[0]); exit(1);}
				{
					// Parse hex
					int digits[6] = {0,0,0,0,0,0};
					for (int i = 0; i < 6; i++) {
						unsigned char this_char = optarg[i];
						if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
						if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
						else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
						else {fprintf(stderr, "Argument to -C must be a six-digit hexidecimal number.\n"); usage(argv[0]); exit(1);}
					}
					color.red =   16 * digits[0] + digits[1];
					color.green = 16 * digits[2] + digits[3];
					color.blue =  16 * digits[4] + digits[5];
				}
				break;
			case 'O':
				if (strlen(optarg) != 2) {fprintf(stderr, "Argument to -O must be a two-digit hexidecimal number.\n"); usage(argv[0]); exit(1);}
				{
					// Parse hex
					int digits[2] = {0,0};
					for (int i = 0; i < 2; i++) {
						unsigned char this_char = optarg[i];
						if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
						if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
						else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
						else {fprintf(stderr, "Argument to -O must be a two-digit hexidecimal number.\n"); usage(argv[0]); exit(1);}
					}
					color.opacity = 16 * digits[0] + digits[1];
				}
				break;
		}
	}
	bool got_required = true;
	if (projection == 0)              {fprintf(stderr, "A projection must be specified.\n");           got_required = false;}
	if (dest_file == 0)               {fprintf(stderr, "An output file is required.\n");               got_required = false;}
	if (gotx == false)                {fprintf(stderr, "BBox value minx must be specified.\n");        got_required = false;}
	if (gotX == false)                {fprintf(stderr, "BBox value maxx must be specified.\n");        got_required = false;}
	if (goty == false)                {fprintf(stderr, "BBox value miny must be specified.\n");        got_required = false;}
	if (gotY == false)                {fprintf(stderr, "BBox value maxy must be specified.\n");        got_required = false;}
	if (dest_width <= 0)              {fprintf(stderr, "The destination image width must be > 0\n");   got_required = false;}
	if (dest_height <= 0)             {fprintf(stderr, "The destination image height must be > 0\n");  got_required = false;}
	if (!got_sub_solar_longitude)     {fprintf(stderr, "A sub-solar longitude must be supplied.\n");   got_required = false;}
	if (!got_sub_solar_latitude)      {fprintf(stderr, "A sub-solar latitude must be supplied.\n");    got_required = false;}

	if (!got_required) {
		usage(argv[0]);
		exit(1);
	}

	create_day_night_layer(dest_file, projection, bbox, dest_width, dest_height, color, sub_solar_longitude, sub_solar_latitude);

	if (projection != 0) free(projection);
	if (dest_file != 0) free(dest_file);

	exit(0);
}
