/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_get_bins.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Mapserver component to get a list of all bin IDs indicating bins which overlap with a projection window
 */

#include <getopt.h>
#include <stdio.h>
#include <string.h>

#include "map_shared_code.h"

#define TOTAL_BINS              2592   // Total # of bins on surface
#define BINS_PER_ROW            72     // 360 divided by 5
#define BINS_PER_COLUMN         36     // 180 divided by 5
#define LONGITUDE_PER_BIN       5      // Bins are 5-degree boxes
#define LATITUDE_PER_BIN        5      // Ditto
#define LATTICE_SIZE            500    // X by X lattice of sample points from the destination -- turn on all bins to which these points project back

void get_bins(bool bins [], const char* projection, BBox bbox) {
	int ret = 0;
	projPJ pj_latlon = pj_init_plus("+proj=latlon +a=1");
	projPJ pj_dest = pj_init_plus(projection);
	if (!pj_dest) {
		fprintf(stderr, "Error: Unable to generate output projection object.\n");
		exit(1);
	}
	validate_bbox(projection, pj_dest, &bbox);

	// Initialize bins as false
	for (int i = 0; i < TOTAL_BINS; i++)
		bins[i] = false;

	// Many bins meet at the poles, so we check them as a special case
	double longitude = 0;
	double latitude = 90;
	if (!pj_is_latlong(pj_dest)) {
		longitude *= DEG_TO_RAD;
		latitude  *= DEG_TO_RAD;
	}
	ret = pj_transform(pj_latlon, pj_dest, 1, 1, &longitude, &latitude, NULL);
	if (ret == 0 && longitude > bbox.minx && longitude < bbox.maxx && latitude > bbox.miny && latitude < bbox.maxy) {
		// The north pole is in the image, so we will turn on all bins connected to the north pole
		for (int i = 0; i < TOTAL_BINS; i += BINS_PER_COLUMN)
			bins[i] = true;
	}
	longitude = 0;
	latitude = -90;
	if (!pj_is_latlong(pj_dest)) {
		longitude *= DEG_TO_RAD;
		latitude  *= DEG_TO_RAD;
	}
	if (!pj_is_latlong(pj_dest)) latitude *= DEG_TO_RAD;
	ret = pj_transform(pj_latlon, pj_dest, 1, 1, &longitude, &latitude, NULL);
	if (ret == 0 && longitude > bbox.minx && longitude < bbox.maxx && latitude > bbox.miny && latitude < bbox.maxy) {
		// The south pole is in the image, so we will turn on all bins connected to the south pole
		for (int i = BINS_PER_COLUMN - 1; i < TOTAL_BINS; i += BINS_PER_COLUMN)
			bins[i] = true;
	}

	// Check all the bin-corners, starting at row 1; row 0 would just be the north pole
	for (int row = 1; row < BINS_PER_COLUMN; row++) {
		// We check every column though (still < and not <=, because the last column would be longitude 360 or 0, which we took care of first)
		for (int column = 0; column < BINS_PER_ROW; column++) {
			longitude = (column * LONGITUDE_PER_BIN);
			latitude = (90 - row * LATITUDE_PER_BIN);
			if (!pj_is_latlong(pj_dest)) {
				longitude *= DEG_TO_RAD;
				latitude  *= DEG_TO_RAD;
			}
			ret = pj_transform(pj_latlon, pj_dest, 1, 1, &longitude, &latitude, NULL);
			if (ret == 0 && longitude > bbox.minx && longitude < bbox.maxx && latitude > bbox.miny && latitude < bbox.maxy) {
				// This corner is in the image, so we turn on all adjacent bins
				bins[column * BINS_PER_COLUMN + row] = true;
				bins[column * BINS_PER_COLUMN + row - 1] = true;
				bins[(column == 0 ? BINS_PER_ROW - 1 : (column - 1)) * BINS_PER_COLUMN + row] = true;
				bins[(column == 0 ? BINS_PER_ROW - 1 : (column - 1)) * BINS_PER_COLUMN + row - 1] = true;
			}
		}
	}

	// Finally, take a fairly fine (but not too fine to be swift) lattice of points from the destination image, and enable the bin that each point belongs to
	for (int bbox_row = 0;        bbox_row < LATTICE_SIZE;    bbox_row++) {
		for (int bbox_column = 0; bbox_column < LATTICE_SIZE; bbox_column++) {
			double longitude = bbox.minx + ((double)bbox_row /    (LATTICE_SIZE - 1)) * (bbox.maxx - bbox.minx);
			double latitude = bbox.miny +  ((double)bbox_column / (LATTICE_SIZE - 1)) * (bbox.maxy - bbox.miny);
			ret = pj_transform(pj_dest, pj_latlon, 1, 1, &longitude, &latitude, NULL);
			if (!pj_is_latlong(pj_dest)) {
				longitude *= RAD_TO_DEG;
				latitude *= RAD_TO_DEG;
			}
			if (ret == 0) {
				if (longitude < 0) longitude += 360;
				int bin_row = BINS_PER_COLUMN - floor((latitude + 90) / LATITUDE_PER_BIN) - 1;
				if (bin_row == BINS_PER_COLUMN) bin_row--; // Boundary case
				int bin_column = floor(longitude / LONGITUDE_PER_BIN);
				if (bin_column == BINS_PER_ROW) bin_column--; // Boundary case
				bins[bin_column * BINS_PER_COLUMN + bin_row] = true;
			}
		}
	}
}

void usage(const char* progname) {
	printf(
			"%s: Get a list of bin IDs of bins which overlap with a projection\n\n"
			"Options:\n"
			"  --help, -h: show help\n"
			"  --projection, -p projection: projection to map in (proj4 format)\n"
			"  --minx, -x n;\n"
			"  --maxx, -X n;\n"
			"  --miny, -y n;\n"
			"  --maxy, -Y n: define the bounding box of the projection.\n", progname);
}

int main(int argc, char **argv) {
	int c, option_index;
	static struct option long_options[] =
	{
		{"help",                 no_argument,         0, 'h'},
		{"projection",           required_argument,   0, 'p'},
		{"minx",                 required_argument,   0, 'x'},
		{"maxx",                 required_argument,   0, 'X'},
		{"miny",                 required_argument,   0, 'y'},
		{"maxy",                 required_argument,   0, 'Y'},
	};

	char *projection = 0;
	BBox bbox = {0,0,0,0};
	bool gotx=(false), gotX=(false), goty=(false), gotY=(false);

	while (1) { 
		c = getopt_long(argc, argv, "hp:x:X:y:Y:", long_options, &option_index);

		if (c == -1) break;

		switch(c) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'p':
				projection = strdup(optarg);
				break;
			case 'x':
				sscanf(optarg, "%lf", &bbox.minx);
				gotx=(true);
				break;
			case 'X':
				sscanf(optarg, "%lf", &bbox.maxx);
				gotX=(true);
				break;
			case 'y':
				sscanf(optarg, "%lf", &bbox.miny);
				goty=(true);
				break;
			case 'Y':
				sscanf(optarg, "%lf", &bbox.maxy);
				gotY=(true);
				break;
		}
	}
	bool got_required = true;
	if (projection == 0)        {fprintf(stderr, "A projection must be specified.\n");          got_required = false;}
	if (gotx == false)          {fprintf(stderr, "BBox value minx must be specified.\n");       got_required = false;}
	if (gotX == false)          {fprintf(stderr, "BBox value maxx must be specified.\n");       got_required = false;}
	if (goty == false)          {fprintf(stderr, "BBox value miny must be specified.\n");       got_required = false;}
	if (gotY == false)          {fprintf(stderr, "BBox value maxy must be specified.\n");       got_required = false;}

	if (!got_required) {
		usage(argv[0]);
		exit(1);
	}

	bool bins[TOTAL_BINS];
	get_bins(bins, projection, bbox);
	// Bin indices are 1-based, so we have to add one to our indices
	for (int i = 0; i < TOTAL_BINS; i++)
		if (bins[i]) printf("%i ", i + 1);
	printf("\n");

	if (projection != 0) free(projection);

	exit(0);
}
