/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_illum_layer.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Illumination layer using Proj4 -- derived closely from lum_gen
 */

#include <ctype.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "map_shared_code.h"

#define USAGE_COLUMN_WIDTH 50

#define SOLAR_RADIUS 696342

typedef struct {
	uint8_t red;
	uint8_t green;
	uint8_t blue;
	uint8_t opacity;
} color;

typedef struct {
	bool full_radii;
	bool meters;
} cube_value_format;

struct pixel_distance_information {
	int x, y;
	double distance_from_sun;
};

/* Given two pixel locations, figure out which one is closer to the sun and
 * should be computed first */
int closer_to_sun(const void *pixel_a_void, const void *pixel_b_void) {
	struct pixel_distance_information *pixel_a = (struct pixel_distance_information *)pixel_a_void;
	struct pixel_distance_information *pixel_b = (struct pixel_distance_information *)pixel_b_void;
	if (pixel_a->distance_from_sun < pixel_b->distance_from_sun) return -1;
	if (pixel_a->distance_from_sun > pixel_b->distance_from_sun) return +1;
	return 0;
}

/*** Photometric functions     ***/

// Pointer to a photometric function
typedef double (*photometric_function) (double, double, double);

// Basic Lambertian
double simple_lambert(double phase, double emission, double incidence) {
	/* In this simplistic model, intensity of illumination is simply
	 * proportional to cosine of incidence angle */
	if (incidence <= 0)  return 1;
	if (incidence >= 90) return 0;
	return cos(incidence * DEG_TO_RAD);
}

// Lommel-Seeliger
double lommel_seeliger(double phase, double emission, double incidence) {
	/* Cap incidence and emission angles from getting too close to 90 degrees,
	 * which causes "speckling" due to the large variability of the formula
	 * when these angles are both very close to 90 degrees */
	if (incidence > 88) incidence = 88;
	if (emission  > 88) emission  = 88;
	double result = cos(incidence * DEG_TO_RAD) / (cos(incidence * DEG_TO_RAD) + cos(emission * DEG_TO_RAD));
	if (result > 1) return 1;
	if (result < 0) return 0;
	/* Apply polynomial curve f(x) = -x^2 + 2x, which is the simplest
	 * polynomial mapping 0 to 0, 1 to 1, and 0.5 to 0.75 */
	return -(result * result) + 2 * result;
}

/*** End photometric functions ***/

// Some basic vector math

double vector_magnitude(double * vector) {
	return sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);
}

void scale_to_unit(double * vector) {
	double magnitude = vector_magnitude(vector);
	vector[0] /= magnitude;
	vector[1] /= magnitude;
	vector[2] /= magnitude;
}

double dot(double * first, double * second) {
	return first[0] * second[0] + first[1] * second[1] + first[2] * second[2];
}

void cross(double * first, double * second, double * result) {
	result[0] = first[1] * second[2] - first[2] * second[1];
	result[1] = first[2] * second[0] - first[0] * second[2];
	result[2] = first[0] * second[1] - first[1] * second[0];
}

double angle_between(double * v1, double * v2) {
	return acos(dot(v1, v2) / (vector_magnitude(v1) * vector_magnitude(v2))) * RAD_TO_DEG;
}

// Get elevation given geographic coordinates
float get_elevation(cube_properties * DEM_info, double longitude, double latitude, cube_value_format DEM_values) {
	float elevation = get_value_at_geographic_point(DEM_info, longitude, latitude);
	if (DEM_values.meters) elevation /= 1000;
	if (!DEM_values.full_radii) elevation += DEM_info->radius / 1000;
	return elevation;
}

// Given a lon/lat starting point, a distance, and an initial bearing along an equator, return the destination point
void find_destination_km(double longitude_a, double latitude_a, double bearing, double distance, double * longitude_b, double * latitude_b, cube_properties * DEM_info) {
	validate_longitude(&longitude_a, DEM_info);
	longitude_a  *= DEG_TO_RAD;
	latitude_a   *= DEG_TO_RAD;
	bearing      *= DEG_TO_RAD;
	*latitude_b   = asin(sin(latitude_a) * cos(distance / (DEM_info->radius / 1000)) + cos(latitude_a) * sin(distance / (DEM_info->radius / 1000)) * cos(bearing));
	*longitude_b  = longitude_a + atan2(sin(bearing) * sin(distance / (DEM_info->radius / 1000)) * cos(latitude_a), cos(distance / (DEM_info->radius / 1000)) - sin(latitude_a) * sin(*latitude_b));
	*longitude_b *= RAD_TO_DEG;
	*latitude_b  *= RAD_TO_DEG;
	validate_longitude(longitude_b, DEM_info);
}

// Get a unit normal vector based on topography
void unit_normal_vector(cube_properties * DEM_info, double longitude, double latitude, double * normal, cube_value_format DEM_values) {
	validate_longitude(&longitude, DEM_info);
	uint64_t image_coords[2];
	double row_fraction = 1 - ((latitude - DEM_info->latitude_minimum) / (DEM_info->latitude_maximum - DEM_info->latitude_minimum));
	if (row_fraction < 0) row_fraction = 0;
	if (row_fraction > 1) row_fraction = 1;
	image_coords[0] = (uint64_t)(row_fraction * (DEM_info->height - 1));
	double column_fraction = (longitude - DEM_info->longitude_minimum) / (DEM_info->longitude_maximum - DEM_info->longitude_minimum);
	if (column_fraction < 0) column_fraction = 0;
	if (column_fraction > 1) column_fraction = 1;
	image_coords[1] = (uint64_t)(column_fraction * (DEM_info->width - 1));
	/* If this pixel does not have a pixel above and/or to the left, move over
	 * to an adjacent pixel -- it's an easy fix and will be close enough */
	if (image_coords[0] == 0) {
		image_coords[0]++;
		latitude -= (DEM_info->latitude_maximum - DEM_info->latitude_minimum) / (DEM_info->height - 1);
	}
	if (image_coords[1] == 0) {
		image_coords[1]++;
		longitude += (DEM_info->longitude_maximum - DEM_info->longitude_minimum) / (DEM_info->width - 1);
	}
	double geo_coords_up[2];
	double geo_coords_left[2];
	geo_coords_up[0] = longitude;
	geo_coords_up[1] = latitude + (DEM_info->latitude_maximum - DEM_info->latitude_minimum) / (DEM_info->height - 1);
	geo_coords_left[0] = longitude - (DEM_info->longitude_maximum - DEM_info->longitude_minimum) / (DEM_info->width - 1);
	geo_coords_left[1] = latitude;
	double elevation;
	double elevation_up;
	double elevation_left;
	elevation      = get_elevation(DEM_info, longitude,          latitude,           DEM_values);
	elevation_up   = get_elevation(DEM_info, geo_coords_up[0],   geo_coords_up[1],   DEM_values);
	elevation_left = get_elevation(DEM_info, geo_coords_left[0], geo_coords_left[1], DEM_values);
	double rect_coords[3];
	double rect_coords_up[3];
	double rect_coords_left[3];
	rect_coords[0]      = elevation      * cos(latitude * DEG_TO_RAD) * cos(longitude * DEG_TO_RAD);
	rect_coords[1]      = elevation      * cos(latitude * DEG_TO_RAD) * sin(longitude * DEG_TO_RAD);
	rect_coords[2]      = elevation      * sin(latitude * DEG_TO_RAD);
	rect_coords_up[0]   = elevation_up   * cos(geo_coords_up[1] * DEG_TO_RAD) * cos(geo_coords_up[0] * DEG_TO_RAD);
	rect_coords_up[1]   = elevation_up   * cos(geo_coords_up[1] * DEG_TO_RAD) * sin(geo_coords_up[0] * DEG_TO_RAD);
	rect_coords_up[2]   = elevation_up   * sin(geo_coords_up[1] * DEG_TO_RAD);
	rect_coords_left[0] = elevation_left * cos(geo_coords_left[1] * DEG_TO_RAD) * cos(geo_coords_left[0] * DEG_TO_RAD);
	rect_coords_left[1] = elevation_left * cos(geo_coords_left[1] * DEG_TO_RAD) * sin(geo_coords_left[0] * DEG_TO_RAD);
	rect_coords_left[2] = elevation_left * sin(geo_coords_left[1] * DEG_TO_RAD);
	double north_vector[3];
	double west_vector[3];
	north_vector[0] = rect_coords_up[0]   - rect_coords[0];
	north_vector[1] = rect_coords_up[1]   - rect_coords[1];
	north_vector[2] = rect_coords_up[2]   - rect_coords[2];
	west_vector[0]  = rect_coords_left[0] - rect_coords[0];
	west_vector[1]  = rect_coords_left[1] - rect_coords[1];
	west_vector[2]  = rect_coords_left[2] - rect_coords[2];
	cross(north_vector, west_vector, normal);
	scale_to_unit(normal);
}

/* Decide whether or not a pixel is shadowed.  An array of bools is passed in
 * which signifies which pixels have already been deemed illuminated, which
 * speeds up the algorithm immensely, and is updated if this pixel appears
 * illuminated as well. */
void determine_whether_shadowed(cube_properties *DEM_info, cube_value_format DEM_values, bool *light_map, double shadow_interval, double elevation_variance, double *solar_vector, double longitude, double latitude, char *projection, projPJ pj_latlon, projPJ pj_dest, BBox bbox, int output_width, int output_height, double dest_scale_x, double dest_scale_y, int x, int y) {
	double current_point[3];
	current_point[0] = cos(latitude * DEG_TO_RAD) * cos(longitude * DEG_TO_RAD);
	current_point[1] = cos(latitude * DEG_TO_RAD) * sin(longitude * DEG_TO_RAD);
	current_point[2] = sin(latitude * DEG_TO_RAD);
	float elevation = get_elevation(DEM_info, longitude, latitude, DEM_values);
	for (int i = 0; i < 3; i++) {
		current_point[i] *= elevation;
	}
	double distance_from_surface = 0;
	while (distance_from_surface < elevation_variance) {
		current_point[0] += shadow_interval * solar_vector[0];
		current_point[1] += shadow_interval * solar_vector[1];
		current_point[2] += shadow_interval * solar_vector[2];
		double current_norm = vector_magnitude(current_point);
		double current_longitude = atan2(current_point[1], current_point[0]) * RAD_TO_DEG;
		double current_latitude = asin(current_point[2] / current_norm) * RAD_TO_DEG;
		validate_longitude(&current_longitude, DEM_info);
		// Clamp longitude and latitude into DEM boundaries, if necessary
		if (current_latitude  > DEM_info->latitude_maximum)  current_latitude  = DEM_info->latitude_maximum;
		if (current_latitude  < DEM_info->latitude_minimum)  current_latitude  = DEM_info->latitude_minimum;
		if (current_longitude > DEM_info->longitude_maximum) current_longitude = DEM_info->longitude_maximum;
		if (current_longitude < DEM_info->longitude_minimum) current_longitude = DEM_info->longitude_minimum;
		float current_elevation = get_elevation(DEM_info, current_longitude, current_latitude, DEM_values);
		distance_from_surface = sqrt(current_point[0] * current_point[0] + current_point[1] * current_point[1] + current_point[2] * current_point[2]) - current_elevation;
		// If we're below the surface, the point in question is shadowed
		if (distance_from_surface < 0) {
			return;
		}
		/* If we made it here, we are above the surface.  If this point is
		 * illuminated, the point we started at must be illuminated as well. */
		int other_point_x;
		int other_point_y;
		if (project_latlon_to_image(projection, &other_point_x, &other_point_y, pj_latlon, pj_dest, bbox, current_longitude, current_latitude, false, dest_scale_x, dest_scale_y)) {
			if (other_point_x >= 0 && other_point_x < output_width) {
				if (other_point_y >= 0 && other_point_y < output_height) {
					if (light_map[other_point_y * output_width + other_point_x]) {
						light_map[y * output_width + x] = true;
						return;
					}
				}
			}
		}
	}
	// We can't find any cause of shadowing for this pixel; assume illuminated
	light_map[y * output_width + x] = true;
	return;
}

void usage_line(const char * left, const char * right, ...) {
	printf("%s", left);
	if (USAGE_COLUMN_WIDTH >= strlen(left)) {
		for (int i = 0; i < USAGE_COLUMN_WIDTH - strlen(left); i++) {
			printf(" ");
		}
	}
	else {
		// Resort to a tab if USAGE_COLUMN_WIDTH is too small
		printf("\t");
	}
	va_list parameters;
	va_start(parameters, right);
	char right_formatted[USAGE_COLUMN_WIDTH * 20];
	vsprintf(right_formatted, right, parameters);
	char * printer = right_formatted;
	int current_char_position = 0;
	while (*printer) {
		if (current_char_position >= USAGE_COLUMN_WIDTH && *printer == ' ') {
			// Advance a line
			printf("\n");
			for (int i = 0; i < USAGE_COLUMN_WIDTH; i++) {
				printf(" ");
			}
			current_char_position = 0;
			while (*printer == ' ') printer++;
		}
		else {
			printf("%c", *printer);
			current_char_position++;
			printer++;
		}
	}
	printf("\n");
}

// Usage info
void usage (char * name) {
	printf("%s: Illumination map layer, based heavily on lum_gen.\n", name);
	printf("\nUsage: %s <parameters>\n", name);

	printf("\nRequired parameters\n-------------------\n");
	usage_line("-i, --input <Input DEM>:", "Path/filename to DEM cube for elevation data.  May be specified more than once; supply multiple resolutions for more efficiency with low-resolution output.");
	usage_line("-L, --longitude <Sub-solar longitude>:", "Longitude of sub-solar location in degrees.");
	usage_line("-l, --latitude <Sub-solar latitude>:", "Latitude of sub-solar location in degrees.");
	usage_line("-x, --min-x <X minimum>:", "Leftmost X coordinate (in projection space defined by -p flag) of output image.");
	usage_line("-X, --max-x <X maximum>:", "Rightmost X coordinate (in projection space defined by -p flag) of output image.");
	usage_line("-y, --min-y <Y minimum>:", "Lowest Y coordinate (in projection space defined by -p flag) of output image.");
	usage_line("-Y, --max-y <Y maximum>:", "Highest Y coordinate (in projection space defined by -p flag) of output image.");
	usage_line("-w, --width <Output width>:", "Output PNG width in pixels.  If supplied without height, height will be automatically computed to match degrees/pixel scale.");
	usage_line("-h, --height <Output height>:", "Output PNG height in pixels.  If supplied without width, width will be automatically computed to match degrees/pixel scale.");
	usage_line("-o, --output <Output PNG>:", "Filename of output PNG to write.");

	printf("\nOptional parameters\n-------------------\n");
	usage_line("-v, --verbose", "Output status information, progress reports, etc.");
	usage_line("-p, --projection <Projection string>:", "PROJ4 Projection string for output image.  Default is \"+proj=latlon\", or equirectangular longitude/latitude.");
	usage_line("--vL, --viewer-longitude:", "Longitude of sub-viewer location in degrees.  Default: center of output projection.");
	usage_line("--vl, --viewer-latitude:", "Latitude of sub-viewer location in degrees.  Default: center of output projection.");
	usage_line("-m, --meters:", "Use this flag if the input data is in meters (kilometers assumed otherwise).");
	usage_line("-r, --radii:", "Use this flag if the input data is in complete radii (offsets assumed otherwise).");
	usage_line("-S, --shadow-color <Shadow color>:", "An 8-digit hexadecimal RGBA color to use for shadowed points.  Default is 000000ff (opaque black).");
	usage_line("-D, --dark-color <Dark color>:", "An 8-digit hexadecimal RGBA color to use for the darkest non-shadowed points.  Default is 000000ff (opaque black).");
	usage_line("-I, --illuminated-color <Illuminated color>:", "An 8-digit hexadecimal RGBA color to use for the most strongly illuminated points.  Default is ffffffff (opaque white).");
	usage_line("--preload:", "Loads the DEM, and albedo map if used, entirely into memory before using it.  This can drastically reduce runtime when the input and/or output is massive in size, but should only be done on a machine with ample memory.");
	usage_line("--photo-function <name>:", "Specify a photometric function to use.  Specify 'help' here to get a list of available photometric functions and exit.");
	usage_line("--albedo-map <Albedo map cube>:", "Path/filename to a cube whose contents should be factored into output illumination values.  May be specified more than once; supply multiple resolutions for more efficiency with low-resolution output.");
	usage_line("--radius <Radius>:", "Radius of planetary body in kilometers.  If not supplied, will attempt to read from DEM header.");
	usage_line("--solar-distance <Distance>:", "The distance from the planetary body to the Sun, in kilometers.  If not supplied, will default to one AU (appropriate for the Earth and the Moon).");
	usage_line("--assume-spherical:", "With this flag enabled, any pixels not covered by the input DEM will be assumed non-shadowed and will be shaded as if sitting on a perfectly smooth sphere.  Default behavior is to use transparent pixels in this scenario.");
	usage_line("--variance <Elevational variation>:", "An upper bound on elevational variance, or vertical difference between highest and lowest points on the surface, in kilometers.  This is used for limiting shadow-casting checks to a sane distance.  Default is 20 (the Moon).");
	usage_line("--shadow-distance <Shadow-casting interval>:", "Interval in kilometers to use when checking for shadowing.  A lower value means more accurate shadows, but these checks also account for the vast majority of program runtime.  By default, five times the input cube header's PixelResolution will be used.  If overriding, keep in mind that this value should be supplied in km, while a cube header's PixelResolution is specified in meters.");
	printf("\n");
}

int main(int argc, char ** argv) {
	// Verbose?
	bool verbose = false;

	// Preload?
	bool preload_cubes = false;

	// Default photometric function
	photometric_function selected_photometric_function = &simple_lambert;

	// Interval distance when scanning for shadow-casting terrain
	double shadow_interval = 0;
	bool shadow_interval_override = false;

	// Radius
	double body_radius;
	bool radius_override = false;

	// Distance to Sun; default is one AU, appropriate for Earth/Moon
	double distance_to_sun = 149597870.7;
	bool distance_to_sun_override = false;

	// An upper bound on the difference between any two elevations, in km
	double elevation_variance = 20;
	bool elevation_variance_override = false;

	// Output parameters
	uint64_t output_width = 0, output_height = 0;
	double sub_solar_longitude, sub_solar_latitude;
	bool got_sub_solar_longitude = false, got_sub_solar_latitude = false;
	double sub_viewer_longitude, sub_viewer_latitude;
	bool got_sub_viewer_longitude = false, got_sub_viewer_latitude = false;
	BBox bbox;
	char * dest_file = NULL;

	// Input DEM
	int dem_pathname_count = 0;
	char ** dem_pathnames = NULL;
	cube_properties DEM_info;

	// Input albedo map
	bool using_albedo_map = false;
	int albedo_map_pathname_count = 0;
	char ** albedo_map_pathnames = NULL;
	cube_properties albedo_map;

	// Default interpretation of DEM pixel values
	cube_value_format DEM_values;
	DEM_values.meters = false;
	DEM_values.full_radii = false;

	// Default colors
	color shadow_color;
	shadow_color.red     = 0;
	shadow_color.green   = 0;
	shadow_color.blue    = 0;
	shadow_color.opacity = 255;

	color minimum_color;
	minimum_color.red     = 0;
	minimum_color.green   = 0;
	minimum_color.blue    = 0;
	minimum_color.opacity = 255;

	color maximum_color;
	maximum_color.red     = 255;
	maximum_color.green   = 255;
	maximum_color.blue    = 255;
	maximum_color.opacity = 255;

	bool assume_spherical = false;

	char * projection = NULL;
	bool got_minimum_x = false;
	bool got_maximum_x = false;
	bool got_minimum_y = false;
	bool got_maximum_y = false;

	for (int current_arg = 1; current_arg < argc; current_arg++) {
		// Input file
		if (strcmp(argv[current_arg], "-i") == 0 || strcmp(argv[current_arg], "--input") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -i.\n\n");
				usage(argv[0]);
				exit(1);
			}
			dem_pathnames = realloc(dem_pathnames, sizeof(char *) * (++dem_pathname_count));
			dem_pathnames[dem_pathname_count - 1] = strdup(argv[current_arg]);
		}
		// Sub-solar coordinates
		else if (strcmp(argv[current_arg], "-L") == 0 || strcmp(argv[current_arg], "--longitude") == 0) {
			if (got_sub_solar_longitude) {
				fprintf(stderr, "Option -L supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -L.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%lf", &sub_solar_longitude);
			if (sub_solar_longitude < -180 || sub_solar_longitude > 360) {
				fprintf(stderr, "Sub-solar longitude must be a number between -180 and +360.\n");
				usage(argv[0]);
				exit(1);
			}
			got_sub_solar_longitude = true;
		}
		else if (strcmp(argv[current_arg], "-l") == 0 || strcmp(argv[current_arg], "--latitude") == 0) {
			if (got_sub_solar_latitude) {
				fprintf(stderr, "Option -l supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -l.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%lf", &sub_solar_latitude);
			if (sub_solar_latitude < -90 || sub_solar_latitude > 90) {
				fprintf(stderr, "Sub-solar latitude must be a number between -90 and +90.\n");
				usage(argv[0]);
				exit(1);
			}
			got_sub_solar_latitude = true;
		}
		// Geographical range
		else if (strcmp(argv[current_arg], "-x") == 0 || strcmp(argv[current_arg], "--min-x") == 0) {
			if (got_minimum_x) {
				fprintf(stderr, "Option -x supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -x.\n\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.minx) != 1) {
				fprintf(stderr, "Minimum X must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_minimum_x = true;
		}
		else if (strcmp(argv[current_arg], "-X") == 0 || strcmp(argv[current_arg], "--max-x") == 0) {
			if (got_maximum_x) {
				fprintf(stderr, "Option -X supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -X.\n\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.maxx) != 1) {
				fprintf(stderr, "Maximum X must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_maximum_x = true;
		}
		else if (strcmp(argv[current_arg], "-y") == 0 || strcmp(argv[current_arg], "--min-y") == 0) {
			if (got_minimum_y) {
				fprintf(stderr, "Option -y supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -y.\n\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.miny) != 1) {
				fprintf(stderr, "Minimum Y must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_minimum_y = true;
		}
		else if (strcmp(argv[current_arg], "-Y") == 0 || strcmp(argv[current_arg], "--max-y") == 0) {
			if (got_maximum_y) {
				fprintf(stderr, "Option -Y supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -Y.\n\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.maxy) != 1) {
				fprintf(stderr, "Maximum Y must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_maximum_y = true;
		}
		// Output width
		else if (strcmp(argv[current_arg], "-w") == 0 || strcmp(argv[current_arg], "--width") == 0) {
			if (output_width != 0) {
				fprintf(stderr, "Option -w supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -w.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%"PRIu64"", &output_width);
			if (output_width < 1) {
				fprintf(stderr, "Output width must be a positive integer.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		// Output height
		else if (strcmp(argv[current_arg], "-h") == 0 || strcmp(argv[current_arg], "--height") == 0) {
			if (output_height != 0) {
				fprintf(stderr, "Option -h supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -h.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%"PRIu64"", &output_height);
			if (output_height < 1) {
				fprintf(stderr, "Output height must be a positive integer.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		// Output file
		else if (strcmp(argv[current_arg], "-o") == 0 || strcmp(argv[current_arg], "--output") == 0) {
			if (dest_file != NULL) {
				fprintf(stderr, "Option -o supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -o.\n\n");
				usage(argv[0]);
				exit(1);
			}
			dest_file = strdup(argv[current_arg]);
		}
		// Verbosity
		else if (strcmp(argv[current_arg], "-v") == 0 || strcmp(argv[current_arg], "--verbose") == 0) {
			verbose = true;
		}
		// Projection string
		else if (strcmp(argv[current_arg], "-p") == 0 || strcmp(argv[current_arg], "--projection") == 0) {
			if (projection != NULL) {
				fprintf(stderr, "Option -p supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -p.\n\n");
				usage(argv[0]);
				exit(1);
			}
			projection = strdup(argv[current_arg]);
		}
		// Sub-viewer coordinates
		else if (strcmp(argv[current_arg], "--vL") == 0 || strcmp(argv[current_arg], "--viewer-longitude") == 0) {
			if (got_sub_viewer_longitude) {
				fprintf(stderr, "Option --vL supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --vL.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%lf", &sub_viewer_longitude);
			if (sub_viewer_longitude < -180 || sub_viewer_longitude > 360) {
				fprintf(stderr, "Sub-viewer longitude must be in the range [-180, 360].\n");
				usage(argv[0]);
				exit(1);
			}
			got_sub_viewer_longitude = true;
		}
		else if (strcmp(argv[current_arg], "--vl") == 0 || strcmp(argv[current_arg], "--viewer-latitude") == 0) {
			if (got_sub_viewer_latitude) {
				fprintf(stderr, "Option --vl supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --vl.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%lf", &sub_viewer_latitude);
			if (sub_viewer_latitude < -90 || sub_viewer_latitude > 90) {
				fprintf(stderr, "Sub-viewer latitude must be in the range [-90, 90].\n");
				usage(argv[0]);
				exit(1);
			}
			got_sub_viewer_latitude = true;
		}
		// Meters?
		else if (strcmp(argv[current_arg], "-m") == 0 || strcmp(argv[current_arg], "--meters") == 0) {
			DEM_values.meters = true;
		}
		// Full radii?
		else if (strcmp(argv[current_arg], "-r") == 0 || strcmp(argv[current_arg], "--radii") == 0) {
			DEM_values.full_radii = true;
		}
		// Colors
		else if (strcmp(argv[current_arg], "-S") == 0 || strcmp(argv[current_arg], "--shadow-color") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -S.\n\n");
				usage(argv[0]);
				exit(1);
			}
			char * shadow_color_string = argv[current_arg];
			// Parse hex
			int digits[8] = {0,0,0,0,0,0,0,0};
			for (int i = 0; i < 8; i++) {
				unsigned char this_char = shadow_color_string[i];
				if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
				if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
				else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
				else {
					fprintf(stderr, "Argument to -S must be an eight-digit hexidecimal number.\n");
					usage(argv[0]);
					exit(1);
				}
			}
			shadow_color.red =      16 * digits[0] + digits[1];
			shadow_color.green =    16 * digits[2] + digits[3];
			shadow_color.blue =     16 * digits[4] + digits[5];
			shadow_color.opacity =  16 * digits[6] + digits[7];
		}
		else if (strcmp(argv[current_arg], "-D") == 0 || strcmp(argv[current_arg], "--dark-color") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -D.\n\n");
				usage(argv[0]);
				exit(1);
			}
			char * minimum_color_string = argv[current_arg];
			// Parse hex
			int digits[8] = {0,0,0,0,0,0,0,0};
			for (int i = 0; i < 8; i++) {
				unsigned char this_char = minimum_color_string[i];
				if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
				if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
				else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
				else {
					fprintf(stderr, "Argument to -D must be an eight-digit hexidecimal number.\n");
					usage(argv[0]);
					exit(1);
				}
			}
			minimum_color.red =      16 * digits[0] + digits[1];
			minimum_color.green =    16 * digits[2] + digits[3];
			minimum_color.blue =     16 * digits[4] + digits[5];
			minimum_color.opacity =  16 * digits[6] + digits[7];
		}
		else if (strcmp(argv[current_arg], "-I") == 0 || strcmp(argv[current_arg], "--illuminated-color") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with -I.\n\n");
				usage(argv[0]);
				exit(1);
			}
			char * maximum_color_string = argv[current_arg];
			// Parse hex
			int digits[8] = {0,0,0,0,0,0,0,0};
			for (int i = 0; i < 8; i++) {
				unsigned char this_char = maximum_color_string[i];
				if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
				if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
				else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
				else {
					fprintf(stderr, "Argument to -I must be an eight-digit hexidecimal number.\n");
					usage(argv[0]);
					exit(1);
				}
			}
			maximum_color.red =      16 * digits[0] + digits[1];
			maximum_color.green =    16 * digits[2] + digits[3];
			maximum_color.blue =     16 * digits[4] + digits[5];
			maximum_color.opacity =  16 * digits[6] + digits[7];
		}
		// Use preloading?
		else if (strcmp(argv[current_arg], "--preload") == 0) {
			preload_cubes = true;
		}
		// Photometric function selection
		else if (strcmp(argv[current_arg], "--photo-function") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --photo-function.\n\n");
				usage(argv[0]);
				exit(1);
			}
			char * lowercase_photo_function = strdup(argv[current_arg]);
			for (char * lowercase = lowercase_photo_function; *lowercase; lowercase++) {
				*lowercase = tolower(*lowercase);
			}
			if (strcmp(lowercase_photo_function, "help") == 0) {
				printf("\nAvailable photometric functions\n-------------------\n");
				usage_line("'lambert':", "Use a basic Lambertian function.  Light intensity is proportional to cosine of incidence angle.  Viewer angle not taken into account.");
				usage_line("'lommel-seeliger':", "Apply Lommel-Seeliger law, cos(incidence) / (cos(incidence) + cos(emission)).");
				printf("\n");
				exit(0);
			}
			else if (strcmp(lowercase_photo_function, "lambert") == 0) {
				selected_photometric_function = &simple_lambert;
			}
			else if (strcmp(lowercase_photo_function, "lommel-seeliger") == 0) {
				selected_photometric_function = &lommel_seeliger;
			}
			else {
				fprintf(stderr, "Photometric function specifier '%s' unrecognized.  Use '--photo-function help' to see a list of available choices.\n", argv[current_arg]);
				exit(1);
			}
		}
		// Albedo map
		else if (strcmp(argv[current_arg], "--albedo-map") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --albedo-map.\n\n");
				usage(argv[0]);
				exit(1);
			}
			albedo_map_pathnames = realloc(albedo_map_pathnames, sizeof(char *) * (++albedo_map_pathname_count));
			albedo_map_pathnames[albedo_map_pathname_count - 1] = strdup(argv[current_arg]);
			using_albedo_map = true;
		}
		// Planetary characteristics
		else if (strcmp(argv[current_arg], "--radius") == 0) {
			if (radius_override) {
				fprintf(stderr, "Option --radius supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --radius.\n\n");
				usage(argv[0]);
				exit(1);
			}
			radius_override = true;
			sscanf(argv[current_arg], "%lf", &body_radius);
			if (body_radius <= 0) {
				fprintf(stderr, "Radius must be a positive number.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		else if (strcmp(argv[current_arg], "--solar-distance") == 0) {
			if (distance_to_sun_override) {
				fprintf(stderr, "Option --solar-distance supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --solar-distance.\n\n");
				usage(argv[0]);
				exit(1);
			}
			distance_to_sun_override = true;
			if (sscanf(argv[current_arg], "%lf", &distance_to_sun) != 1 || distance_to_sun <= 0) {
				fprintf(stderr, "Solar distance must be a positive number.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		// Assume smooth sphere in unknown-pixel scenario?
		else if (strcmp(argv[current_arg], "--assume-spherical") == 0) {
			assume_spherical = true;
		}
		else if (strcmp(argv[current_arg], "--variance") == 0) {
			if (elevation_variance_override) {
				fprintf(stderr, "Option --variance supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --variance.\n\n");
				usage(argv[0]);
				exit(1);
			}
			elevation_variance_override = true;
			sscanf(argv[current_arg], "%lf", &elevation_variance);
			if (elevation_variance <= 0) {
				fprintf(stderr, "Elevational variance must be a positive number.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		// Shadow-checking interval
		else if (strcmp(argv[current_arg], "--shadow-distance") == 0) {
			if (shadow_interval_override) {
				fprintf(stderr, "Option --shadow-distance supplied more than once.\n\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "No parameter supplied with --shadow-distance.\n\n");
				usage(argv[0]);
				exit(1);
			}
			sscanf(argv[current_arg], "%lf", &shadow_interval);
			if (shadow_interval <= 0) {
				fprintf(stderr, "Shadow-checking interval must be a positive number.\n");
				usage(argv[0]);
				exit(1);
			}
			shadow_interval_override = true;
		}
		// Whatcha talkin' bout, Willis?
		else {
			fprintf(stderr, "Option unrecognized: %s\n", argv[current_arg]);
			usage(argv[0]);
			exit(1);
		}
	}
	if (dem_pathname_count == 0) {
		fprintf(stderr, "Error: No input DEMs were specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (dest_file == NULL) {
		fprintf(stderr, "Error: No output filename was specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (projection == NULL) {
		// No output projection was given; default to equirectangular
		projection = strdup("+proj=latlon +a=1");
	}
	if (!(got_minimum_x && got_maximum_x && got_minimum_y && got_maximum_y)) {
		fprintf(stderr, "Error: Output boundaries (-x, -X, -y, -Y) were not completely specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (output_width == 0 && output_height == 0) {
		fprintf(stderr, "Error: Neither output width nor height was specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (!got_sub_solar_longitude) {
		fprintf(stderr, "Error: No sub-solar longitude was specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (!got_sub_solar_latitude) {
		fprintf(stderr, "Error: No sub-solar latitude was specified.\n");
		usage(argv[0]);
		exit(1);
	}
	if (!radius_override != !elevation_variance_override) {
		fprintf(stderr, "Error: If specifying non-default radius or elevational variance, the other should be specified as well.\n");
		usage(argv[0]);
		exit(1);
	}
	/* Now that we know we have an output size in at least one dimension, and
	 * the latitude/longitude constraints have been validated, compute the
	 * other dimension of size if needed */
	if (output_width == 0 && output_height != 0) {
		double units_per_pixel = (bbox.maxy - bbox.miny) / output_height;
		output_width = (bbox.maxx - bbox.minx) / units_per_pixel;
	}
	else if (output_width != 0 && output_height == 0) {
		double units_per_pixel = (bbox.maxx - bbox.minx) / output_width;
		output_height = (bbox.maxy - bbox.miny) / units_per_pixel;
	}
	// Projection objects
	projPJ pj_dest = pj_init_plus(projection);
	if (!pj_dest) {
		fprintf(stderr, "Error: Unable to generate output projection object.\n");
		exit(1);
	}
	projPJ pj_latlon = pj_init_plus("+proj=latlon +a=1");
	validate_bbox(projection, pj_dest, &bbox);

	// Compute destination scales
	double dest_scale_x = (bbox.maxx - bbox.minx) / output_width;
	double dest_scale_y = (bbox.miny - bbox.maxy) / output_height;
	/* If either sub-viewer coordinate was not specified, set it to the
	 * default, at the center of the output projection */
	if (!got_sub_viewer_longitude || !got_sub_viewer_latitude) {
		double default_longitude;
		double default_latitude;
		project_other_projection_to_latlon(projection, &default_longitude, &default_latitude, pj_latlon, pj_dest, 0, 0);
		if (!got_sub_viewer_longitude) sub_viewer_longitude = default_longitude;
		if (!got_sub_viewer_latitude)  sub_viewer_latitude  = default_latitude;
	}
	/* Choose the best DEM to use, and initialize the object, converting from
	 * km to meters if we have a radius override */
	SAY("Selecting a DEM...\n");
	char * dem_pathname = pick_a_cube(dem_pathname_count, dem_pathnames, projection, bbox, output_width, output_height);
	SAY("Selected %s, loading...\n", dem_pathname);
	// Use a 256 MB cache unless we're preloading the entire cube
	init_cube(&DEM_info, dem_pathname, (radius_override ? body_radius * 1000 : 0), (preload_cubes ? 0 : 256), preload_cubes, verbose);
	// If the user wants to use an albedo map, do the same for the choices
	if (using_albedo_map) {
		SAY("Selecting an albedo map...\n");
		char * albedo_map_pathname = pick_a_cube(albedo_map_pathname_count, albedo_map_pathnames, projection, bbox, output_width, output_height);
		SAY("Selected %s, loading...\n", albedo_map_pathname);
		// Use a 256 MB cache unless we're preloading the entire cube
		init_cube(&albedo_map, albedo_map_pathname, (radius_override ? body_radius * 1000 : 0), (preload_cubes ? 0 : 256), preload_cubes, verbose);
	}
	/* Now that the cube has been initialized, set default shadow-checking
	 * interval, if it wasn't overridden */
	if (!shadow_interval_override) {
		shadow_interval = DEM_info.meters_per_pixel * 5 / 1000;
	}
	// Determine rectangular-coordinate unit vector toward Sun
	double mean_solar_vector[3];
	mean_solar_vector[0] = cos(sub_solar_latitude * DEG_TO_RAD) * cos(sub_solar_longitude * DEG_TO_RAD);
	mean_solar_vector[1] = cos(sub_solar_latitude * DEG_TO_RAD) * sin(sub_solar_longitude * DEG_TO_RAD);
	mean_solar_vector[2] = sin(sub_solar_latitude * DEG_TO_RAD);
	// Determine location of Sun W.R.T. planetary body
	double sun_location[3];
	sun_location[0] = mean_solar_vector[0] * distance_to_sun;
	sun_location[1] = mean_solar_vector[1] * distance_to_sun;
	sun_location[2] = mean_solar_vector[2] * distance_to_sun;
	// Determine rectangular-coordinate unit vector toward viewer
	double viewer_vector[3];
	viewer_vector[0] = cos(sub_viewer_latitude * DEG_TO_RAD) * cos(sub_viewer_longitude * DEG_TO_RAD);
	viewer_vector[1] = cos(sub_viewer_latitude * DEG_TO_RAD) * sin(sub_viewer_longitude * DEG_TO_RAD);
	viewer_vector[2] = sin(sub_viewer_latitude * DEG_TO_RAD);
	// Build the output
	SAY("Computing output...\n");
	uint64_t output_size = output_width * output_height;
	uint32_t * buffer = calloc(output_size, sizeof (uint32_t));
	uint64_t pixels_finished = 0;
	int previous_percent = -1;
	bool *light_map = calloc(output_size, sizeof(bool));
	// Evaluate pixels that are closest to the sun first, and farthest from the sun last.
	struct pixel_distance_information *pixel_order = calloc(output_size, sizeof(struct pixel_distance_information));
	for (int y = 0; y < output_height; y++) {
		for (int x = 0; x < output_width; x++) {
			pixel_order[y * output_width + x].x = x;
			pixel_order[y * output_width + x].y = y;
			double longitude;
			double latitude;
			if (project_image_to_latlon(projection, &longitude, &latitude, pj_latlon, pj_dest, bbox, x, y, dest_scale_x, dest_scale_y)) {
				pixel_order[y * output_width + x].distance_from_sun = short_arc_length(longitude, latitude, sub_solar_longitude, sub_solar_latitude);
			}
			else {
				pixel_order[y * output_width + x].distance_from_sun = 0;
			}
		}
	}
	qsort(pixel_order, output_size, sizeof(struct pixel_distance_information), &closer_to_sun);
	for (int pixel_index = 0; pixel_index < output_size; pixel_index++) {
		int x = pixel_order[pixel_index].x;
		int y = pixel_order[pixel_index].y;
		// Calculate current geographic location
		double longitude = 0;
		double latitude = 0;
		if (project_image_to_latlon(projection, &longitude, &latitude, pj_latlon, pj_dest, bbox, x, y, dest_scale_x, dest_scale_y)) {
			validate_longitude(&longitude, &DEM_info);
			double spherical_normal[3];
			spherical_normal[0] = cos(latitude * DEG_TO_RAD) * cos(longitude * DEG_TO_RAD);
			spherical_normal[1] = cos(latitude * DEG_TO_RAD) * sin(longitude * DEG_TO_RAD);
			spherical_normal[2] = sin(latitude * DEG_TO_RAD);
			double zenith_location[3];
			zenith_location[0] = spherical_normal[0] * distance_to_sun;
			zenith_location[1] = spherical_normal[1] * distance_to_sun;
			zenith_location[2] = spherical_normal[2] * distance_to_sun;
			double sun_to_zenith[3];
			sun_to_zenith[0] = zenith_location[0] - sun_location[0];
			sun_to_zenith[1] = zenith_location[1] - sun_location[1];
			sun_to_zenith[2] = zenith_location[2] - sun_location[2];
			scale_to_unit(sun_to_zenith);
			/* Choose a Solar vector that is adjusted by one Sun-radius
			 * away from the mean Solar vector, in the direction of the
			 * current zenith.  This should be a sufficient approximation
			 * of a vector toward the "edge" of the Sun that's "highest in
			 * the sky" and thus most significant. */
			double solar_vector[3];
			solar_vector[0] = sun_location[0] + SOLAR_RADIUS * sun_to_zenith[0];
			solar_vector[1] = sun_location[1] + SOLAR_RADIUS * sun_to_zenith[1];
			solar_vector[2] = sun_location[2] + SOLAR_RADIUS * sun_to_zenith[2];
			scale_to_unit(solar_vector);
			// If we're outside of the DEM's boundaries, but less than a pixel (in input DEM space) away, clamp to the closest pixel
			if (longitude > DEM_info.longitude_maximum || longitude < DEM_info.longitude_minimum || latitude > DEM_info.latitude_maximum || latitude < DEM_info.latitude_minimum) {
				if (longitude < DEM_info.longitude_minimum && (DEM_info.longitude_minimum - longitude) * DEM_info.pixels_per_degree < 1) longitude = DEM_info.longitude_minimum;
				if (longitude > DEM_info.longitude_maximum && (longitude - DEM_info.longitude_maximum) * DEM_info.pixels_per_degree < 1) longitude = DEM_info.longitude_maximum;
				if (latitude  < DEM_info.latitude_minimum  && (DEM_info.latitude_minimum - latitude)   * DEM_info.pixels_per_degree < 1) latitude  = DEM_info.latitude_minimum;
				if (latitude  > DEM_info.latitude_maximum  && (latitude - DEM_info.latitude_maximum)   * DEM_info.pixels_per_degree < 1) latitude  = DEM_info.latitude_maximum;
			}
			/* If we're still out of bounds (the above check did not make a
			 * correction) transparent pixel unless '--assume-spherical' flag is given --
			 * in which case, assume non-shadowed and 0 elevation */
			if (longitude > DEM_info.longitude_maximum || longitude < DEM_info.longitude_minimum || latitude > DEM_info.latitude_maximum || latitude < DEM_info.latitude_minimum) {
				if (assume_spherical) {
					double surface_normal[3];
					surface_normal[0] = cos(latitude * DEG_TO_RAD) * cos(longitude * DEG_TO_RAD);
					surface_normal[1] = cos(latitude * DEG_TO_RAD) * sin(longitude * DEG_TO_RAD);
					surface_normal[2] = sin(latitude * DEG_TO_RAD);
					double phase_angle = angle_between(viewer_vector, solar_vector);
					double emission_angle = angle_between(surface_normal, viewer_vector);
					double incidence_angle = angle_between(surface_normal, solar_vector);
					double light_intensity = selected_photometric_function(phase_angle, emission_angle, incidence_angle);
					// If we're using an albedo map, multiply by its value at this location
					if (using_albedo_map) {
						double albedo_map_longitude = longitude;
						validate_longitude(&albedo_map_longitude, &albedo_map);
						if (albedo_map_longitude >= albedo_map.longitude_minimum && albedo_map_longitude <= albedo_map.longitude_maximum) {
							if (latitude >= albedo_map.latitude_minimum && latitude <= albedo_map.latitude_maximum) {
								light_intensity *= get_value_at_geographic_point(&albedo_map, albedo_map_longitude, latitude);
							}
						}
					}
					if (light_intensity <= 0) {
						// Use the minimum shade color
						char *byte_ptr = (char *)(&(buffer[y * output_width + x]));
						byte_ptr[0] = minimum_color.blue;
						byte_ptr[1] = minimum_color.green;
						byte_ptr[2] = minimum_color.red;
						byte_ptr[3] = minimum_color.opacity;
					}
					else {
						// Calculate the appropriate color
						char *byte_ptr = (char *)(&(buffer[y * output_width + x]));
						byte_ptr[0] = (uint8_t)((maximum_color.blue    - minimum_color.blue)    * light_intensity + minimum_color.blue);
						byte_ptr[1] = (uint8_t)((maximum_color.green   - minimum_color.green)   * light_intensity + minimum_color.green);
						byte_ptr[2] = (uint8_t)((maximum_color.red     - minimum_color.red)     * light_intensity + minimum_color.red);
						byte_ptr[3] = (uint8_t)((maximum_color.opacity - minimum_color.opacity) * light_intensity + minimum_color.opacity);
					}
				}
				else {
					memset(&(buffer[y * output_width + x]), 0, 4);
				}
			}
			else {
				// Is this pixel shadowed or unshadowed?
				determine_whether_shadowed(&DEM_info, DEM_values, light_map, shadow_interval, elevation_variance, solar_vector, longitude, latitude, projection, pj_latlon, pj_dest, bbox, output_width, output_height, dest_scale_x, dest_scale_y, x, y);
				if (!light_map[y * output_width + x]) {
					// Use the shadow color
					char *byte_ptr = (char *)(&(buffer[y * output_width + x]));
					byte_ptr[0] = shadow_color.blue;
					byte_ptr[1] = shadow_color.green;
					byte_ptr[2] = shadow_color.red;
					byte_ptr[3] = shadow_color.opacity;
				}
				else {
					// Not shadowed -- call the photometric function
					double surface_normal[3];
					unit_normal_vector(&DEM_info, longitude, latitude, surface_normal, DEM_values);
					double phase_angle = angle_between(viewer_vector, solar_vector);
					double emission_angle = angle_between(surface_normal, viewer_vector);
					double incidence_angle = angle_between(surface_normal, solar_vector);
					double light_intensity = selected_photometric_function(phase_angle, emission_angle, incidence_angle);
					// If we're using an albedo map, multiply by its value at this location
					if (using_albedo_map) {
						double albedo_map_longitude = longitude;
						validate_longitude(&albedo_map_longitude, &albedo_map);
						if (albedo_map_longitude >= albedo_map.longitude_minimum && albedo_map_longitude <= albedo_map.longitude_maximum) {
							if (latitude >= albedo_map.latitude_minimum && latitude <= albedo_map.latitude_maximum) {
								light_intensity *= get_value_at_geographic_point(&albedo_map, albedo_map_longitude, latitude);
							}
						}
					}
					if (light_intensity <= 0) {
						// Use the minimum shade color
						char *byte_ptr = (char *)(&(buffer[y * output_width + x]));
						byte_ptr[0] = minimum_color.blue;
						byte_ptr[1] = minimum_color.green;
						byte_ptr[2] = minimum_color.red;
						byte_ptr[3] = minimum_color.opacity;
					}
					else {
						// Calculate the appropriate color
						char *byte_ptr = (char *)(&(buffer[y * output_width + x]));
						byte_ptr[0] = (uint8_t)((maximum_color.blue    - minimum_color.blue)    * light_intensity + minimum_color.blue);
						byte_ptr[1] = (uint8_t)((maximum_color.green   - minimum_color.green)   * light_intensity + minimum_color.green);
						byte_ptr[2] = (uint8_t)((maximum_color.red     - minimum_color.red)     * light_intensity + minimum_color.red);
						byte_ptr[3] = (uint8_t)((maximum_color.opacity - minimum_color.opacity) * light_intensity + minimum_color.opacity);
					}
				}
			}
		}
		else {
			// Didn't back-project, so use a transparent pixel
			memset(&(buffer[y * output_width + x]), 0, 4);
		}
		pixels_finished++;
		int percent_finished = (int)(pixels_finished * 100 / output_size);
		if (percent_finished > previous_percent) {
			SAY("\r%d%% complete...", percent_finished);
			fflush(stdout);
			previous_percent = percent_finished;
		}
	}
	SAY("\nFinished.\n");
	// Write the output
	SAY("Writing output to %s...", dest_file);
	Imlib_Image img_out = imlib_create_image_using_copied_data(output_width, output_height, buffer);
	imlib_context_set_image(img_out);
	imlib_image_set_has_alpha('1');
	imlib_image_set_format("png");

	if (img_out == NULL) {
		printf("Imlib2 Failed to generate out image!\n");
		exit(1);
	}
	imlib_context_set_image(img_out);
	imlib_save_image(dest_file);
	SAY("\nFinished.\n");

	imlib_free_image();
	free(buffer);
	// Clean up
	free(light_map);
	free(pixel_order);
	free(projection);
	pj_free(pj_latlon);
	pj_free(pj_dest);
	for (int i = 0; i < dem_pathname_count; i++) {
		free(dem_pathnames[i]);
	}
	free(dem_pathnames);
	free_cube(&DEM_info);
	if (using_albedo_map) free_cube(&albedo_map);
	exit(0);
}
