/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_numeric_layer.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Produce VICAR output from ISIS cube file input
 */

#include <getopt.h>
#include <string.h>
#include <float.h>

#include "map_shared_code.h"

#define LABEL_SIZE  512  // Length in bytes to use for the label at the beginning of VICAR output
//Constants for CESIUM terrain output
#define INITIAL_X_TILES 2 // The number of tiles in zoom level 0
#define INITIAL_Y_TILES 1
#define START_TILE_LAT -90.00 //The UL corner of the first tile
#define START_TILE_LON -180.00

float get_value_for_dest_pixel(int x, int y, cube_properties * cube_info, const char* projection, BBox bbox, double scale, double base, projPJ pj_latlon, projPJ pj_dest, double dest_scale_x, double dest_scale_y) {
        // Calculate current geographic location
        double longitude = 0;
        double latitude = 0;
        // Try back-projecting into cube
        if (project_image_to_latlon(projection, &longitude, &latitude, pj_latlon, pj_dest, bbox, x, y, dest_scale_x, dest_scale_y)) {
                validate_longitude(&longitude, cube_info);
                // If we're outside of the cube's boundaries, but less than a pixel (in input cube space) away, clamp to the closest pixel
                if (longitude > cube_info->longitude_maximum || longitude < cube_info->longitude_minimum || latitude > cube_info->latitude_maximum || latitude < cube_info->latitude_minimum) {
                        if (longitude < cube_info->longitude_minimum && (cube_info->longitude_minimum - longitude) * cube_info->pixels_per_degree < 1) longitude = cube_info->longitude_minimum;
                        if (longitude > cube_info->longitude_maximum && (longitude - cube_info->longitude_maximum) * cube_info->pixels_per_degree < 1) longitude = cube_info->longitude_maximum;
                        if (latitude  < cube_info->latitude_minimum  && (cube_info->latitude_minimum - latitude)   * cube_info->pixels_per_degree < 1) latitude  = cube_info->latitude_minimum;
                        if (latitude  > cube_info->latitude_maximum  && (latitude - cube_info->latitude_maximum)   * cube_info->pixels_per_degree < 1) latitude  = cube_info->latitude_maximum;
                }
                // If we're still outside of the cube's boundaries, give this pixel a 0 value
                if (longitude > cube_info->longitude_maximum || longitude < cube_info->longitude_minimum || latitude > cube_info->latitude_maximum || latitude < cube_info->latitude_minimum) {
                        return 0.0;
                }
                else {
                        return get_value_at_geographic_point(cube_info, longitude, latitude) * scale + base;
                }
        }
        else {
                // We didn't back-project into the cube, so give this pixel a 0 value
                return 0.0;
        }
}

int cube_bbox_to_cesium(char * cube, cube_properties * cube_info, const char * output_file, const char* projection, BBox bbox, int dest_width, int dest_height, double scale, double base, bool highest_res) {
	// Prepare the output CESIUM file
	FILE * output = fopen(output_file, "w");
	if (output == NULL) {
		fprintf(stderr, "Error: could not open %s for writing!\n", output_file);
		exit(1);
	}

	// Prepare destination and latlon projection objects
	projPJ pj_latlon = pj_init_plus("+proj=latlong +a=1");
	if (pj_latlon == NULL) {
		fprintf(stderr, "Unable to init latlon projection.\n");
		exit(1);
	}
	projPJ pj_dest = pj_init_plus(projection);
	if (pj_dest == NULL) {
		fprintf(stderr, "Unable to init destination projection.\n");
		exit(1);
	}
	validate_bbox(projection, pj_dest, &bbox);

	// Compute scales
	double dest_scale_x = (bbox.maxx - bbox.minx) / dest_width; 
	double dest_scale_y = (bbox.miny - bbox.maxy) / dest_height;

	for (int y=0; y < dest_height; y++) {
		for (int x=0; x < dest_width; x++) {
			float px = get_value_for_dest_pixel(x, y, cube_info, projection, bbox, scale, base, pj_latlon, pj_dest, dest_scale_x, dest_scale_y);
			//Change from full radius, to an offset value
			px -= cube_info->radius;
			if (px == 0.0)
				for (int i = 0; i < 4; i++) fputc(0, output);
			else{
				char * write_byte = (char *)(&px);
				for (int i = 0; i < 4; i++) fputc(write_byte[i], output);
			}
		}
	}
	// If this is the highest resolution tile which we have, then 
	// we will set the last byte in the file, to let cesium know
	// not to request another tile.
	if (highest_res)
		fputc(0xff,output);
	else
		fputc(0x00,output);
	// Cleanup
	fclose(output);
	pj_free(pj_latlon);
	pj_free(pj_dest);
	return 0;
}

int cube_bbox_to_tif(char * cube, cube_properties * cube_info, const char * output_file, const char* projection, BBox bbox, int dest_width, int dest_height, double scale, double base) {
        // Open the input
        TIFF * output_tiff = TIFFOpen(output_file, "w");
        if (output_tiff == NULL) {
                fprintf(stderr, "Error: Unable to open %s for writing.\n", output_file);
                exit(1);
        }

        // Prepare destination and latlon projection objects
        projPJ pj_latlon = pj_init_plus("+proj=latlong +a=1");
        if (pj_latlon == NULL) {
                fprintf(stderr, "Unable to init latlon projection.\n");
                exit(1);
        }
        projPJ pj_dest = pj_init_plus(projection);
        if (pj_dest == NULL) {
                fprintf(stderr, "Unable to init destination projection.\n");
                exit(1);
        }

        // Compute scales
        double dest_scale_x = (bbox.maxx - bbox.minx) / dest_width;
        double dest_scale_y = (bbox.miny - bbox.maxy) / dest_height;

        // Set TIFF header info
        TIFFSetField(output_tiff, TIFFTAG_IMAGEWIDTH, dest_width);
        TIFFSetField(output_tiff, TIFFTAG_IMAGELENGTH, dest_height);
        TIFFSetField(output_tiff, TIFFTAG_BITSPERSAMPLE, 32);
        TIFFSetField(output_tiff, TIFFTAG_SAMPLESPERPIXEL, 1);
        TIFFSetField(output_tiff, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP);
        TIFFSetField(output_tiff, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(output_tiff, dest_width));
        TIFFSetField(output_tiff, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
        TIFFSetField(output_tiff, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
        TIFFSetField(output_tiff, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);

        // Build the TIFF
        for (int y=0; y < dest_height; y++) {
		float *line_data = malloc(dest_width * sizeof(float));
		memset(line_data, 0, (dest_width * sizeof(float)));
                for (int x=0; x < dest_width; x++) {
                        line_data[x] = get_value_for_dest_pixel(x, y, cube_info, projection, bbox, scale, base, pj_latlon, pj_dest, dest_scale_x, dest_scale_y);
                }
                TIFFWriteScanline(output_tiff, line_data, y, 0);
		free(line_data);
        }
        TIFFClose(output_tiff);
        return 0;
}

int cube_bbox_to_vicar(char * cube, cube_properties * cube_info, const char * output_file, const char* projection, BBox bbox, int dest_width, int dest_height, double scale, double base) {
	// Prepare the output VICAR file
	FILE * output = fopen(output_file, "w");
	if (output == NULL) {
		fprintf(stderr, "Error: could not open %s for writing!\n", output_file);
		exit(1);
	}

	// Prepare destination and latlon projection objects
	projPJ pj_latlon = pj_init_plus("+proj=latlong +a=1");
	if (pj_latlon == NULL) {
		fprintf(stderr, "Unable to init latlon projection.\n");
		exit(1);
	}
	projPJ pj_dest = pj_init_plus(projection);
	if (pj_dest == NULL) {
		fprintf(stderr, "Unable to init destination projection.\n");
		exit(1);
	}

	// Compute scales
	double dest_scale_x = (bbox.maxx - bbox.minx) / dest_width; 
	double dest_scale_y = (bbox.miny - bbox.maxy) / dest_height;

	// Create the label
	char label[LABEL_SIZE + 1];
	sprintf(label, "LBLSIZE=%i  INTFMT='LOW'  REALFMT='RIEEE'  FORMAT='REAL'  TYPE='IMAGE'  RECSIZE=%i  BUFSIZE=%i  DIM=3  EOL=0  ORG='BSQ'  NB=1  NL=%i  NS=%i  N3=1  N2=%i  N1=%i  NBB=0  NLB=0", LABEL_SIZE, 4 * dest_width, 4 * dest_width, dest_height, dest_width, dest_height, dest_width);

	// Find the end of the sprintf'd label and pad the rest of the label space with blanks, then write it
	bool found_null = false;
	for (int i = 0; i < LABEL_SIZE + 1; i++) {
		if (label[i] == '\0') found_null = true;
		if (found_null) label[i] = ' ';
		if (i == LABEL_SIZE) label[i] = '\0';
	}
	fprintf(output, "%s", label);

	for (int y=0; y < dest_height; y++) {
		for (int x=0; x < dest_width; x++) {
			float px = get_value_for_dest_pixel(x, y, cube_info, projection, bbox, scale, base, pj_latlon, pj_dest, dest_scale_x, dest_scale_y);
			if (px == 0.0)
				for (int i = 0; i < 4; i++) fputc(0, output);
			else{
				char * write_byte = (char *)(&px);
				for (int i = 0; i < 4; i++) fputc(write_byte[i], output);
			}
		}
	}
	// Cleanup
	fclose(output);
	pj_free(pj_latlon);
	pj_free(pj_dest);
	return 0;
}

// This function is used for Cesium output.
// It decides that if the Cesium terrain provider should request the next zoom level worth of terrain
// or if we are out of resolution
bool check_next_zoom_level(const char *current_best_cube, cube_properties cube_info, BBox bbox, int num_cubes, char **cubes, const char* projection, int dest_width, int dest_height){
	//find the highest resolution cub we have
	int highest_res_cube_index = -1;
	double highest_res_mpp = DBL_MAX;
	for (int i = 0; i < num_cubes; i++) {
		cube_properties test_cube_info;
		init_cube(&test_cube_info, cubes[i], 0, 32, false, false);
		if (test_cube_info.meters_per_pixel < highest_res_mpp){
			highest_res_cube_index = i;
			highest_res_mpp = test_cube_info.meters_per_pixel;
		}
		free_cube(&test_cube_info);
	}

	// Check to see if we're on the highest resolution cub yet
	if (strcmp(current_best_cube, cubes[highest_res_cube_index]) == 0)
		return false;
	
	// If we're here, we have to make sure that there is enough 
	// resolution in the current cub to pull another layer out of it
	// Figure out what the next zoom levels bbox will be
	BBox next_bbox = {bbox.minx, bbox.miny, bbox.minx + (bbox.maxx - bbox.minx)/2 , bbox.miny + (bbox.maxy - bbox.miny)/2};
	// And what the source cub's bbox is
	BBox cube_bbox = {cube_info.longitude_minimum, cube_info.longitude_maximum, cube_info.latitude_minimum, cube_info.latitude_maximum};

	projPJ pj_dest, pj_latlon;
	pj_latlon = pj_init_plus("+proj=latlong +a=1");
	if (pj_latlon == NULL){
		fprintf(stderr, "Fatal error: Failed to init latlon projection.\n");
		exit(1);
	}

	pj_dest = pj_init_plus(projection);
	if(pj_dest == NULL) {
		fprintf(stderr, "Fatal error: Failed to init dest projection.\n");
		exit(1);
	}

	//Check the number of unique pixels hit the source image from the next levels bbox, inside of our current dest image
	int orig_px_hit =resolution_quality(pj_latlon, projection, pj_dest,  "+proj=latlong", pj_latlon, next_bbox, cube_bbox, dest_width/2, dest_height/2, cube_info.width, cube_info.height);

	//Check how many pixels will hit in the next zoom level (same bbox, but full width and height)
	int next_level_px_hit =resolution_quality(pj_latlon, projection, pj_dest,  "+proj=latlong", pj_latlon, next_bbox, cube_bbox, dest_width, dest_height, cube_info.width, cube_info.height);


	//If the same number of unique pixels hit in the source image, then we are out of resolution in this image
	if (next_level_px_hit == orig_px_hit) 
		return true;
	else
		return false;
}

void usage(const char* progname) {
	printf(
			"%s: Create a TIFF, LROC-CESIUM or VICAR file containing a bbox from one or more possible ISIS cube input files\n\n"
			"Options:\n"
			"  --help, -h: show help\n"
			"  --projection, -p projection: projection to map in (proj4 format)\n"
			"  --minx, -x n;\n"
			"  --maxx, -X n;\n"
			"  --miny, -y n;\n"
			"  --maxy, -Y n: define the bounding box for the output image.\n"
			"  --scale, -s n: define a scaling factor by which to multiply values (input of 1 becomes n, 2 becomes 2 * n, etc.) (default is 1)\n"
			"  --base, -b n: define a base by which to offset values (input of 0 becomes n, 5 becomes 5 + n, etc.) (default is 0)\n"
			"                 Note: scale is applied first, example: input = 5, scale = 2, base = 10, result = 5 * 2 + 10 = 20\n"
			"  --dest_width, -W size: width of the destination image.\n" 
			"  --dest_height, -H size: height of the destination image.\n"
			"  --output, -o filename: file to render the image into. If given a tif extension, a tif will be generated.\n"
			"  --cube, -c name: cube input file;\n"
			"            may be specified multiple times.\n", progname);
}

int main(int argc, char **argv) { 
	int c, option_index;
	static struct option long_options[] =
	{
		{"help",        no_argument,       0, 'h'},
		{"projection",  required_argument, 0, 'p'},
		{"minx",        required_argument, 0, 'x'},
		{"maxx",        required_argument, 0, 'X'},
		{"miny",        required_argument, 0, 'y'},
		{"maxy",        required_argument, 0, 'Y'},
		{"scale",       required_argument, 0, 's'},
		{"base",        required_argument, 0, 'b'},
		{"dest_width",  required_argument, 0, 'W'},
		{"dest_height", required_argument, 0, 'H'},
		{"output",      required_argument, 0, 'o'},
		{"cube",        required_argument, 0, 'c'}
	};

	int num_cubes = 0;
	char ** cubes = NULL;

	char *projection = 0, *dest_file = 0;
	bool generate_vicar = false, generate_tiff = false, generate_cesium = false;
	int dest_width=0, dest_height=0;
	BBox bbox = {0,0,0,0};
	bool gotx=(false), gotX=(false), goty=(false), gotY=(false);
	double scale = 1;
	double base = 0;

	while (1) { 
		c = getopt_long(argc, argv, "hp:x:X:y:Y:s:b:W:H:o:c:", long_options, &option_index);

		if (c == -1) break;

		switch(c) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'p':
				projection = strdup(optarg);
				break;
			case 'x':
				sscanf(optarg, "%lf", &bbox.minx);
				gotx=(true);
				break;
			case 'X':
				sscanf(optarg, "%lf", &bbox.maxx);
				gotX=(true);
				break;
			case 'y':
				sscanf(optarg, "%lf", &bbox.miny);
				goty=(true);
				break;
			case 'Y':
				sscanf(optarg, "%lf", &bbox.maxy);
				gotY=(true);
				break;
			case 's':
				sscanf(optarg, "%lf", &scale);
				break;
			case 'b':
				sscanf(optarg, "%lf", &base);
				break;
			case 'W':
				dest_width = atoi(optarg);
				break;
			case 'H':
				dest_height = atoi(optarg);
				break;
			case 'o':
				dest_file = strdup(optarg);
				break;
			case 'c': {
				cubes = realloc(cubes, sizeof(char *) * ++num_cubes);
				cubes[num_cubes - 1] = strdup(optarg);
				break;
			}
		}
	}
	bool got_required = true;
	if (projection == 0)    {fprintf(stderr, "A projection must be specified.\n");             got_required = false;}
	if (dest_file == 0)     {fprintf(stderr, "An output file is required.\n");                 got_required = false;}
	if (gotx == false)      {fprintf(stderr, "BBox value minx must be specified.\n");          got_required = false;}
	if (gotX == false)      {fprintf(stderr, "BBox value maxx must be specified.\n");          got_required = false;}
	if (goty == false)      {fprintf(stderr, "BBox value miny must be specified.\n");          got_required = false;}
	if (gotY == false)      {fprintf(stderr, "BBox value maxy must be specified.\n");          got_required = false;}
	if (dest_width <= 0)    {fprintf(stderr, "The destination image width must be > 0\n");     got_required = false;}
	if (dest_height <= 0)   {fprintf(stderr, "The destination image height must be > 0\n");    got_required = false;}
	if (num_cubes == 0)     {fprintf(stderr, "At least one cube file must be supplied.\n");    got_required = false;}

	if (!got_required) {
		usage(argv[0]);
		exit(1);
	}

	// Test to see if we are going to be making a tiff instead of VICAR
	char * dest_test = strdup(dest_file);
	// Down-Case it
	for (int i = 0; i < strlen(dest_test); i++) {
		if (dest_test[i] >= 'A' && dest_test[i] <= 'Z') dest_test[i] += 'a' - 'A';
	}
	if (strstr(dest_test, ".tif") || strstr(dest_test, ".tiff")) 
		generate_tiff = true;
	else if (strstr(dest_test, ".terrain"))
		generate_cesium = true;
	else
		generate_vicar = true;
	free(dest_test);

	char * best_cube = pick_a_cube(num_cubes, cubes, projection, bbox, dest_width, dest_height);
	// Initialize the cube -- let the parser find the radius
	cube_properties cube_info;
	init_cube(&cube_info, best_cube, 0, 32, false, false);
	// Convert it
	if(generate_tiff)
		cube_bbox_to_tif(best_cube, &cube_info, dest_file, projection, bbox, dest_width, dest_height, scale, base);
	else if (generate_cesium){
		bool highest_res = check_next_zoom_level(best_cube, cube_info, bbox, num_cubes, cubes, projection, dest_width, dest_height);
		cube_bbox_to_cesium(best_cube, &cube_info, dest_file, projection, bbox, dest_width, dest_height, scale, base, highest_res);
	}
	else if (generate_vicar)
		cube_bbox_to_vicar(best_cube, &cube_info, dest_file, projection, bbox, dest_width, dest_height, scale, base);

	// Cleanup
	free_cube(&cube_info);

	if (projection != 0) free(projection);
	if (dest_file != 0) free(dest_file);
	for (int i = 0; i < num_cubes; i++) {
		free(cubes[i]);
	}
	if (num_cubes > 0) free(cubes);
	exit(0);
}
