/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_query_shapefile.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Query tool for shapefile layer
 *
 */

#include <stdio.h>
#include <string.h>

#include <ogr_srs_api.h>
#include <shapefil.h>

#include "map_shared_code.h"

#define CIRCLE_RESOLUTION 1000

void query_shapefile(int shapefile_count, char **shapefile_pathnames, double longitude, double latitude, double radius, char *proj_str, BBox output_bbox) {
	projPJ pj_latlon = pj_init_plus("+proj=latlon +a=1");
	if (pj_latlon == NULL) {
		fprintf(stderr, "Error: Unable to generate latlon projection object.\n");
		exit(1);
	}
	projPJ output_projection = pj_init_plus(proj_str);
	if (output_projection == NULL) {
		fprintf(stderr, "Error: Unable to generate output projection object.\n");
		exit(1);
	}
	// Work in an orthographic projection around the queried location
	char ortho_projection_str[1024] = "";
	projPJ ortho_projection = NULL;
	double circumference_ortho_x[CIRCLE_RESOLUTION];
	double circumference_ortho_y[CIRCLE_RESOLUTION];
	sprintf(ortho_projection_str, "+proj=ortho +lon_0=%lf +lat_0=%lf +ellps=WGS84", longitude, latitude);
	ortho_projection = pj_init_plus(ortho_projection_str);
	if (ortho_projection == NULL) {
		fprintf(stderr, "Error when trying to initialize orthographic projection object centered on (%lf, %lf).\n", longitude, latitude);
		exit(1);
	}
	// Construct a polygonal approximation of the circle
	double circumference_longitude[CIRCLE_RESOLUTION];
	double circumference_latitude[CIRCLE_RESOLUTION];
	for (int i = 0; i < CIRCLE_RESOLUTION; i++) {
		double bearing = (double)(i) * 360 / CIRCLE_RESOLUTION;
		find_destination(longitude, latitude, bearing, radius, &circumference_longitude[i], &circumference_latitude[i]);
	}
	/* Project into our ortho space -- in this function, the bbox doesn't
	 * actually matter for a destination of ortho, so we can just pass a phony
	 * bbox */
	BBox phony_bbox = {0, 0, 0, 0};
	for (int i = 0; i < CIRCLE_RESOLUTION; i++) {
		project_latlon_to_other_projection(ortho_projection_str, &circumference_ortho_x[i], &circumference_ortho_y[i], pj_latlon, ortho_projection, phony_bbox, circumference_longitude[i], circumference_latitude[i], false, 0, 0);
	}
	// YAML heading
	printf("---\n");
	printf("shapes:\n");
	int index_into_all_results = 0;
	// Process the shapefiles
	for (int shapefile_index = 0; shapefile_index < shapefile_count; shapefile_index++) {
		char *shapefile_pathname = shapefile_pathnames[shapefile_index];
		SHPHandle shapefile_handle = SHPOpen(shapefile_pathname, "rb");
		if (shapefile_handle == NULL) {
			fprintf(stderr, "Error: could not read shapefile %s.\n", shapefile_pathname);
			exit(1);
		}
		// By default, assume shapefile is in lon/lat
		char * shapefile_projection_str = "+proj=longlat +a=1";
		// Derive projection file name
		char shapefile_projection_filename[strlen(shapefile_pathname) + 1];
		strcpy(shapefile_projection_filename, shapefile_pathname);
		char * extension = &(shapefile_projection_filename[strlen(shapefile_pathname) - 3]);
		if (strcmp(extension, "shp") == 0) {
			strcpy(extension, "prj");
		}
		else if (strcmp(extension, "Shp") == 0) {
			strcpy(extension, "Prj");
		}
		else if (strcmp(extension, "SHP") == 0) {
			strcpy(extension, "PRJ");
		}
		// Check to see if the shapefile has an accompanying projection file
		FILE * shapefile_projection_file = fopen(shapefile_projection_filename, "r");
		if (shapefile_projection_file) {
			// How big is the projection file?
			fseek(shapefile_projection_file, 0, SEEK_END);
			int shapefile_projection_file_length = ftell(shapefile_projection_file);
			fseek(shapefile_projection_file, 0, SEEK_SET);
			if (shapefile_projection_file_length > 0) {
				// Allocate a big enough buffer for the WKT, and read it
				char * WKT = malloc(shapefile_projection_file_length + 1);
				if (WKT == NULL) {
					fprintf(stderr, "Error: Could not allocate memory for well-known text.\n");
					exit(1);
				}
				if (fgets(WKT, shapefile_projection_file_length, shapefile_projection_file) != WKT) {
					fprintf(stderr, "There was an error when attempting to read WKT from file %s\n", shapefile_projection_filename);
					exit(1);
				}
				fclose(shapefile_projection_file);
				/* Special case: this French phrase is confusing to the
				 * conversion routine, so if it's present, change it to
				 * English. */
				const char *FRENCH_PHRASE = "Plate_Carree";
				const char *ENGLISH_REPLACEMENT = "Equirectangular";
				if (strstr(WKT, FRENCH_PHRASE)) {
					char *new_WKT = malloc(strlen(WKT) + 32);
					char *writer = new_WKT;
					char *reader = WKT;
					while (*reader) {
						if (strncmp(reader, FRENCH_PHRASE, strlen(FRENCH_PHRASE)) == 0) {
							strcpy(writer, ENGLISH_REPLACEMENT);
							reader += strlen(FRENCH_PHRASE);
							writer += strlen(ENGLISH_REPLACEMENT);
						}
						else {
							*(writer++) = *(reader++);
						}
					}
					*writer = '\0';
					free(WKT);
					WKT = new_WKT;
				}
				// Translate WKT to proj4
				OGRSpatialReferenceH spatial_reference = OSRNewSpatialReference(NULL);
				/* OSRImportFromWkt modifies the WKT pointer, so we make a copy of
				 * it in order to be able to free its target later... */
				char * WKT_copy = WKT;
				OSRImportFromWkt(spatial_reference, &WKT);
				free(WKT_copy);
				OSRExportToProj4(spatial_reference, &shapefile_projection_str);
			}
		}
		// Initialize shapefile projection object from projection string
		projPJ shapefile_projection = pj_init_plus(shapefile_projection_str);
		if (shapefile_projection == NULL) {
			fprintf(stderr, "Error: Unable to generate source projection object.\n");
			exit(1);
		}
		DBFHandle database_handle = DBFOpen(shapefile_pathname, "rb");
		int shape_count, shape_type;
		SHPGetInfo(shapefile_handle, &shape_count, &shape_type, NULL, NULL);
		int * results = NULL;
		int result_count = 0;
		for (int shape_index = 0; shape_index < shape_count; shape_index++) {
			SHPObject * shape = SHPReadObject(shapefile_handle, shape_index);
			if (shape -> nSHPType == SHPT_POINT ||  shape -> nSHPType == SHPT_MULTIPOINT || shape -> nSHPType == SHPT_POINTZ || shape -> nSHPType == SHPT_MULTIPOINTZ || shape -> nSHPType == SHPT_POINTM || shape -> nSHPType == SHPT_MULTIPOINTM) {
				/* This shape is just (a) point(s); check to see whether any point
				 * lies within the radius */
				for (int vertex = 0; vertex < shape->nVertices; vertex++) {
					double proj_x = shape->padfX[vertex];
					double proj_y = shape->padfY[vertex];
					double point_longitude;
					double point_latitude;
					if (project_other_projection_to_latlon(shapefile_projection_str, &point_longitude, &point_latitude, pj_latlon, shapefile_projection, proj_x, proj_y)) {
						double arc_length = short_arc_length(longitude, latitude, point_longitude, point_latitude);
						if (arc_length <= radius) {
							// This point is in range; add it to the results collection
							results = realloc(results, sizeof(int) * ++result_count);
							results[result_count - 1] = shape_index;
						}
					}
				}
			}
			else if (shape -> nSHPType == SHPT_ARC || shape -> nSHPType == SHPT_POLYGON || shape -> nSHPType == SHPT_ARCZ || shape -> nSHPType == SHPT_POLYGONZ || shape -> nSHPType == SHPT_ARCM || shape -> nSHPType == SHPT_POLYGONM) {
				bool found_intersection = false;
				for (int part = 0; part < shape->nParts && !found_intersection; part++) {
					// Information about this part
					int first_vertex = shape -> panPartStart[part];
					int last_vertex = (part == shape -> nParts - 1 ? shape -> nVertices - 1 : shape -> panPartStart[part + 1] - 1);
					int num_vertices = last_vertex - first_vertex + 1;
					// Project the part into our ortho workspace
					double part_longitude[num_vertices];
					double part_latitude[num_vertices];
					double part_ortho_x[num_vertices];
					double part_ortho_y[num_vertices];
					bool in_projection = true;
					for (int vertex = 0; vertex < num_vertices; vertex++) {
						project_other_projection_to_latlon(shapefile_projection_str, &part_longitude[vertex], &part_latitude[vertex], pj_latlon, shapefile_projection, shape->padfX[first_vertex + vertex], shape->padfY[first_vertex + vertex]);
						if (!project_latlon_to_other_projection(ortho_projection_str, &part_ortho_x[vertex], &part_ortho_y[vertex], pj_latlon, ortho_projection, phony_bbox, part_longitude[vertex], part_latitude[vertex], false, 0, 0)) {
							in_projection = false;
							break;
						}
					}
					if (in_projection) {
						/* Find a bbox of the polygon to cut down on how many calls we
						 * need to make to point_in_poly */
						double minx = part_ortho_x[0];
						double maxx = part_ortho_x[0];
						double miny = part_ortho_y[0];
						double maxy = part_ortho_y[0];
						for (int vertex = 1; vertex < num_vertices; vertex++) {
							if (part_ortho_x[vertex] < minx) minx = part_ortho_x[vertex];
							if (part_ortho_x[vertex] > maxx) maxx = part_ortho_x[vertex];
							if (part_ortho_y[vertex] < miny) miny = part_ortho_y[vertex];
							if (part_ortho_y[vertex] > maxy) maxy = part_ortho_y[vertex];
						}
						for (int circle_point = 0; circle_point < CIRCLE_RESOLUTION && !found_intersection; circle_point++) {
							if (circumference_ortho_x[circle_point] < minx ||
									circumference_ortho_x[circle_point] > maxx ||
									circumference_ortho_y[circle_point] < miny ||
									circumference_ortho_y[circle_point] > maxy)
							{
								continue;
							}
							if (point_in_poly(circumference_ortho_x[circle_point], circumference_ortho_y[circle_point], num_vertices, part_ortho_x, part_ortho_y)) {
								found_intersection = true;
							}
						}
						for (int poly_point = 0; poly_point < num_vertices && !found_intersection; poly_point++) {
							if (short_arc_length(longitude, latitude, part_longitude[poly_point], part_latitude[poly_point]) <= radius) {
								found_intersection = true;
							}
						}
						if (found_intersection) {
							/* We found an intersection between the circle
							 * and the polygon; add this shape to the
							 * results list */
							results = realloc(results, sizeof(int) * ++result_count);
							results[result_count - 1] = shape_index;
						}
					}
				}
			}
			else if (shape -> nSHPType == SHPT_MULTIPATCH) {
				bool found_intersection = false;
				for (int part = 0; part < shape -> nParts && !found_intersection; part++) {
					// Information about this part
					int first_vertex = shape -> panPartStart[part];
					int last_vertex = (part == shape -> nParts - 1 ? shape -> nVertices - 1 : shape -> panPartStart[part + 1] - 1);
					int num_vertices = last_vertex - first_vertex + 1;
					double part_longitude[num_vertices];
					double part_latitude[num_vertices];
					double part_ortho_x[num_vertices];
					double part_ortho_y[num_vertices];
					bool in_projection = true;
					for (int vertex = 0; vertex < num_vertices; vertex++) {
						project_other_projection_to_latlon(shapefile_projection_str, &part_longitude[vertex], &part_latitude[vertex], pj_latlon, shapefile_projection, shape->padfX[first_vertex + vertex], shape->padfY[first_vertex + vertex]);
						if (!project_latlon_to_other_projection(ortho_projection_str, &part_ortho_x[vertex], &part_ortho_y[vertex], pj_latlon, ortho_projection, phony_bbox, part_longitude[vertex], part_latitude[vertex], false, 0, 0)) {
							in_projection = false;
							break;
						}
					}
					if (in_projection) {
						// Tristrip
						if (shape -> panPartType[part] == SHPP_TRISTRIP) {
							// Consider each triangle in the strip
							for (int triangle = 0; triangle < num_vertices - 2 && !found_intersection; triangle++) {
								double x[3] = {part_ortho_x[triangle], part_ortho_x[triangle + 1], part_ortho_x[triangle + 2]};
								double y[3] = {part_ortho_y[triangle], part_ortho_y[triangle + 1], part_ortho_y[triangle + 2]};
								/* Find a bbox of the polygon to cut down on how many calls we
								 * need to make to point_in_poly */
								double minx = x[0];
								double maxx = x[0];
								double miny = y[0];
								double maxy = y[0];
								for (int vertex = 1; vertex < 3; vertex++) {
									if (x[vertex] < minx) minx = x[vertex];
									if (x[vertex] > maxx) maxx = x[vertex];
									if (y[vertex] < miny) miny = y[vertex];
									if (y[vertex] > maxy) maxy = y[vertex];
								}
								for (int circle_point = 0; circle_point < CIRCLE_RESOLUTION && !found_intersection; circle_point++) {
									if (circumference_ortho_x[circle_point] < minx ||
											circumference_ortho_x[circle_point] > maxx ||
											circumference_ortho_y[circle_point] < miny ||
											circumference_ortho_y[circle_point] > maxy)
									{
										continue;
									}
									if (point_in_poly(circumference_ortho_x[circle_point], circumference_ortho_y[circle_point], 3, x, y)) {
										found_intersection = true;
									}
								}
								for (int poly_point = 0; poly_point < num_vertices && !found_intersection; poly_point++) {
									if (short_arc_length(longitude, latitude, part_longitude[poly_point], part_latitude[poly_point]) <= radius) {
										found_intersection = true;
									}
								}
								if (found_intersection) {
									/* We found an intersection between the circle
									 * and the polygon; add this shape to the
									 * results list */
									results = realloc(results, sizeof(int) * ++result_count);
									results[result_count - 1] = shape_index;
								}
							}
						}
						// Trifan
						if (shape -> panPartType[part] == SHPP_TRIFAN) {
							// Consider each triangle in the fan
							for (int triangle = 0; triangle < num_vertices - 2 && !found_intersection; triangle++) {
								double x[3] = {part_ortho_x[0], part_ortho_x[triangle + 1], part_ortho_x[triangle + 2]};
								double y[3] = {part_ortho_y[0], part_ortho_y[triangle + 1], part_ortho_y[triangle + 2]};
								/* Find a bbox of the polygon to cut down on how many calls we
								 * need to make to point_in_poly */
								double minx = x[0];
								double maxx = x[0];
								double miny = y[0];
								double maxy = y[0];
								for (int vertex = 1; vertex < 3; vertex++) {
									if (x[vertex] < minx) minx = x[vertex];
									if (x[vertex] > maxx) maxx = x[vertex];
									if (y[vertex] < miny) miny = y[vertex];
									if (y[vertex] > maxy) maxy = y[vertex];
								}
								for (int circle_point = 0; circle_point < CIRCLE_RESOLUTION && !found_intersection; circle_point++) {
									if (circumference_ortho_x[circle_point] < minx ||
											circumference_ortho_x[circle_point] > maxx ||
											circumference_ortho_y[circle_point] < miny ||
											circumference_ortho_y[circle_point] > maxy)
									{
										continue;
									}
									if (point_in_poly(circumference_ortho_x[circle_point], circumference_ortho_y[circle_point], 3, x, y)) {
										found_intersection = true;
									}
								}
								for (int poly_point = 0; poly_point < num_vertices && !found_intersection; poly_point++) {
									if (short_arc_length(longitude, latitude, part_longitude[poly_point], part_latitude[poly_point]) <= radius) {
										found_intersection = true;
									}
								}
								if (found_intersection) {
									/* We found an intersection between the circle
									 * and the polygon; add this shape to the
									 * results list */
									results = realloc(results, sizeof(int) * ++result_count);
									results[result_count - 1] = shape_index;
								}
							}
						}
						// Rings (polygons)
						if (shape -> panPartType[part] == SHPP_RING || shape -> panPartType[part] == SHPP_OUTERRING || shape -> panPartType[part] == SHPP_INNERRING || shape -> panPartType[part] == SHPP_FIRSTRING) {
							for (int part = 0; part < shape->nParts && !found_intersection; part++) {
								// Project the part into our ortho workspace
								double part_longitude[num_vertices];
								double part_latitude[num_vertices];
								double part_ortho_x[num_vertices];
								double part_ortho_y[num_vertices];
								bool in_projection = true;
								for (int vertex = 0; vertex < num_vertices; vertex++) {
									project_other_projection_to_latlon(shapefile_projection_str, &part_longitude[vertex], &part_latitude[vertex], pj_latlon, shapefile_projection, shape->padfX[first_vertex + vertex], shape->padfY[first_vertex + vertex]);
									if (!project_latlon_to_other_projection(ortho_projection_str, &part_ortho_x[vertex], &part_ortho_y[vertex], pj_latlon, ortho_projection, phony_bbox, part_longitude[vertex], part_latitude[vertex], false, 0, 0)) {
										in_projection = false;
										break;
									}
								}
								if (in_projection) {
									/* Find a bbox of the polygon to cut down on how many calls we
									 * need to make to point_in_poly */
									double minx = part_ortho_x[0];
									double maxx = part_ortho_x[0];
									double miny = part_ortho_y[0];
									double maxy = part_ortho_y[0];
									for (int vertex = 1; vertex < num_vertices; vertex++) {
										if (part_ortho_x[vertex] < minx) minx = part_ortho_x[vertex];
										if (part_ortho_x[vertex] > maxx) maxx = part_ortho_x[vertex];
										if (part_ortho_y[vertex] < miny) miny = part_ortho_y[vertex];
										if (part_ortho_y[vertex] > maxy) maxy = part_ortho_y[vertex];
									}
									for (int circle_point = 0; circle_point < CIRCLE_RESOLUTION && !found_intersection; circle_point++) {
										if (circumference_ortho_x[circle_point] < minx ||
												circumference_ortho_x[circle_point] > maxx ||
												circumference_ortho_y[circle_point] < miny ||
												circumference_ortho_y[circle_point] > maxy)
										{
											continue;
										}
										if (point_in_poly(circumference_ortho_x[circle_point], circumference_ortho_y[circle_point], num_vertices, part_ortho_x, part_ortho_y)) {
											found_intersection = true;
										}
									}
									for (int poly_point = 0; poly_point < num_vertices && !found_intersection; poly_point++) {
										if (short_arc_length(longitude, latitude, part_longitude[poly_point], part_latitude[poly_point]) <= radius) {
											found_intersection = true;
										}
									}
									if (found_intersection) {
										/* We found an intersection between the circle
										 * and the polygon; add this shape to the
										 * results list */
										results = realloc(results, sizeof(int) * ++result_count);
										results[result_count - 1] = shape_index;
									}
								}
							}
						}
					}
				}
			}
			SHPDestroyObject(shape);
		}
		for (int result = 0; result < result_count; result++) {
			printf("  - index: %d\n", index_into_all_results++);
			SHPObject * shape = SHPReadObject(shapefile_handle, results[result]);
			// Bounding box in specified projection
			int num_vertices = shape->nVertices;
			double shape_longitude[num_vertices];
			double shape_latitude[num_vertices];
			double shape_output_x[num_vertices];
			double shape_output_y[num_vertices];
			bool in_projection = true;
			for (int vertex = 0; vertex < num_vertices; vertex++) {
				project_other_projection_to_latlon(shapefile_projection_str, &shape_longitude[vertex], &shape_latitude[vertex], pj_latlon, shapefile_projection, shape->padfX[vertex], shape->padfY[vertex]);
				if (!project_latlon_to_other_projection(proj_str, &shape_output_x[vertex], &shape_output_y[vertex], pj_latlon, output_projection, output_bbox, shape_longitude[vertex], shape_latitude[vertex], false, 0, 0)) {
					in_projection = false;
				}
			}
			double output_proj_minx = shape_output_x[0];
			double output_proj_maxx = shape_output_x[0];
			double output_proj_miny = shape_output_y[0];
			double output_proj_maxy = shape_output_y[0];
			for (int vertex = 0; vertex < num_vertices; vertex++) {
				if (shape_output_x[vertex] < output_proj_minx) output_proj_minx = shape_output_x[vertex];
				if (shape_output_x[vertex] > output_proj_maxx) output_proj_maxx = shape_output_x[vertex];
				if (shape_output_y[vertex] < output_proj_miny) output_proj_miny = shape_output_y[vertex];
				if (shape_output_y[vertex] > output_proj_maxy) output_proj_maxy = shape_output_y[vertex];
			}
			printf("    :bbox:\n");
			if (in_projection) {
				printf("    - %lf\n", output_proj_minx);
				printf("    - %lf\n", output_proj_miny);
				printf("    - %lf\n", output_proj_maxx);
				printf("    - %lf\n", output_proj_maxy);
			}
			else {
				printf("    - not_in_projection\n");
				printf("    - not_in_projection\n");
				printf("    - not_in_projection\n");
				printf("    - not_in_projection\n");
			}
			if (database_handle) {
				// Print out the database record content for the shape
				char field_name[12];
				for (int field = 0; field < DBFGetFieldCount(database_handle); field++) {
					bool field_is_string_data = DBFGetFieldInfo(database_handle, field, field_name, NULL, NULL) == FTString;
					printf(field_is_string_data ? "    %s: \"%s\"\n" : "    %s: %s\n", field_name, DBFReadStringAttribute(database_handle, results[result], field));
				}
			}
			SHPDestroyObject(shape);
		}
		// Cleanup
		free(results);
		if (database_handle) DBFClose(database_handle);
		SHPClose(shapefile_handle);
	}
	// Cleanup
	pj_free(ortho_projection);
}

void usage(const char* progname) {
	printf(
			"%s: Query one or more shapefiles to find out which features lie in a region.\n\n"
			"Options (all measurements are in degrees):\n\n"
			"-s <shapefile>:  Path to shapefile.  May  be specified more than once\n"
			"-L <longitude>:  Longitude of center of region to query\n"
			"-l <latitude>:   Latitude of center of region to query\n"
			"-r <radius>:     Radius of region to query\n"
			"-p <proj4>:      Proj.4 string in which bounding boxes should be reported\n"
			"-x <minx>:       Leftmost X boundary for bounding box projection space\n"
			"-X <maxx>:       Rightmost X boundary for bounding box projection space\n"
			"-y <miny>:       Lower Y boundary for bounding box projection space\n"
			"-Y <maxy>:       Upper Y boundary for bounding box projection space\n\n", progname);
}

int main(int argc, char **argv) {

	bool got_shapefile_pathnames = false;
	char **shapefile_pathnames = NULL;
	int shapefile_count = 0;

	bool got_longitude = false, got_latitude = false, got_radius = false;
	double longitude, latitude, radius;

	bool got_proj_str = false;
	char *proj_str = NULL;

	bool got_minx = false, got_maxx = false, got_miny = false, got_maxy = false;
	BBox bbox;

	for (int current_arg = 1; current_arg < argc; current_arg++) {
		if (strcmp(argv[current_arg], "-s") == 0) {
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-s option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			got_shapefile_pathnames = true;
			shapefile_pathnames = realloc(shapefile_pathnames, ++shapefile_count * sizeof(char *));
			shapefile_pathnames[shapefile_count - 1] = strdup(argv[current_arg]);
		}
		else if (strcmp(argv[current_arg], "-L") == 0) {
			if (got_longitude) {
				fprintf(stderr, "-L option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-L option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			got_longitude = true;
			if (sscanf(argv[current_arg], "%lf", &longitude) != 1) {
				fprintf(stderr, "Argument to -L must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		else if (strcmp(argv[current_arg], "-l") == 0) {
			if (got_latitude) {
				fprintf(stderr, "-l option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-l option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			got_latitude = true;
			if (sscanf(argv[current_arg], "%lf", &latitude) != 1) {
				fprintf(stderr, "Argument to -l must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		else if (strcmp(argv[current_arg], "-r") == 0) {
			if (got_radius) {
				fprintf(stderr, "-r option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-r option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			got_radius = true;
			if (sscanf(argv[current_arg], "%lf", &radius) != 1) {
				fprintf(stderr, "Argument to -r must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
		}
		else if (strcmp(argv[current_arg], "-p") == 0) {
			if (got_proj_str) {
				fprintf(stderr, "-p option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-p option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			got_proj_str = true;
			proj_str = strdup(argv[current_arg]);
		}
		else if (strcmp(argv[current_arg], "-x") == 0) {
			if (got_minx) {
				fprintf(stderr, "-x option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-x option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.minx) != 1) {
				fprintf(stderr, "Argument to -x must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_minx = true;
		}
		else if (strcmp(argv[current_arg], "-X") == 0) {
			if (got_maxx) {
				fprintf(stderr, "-X option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-X option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.maxx) != 1) {
				fprintf(stderr, "Argument to -X must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_maxx = true;
		}
		else if (strcmp(argv[current_arg], "-y") == 0) {
			if (got_miny) {
				fprintf(stderr, "-y option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-y option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.miny) != 1) {
				fprintf(stderr, "Argument to -y must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_miny = true;
		}
		else if (strcmp(argv[current_arg], "-Y") == 0) {
			if (got_maxy) {
				fprintf(stderr, "-Y option was specified twice.\n");
				usage(argv[0]);
				exit(1);
			}
			current_arg++;
			if (current_arg >= argc) {
				fprintf(stderr, "-Y option requires an argument.\n");
				usage(argv[0]);
				exit(1);
			}
			if (sscanf(argv[current_arg], "%lf", &bbox.maxy) != 1) {
				fprintf(stderr, "Argument to -Y must be numerical.\n");
				usage(argv[0]);
				exit(1);
			}
			got_maxy = true;
		}
		else {
			fprintf(stderr, "Error: option unrecognized: %s\n", argv[current_arg]);
			usage(argv[0]);
			exit(1);
		}
	}

	if (!got_shapefile_pathnames ||
			!got_longitude ||
			!got_latitude ||
			!got_radius ||
			!got_proj_str ||
			!got_minx ||
			!got_maxx ||
			!got_miny ||
			!got_maxy)
	{
		fprintf(stderr, "All parameters must be specified.\n");
		usage(argv[0]);
		exit(1);
	}

	query_shapefile(shapefile_count, shapefile_pathnames, longitude, latitude, radius, proj_str, bbox);

	for (int i = 0; i < shapefile_count; i++) {
		free(shapefile_pathnames[i]);
	}
	free(shapefile_pathnames);
	free(proj_str);

	exit(0);
}
