/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_raster_layer.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Mapserver Raster layer using proj4, supporting both full basemaps and smaller "stamps"
 *
 * Magic JMARS proj4 string: From proj4_test.rb 
 *+proj=ob_tran +o_proj=eqc +o_lat_p=#{lat0} +o_lon_p=0.0 +lon_0=#{lon0} +R=57.2957795131 
 *
 */

// #define _GNU_SOURCE

#include <errno.h>
#include <assert.h>
#include <getopt.h>
#include <string.h>
#include <pthread.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/queue.h>

#include <gdal.h>
#include <ogr_srs_api.h>

#include "map_shared_code.h"

#define IS_UPPER(x)                     (x >= 'A' && x <= 'Z')
#define FILENAME_BUFFER_LENGTH          4096 // How much memory to allow for each file pathname on stdin

/* We'll use this to store the original TIFF warning handler after setting our
 * custom one */
TIFFErrorHandler original_tiff_warning_handler;

void suppress_geotiff_complaints(const char* module, const char* fmt, va_list ap) {
	/* Instead of getting (an) integer code(s) unambiguously signifying the
	 * error/warning type, we just get an error message string and have to
	 * analyze it.  Thanks libtiff! */
	char assembled_error_message[4096];
	sprintf(assembled_error_message, fmt, ap);
	char *unknown_field_section = strstr(assembled_error_message, "unknown field with tag");
	if (unknown_field_section) {
		// Ignore any GeoTIFF tags!
		int tag_type;
		if (sscanf(unknown_field_section, "unknown field with tag %d ", &tag_type) == 1) {
			// The following codes all signify GeoTIFF tags.
			if (tag_type == 33550) return;
			if (tag_type == 33922) return;
			if (tag_type == 34735) return;
			if (tag_type == 34736) return;
			if (tag_type == 34737) return;
		}
	}
	// Let the original handler react to the problem, if we didn't suppress it
	original_tiff_warning_handler(module, fmt, ap);
}

// strrstr, to strstr as strrchr is to strchr
char * strrstr (char * haystack, char * needle) {
    if (strlen(needle) > strlen(haystack)) return NULL;
    for (char * search = haystack + strlen(haystack) - strlen(needle); search >= haystack; search--) {
        if (strncmp(search, needle, strlen(needle)) == 0) return search;
    }
    return NULL;
}

int project_basemaps(Basemap* basemaps, const char* dest_file, const char* projection, BBox bbox, int dest_width, int dest_height) {
    Imlib_Image img_out = NULL;
    DATA32 *buffer, *bp, value;
    projPJ pj_dest, pj_latlon;

    pj_latlon = pj_init_plus("+proj=latlon +a=1");
    pj_dest = pj_init_plus(projection);

    validate_bbox(projection, pj_dest, &bbox);

    double dest_scale_x =  (((double) (bbox.maxx-bbox.minx))/dest_width); 
    double dest_scale_y =  (((double) (bbox.miny-bbox.maxy))/dest_height);

    int base_x = -1, base_y = -1;

    buffer = malloc(dest_width * dest_height * sizeof(*buffer));
    bp = buffer;
    Basemap* current_basemap = basemaps;

    for (int y=0; y< dest_height; y++) {
        for (int x=0; x<dest_width; x++) {
            double longitude;
            double latitude;
            bool good_point = project_image_to_latlon(projection, &longitude, &latitude, pj_latlon, pj_dest, bbox, x, y, dest_scale_x, dest_scale_y);
            if (good_point) {
                // Clamp coordinates
                if (latitude >= 90) latitude = 89.9999;
                if (latitude <= -90) latitude= -89.9999;
                if (longitude >= 180) longitude -= 360;
                if (longitude <= -180) longitude = -179.9999;
                current_basemap = basemaps;
                while (current_basemap != 0) {
                    if (project_latlon_to_image(current_basemap->projection_string, &base_x, &base_y, pj_latlon, current_basemap->projection, current_basemap->extents, longitude, latitude, false, current_basemap->scalex, current_basemap->scaley)) {
                        // Check to see if we are within the correct basemap tile
                        if (base_x >= 0 && base_x < current_basemap->width && base_y >= 0 && base_y < current_basemap->height) {
                            // Only paint here if the basemap is non-masked or the mask says we should
                            if (current_basemap->mask == NULL || get_value(current_basemap->mask, base_x, base_y)) {
                                value = get_value(current_basemap, base_x, base_y);
                                *bp = value;
                                // We found the pixel we want, so we don't need to keep looking
                                break;
                            }
                            else *bp = 0x00000000; // Transparent pixel
                        }
                    }
                    // Pixel wasn't in this basemap tile; try the next
                    current_basemap = current_basemap->next;
                    // Use transparent pixels for areas not covered by the basemaps
                    if (current_basemap == 0) *bp = 0x00000000;
                }
            }
            else *bp = 0x00000000; // Transparent pixel
            bp++;
        }
    }
    // Write image out into destination file
    img_out = imlib_create_image_using_copied_data(dest_width, dest_height, buffer);
    imlib_context_set_image(img_out);
    imlib_image_set_has_alpha('1');
    imlib_image_set_format("png");

    if (img_out == NULL) {
        printf("Imlib2 Failed to generate out image!\n");
        return 1;
    }
    imlib_context_set_image(img_out);
    imlib_save_image(dest_file);

    imlib_free_image();
    free(buffer);
    pj_free(pj_dest);
    pj_free(pj_latlon);
    return 0;
}

int load_extents_and_mask(Basemap * basemap, char * projection, BBox bbox, int dest_width, int dest_height) {
    bool got_geography = false;
    // If there is an extents file, use it
    FILE *ext_file;
    char *ext_name, *cp;
    int ext_lines= 0;
    if ((ext_name = strdup(basemap->name)) != NULL){
        cp = strrchr(ext_name, '.');
        if (cp != NULL) {
            strcpy(cp, IS_UPPER(cp[1]) ? ".EXT" : ".ext");
            ext_file=(fopen(ext_name, "r"));
            if (ext_file != NULL) {
                char proj_buffer[1024];
                // Default projection for basemap
                strncpy(proj_buffer, "+proj=latlon +a=1", sizeof(proj_buffer));
                ext_lines=fscanf(ext_file, "%lf\n%lf\n%lf\n%lf\n%[^\n]\n", &basemap->extents.minx, &basemap->extents.maxx, &basemap->extents.miny, &basemap->extents.maxy, proj_buffer);
                basemap->projection = pj_init_plus(proj_buffer);
                basemap->projection_string = strdup(proj_buffer);
                fclose(ext_file);
                free(ext_name);
                if (ext_lines < 4){
                    fprintf(stderr, "Failed to read at least 4 lines from extents file.\n");
                    return 1;
                }
                got_geography = true;
            }
        }
    }
    if (!got_geography && !basemap->got_gtiff_tags) {
        // Neither method worked (the other method is the initial method of pulling all possible GeoTIFF tags before this threading stuff
        fprintf(stderr, "Basemap geography not known for %s -- extents file could not be found or could not be open for reading, and GeoTIFF tags could not be read from basemap image file's header.\n", basemap->name);
        return 1;
    }
    // Check for a mask
    char mask_file[strlen(basemap->name) + 10]; // Original file name plus some room to expand; we go a little overboard just to err on the safe side
    strcpy(mask_file, basemap->name);
    if (strrstr(mask_file, ".pyr.") || strrstr(mask_file, ".PYR.")) {
        char * pyr_part = strrstr(mask_file, ".pyr.");
        char * mask_extension = ".mask.";
        if (pyr_part == NULL) {
            pyr_part = strrstr(mask_file, ".PYR.");
            mask_extension = ".MASK.";
        }
        // Save the rest
        char * rest_of_extension = strdup(pyr_part + 5);
        // Replace ".pyr." with ".mask."
        strcpy(pyr_part, mask_extension);
        // Add the rest back in
        strcpy(pyr_part + 6, rest_of_extension);
        // Clean up mess
        free(rest_of_extension);
        // Make sure the mask exists and can be read
        FILE * check_readability = fopen(mask_file, "r");
        if (check_readability == NULL) {
            fprintf(stderr, "Warning: filename %s contains .pyr./.PYR., but mask file %s does not exist or could not be read, thus not using a mask.\n", basemap->name, mask_file);
        }
        else {
            fclose(check_readability);
            // Register mask
            basemap->mask = register_basemap(mask_file);
            // Copy mask mapping from basemap
            basemap->mask->projection_string = strdup(basemap->projection_string);
            basemap->mask->projection = pj_init_plus(basemap->mask->projection_string);
            basemap->mask->extents = basemap->extents;
            // Load basemap image data
            return load_basemap(basemap->mask, projection, bbox, dest_width, dest_height);
        }
    }

    // Success
    return 0;
}

void load_and_project_basemaps(unsigned int basemap_count, char ** basemap_names, char * dest_file, char * projection, BBox bbox, int dest_width, int dest_height) {
    if (basemap_count == 0) {
        // No basemaps were given on command line, need to read them from stdin
        // Line buffer
        char basemap_name[FILENAME_BUFFER_LENGTH];
        // Read from stdin line-by-line until there's no more
        while (fgets(basemap_name, FILENAME_BUFFER_LENGTH, stdin)) {
            // Ignore a trailing "partial line", and chop the newline character away from a full line
            if (basemap_name[strlen(basemap_name) - 1] == '\n') {
                basemap_name[strlen(basemap_name) - 1] = 0;
                basemap_names = realloc(basemap_names, ++basemap_count * sizeof(char *));
                basemap_names[basemap_count - 1] = strdup(basemap_name);
            }
        }

        if(ferror(stdin)) {
            // 'strerror' is not thread-safe since it works on an internal static buffer
            fprintf(stderr, "Error: %s\n", strerror(errno));
            exit(1);
        }
    }

    int i = 0;
    Basemap * last_basemap = NULL;
    Basemap * first_basemap = NULL;
    struct rlimit file_count_limit;

    // Check how many file descriptors we can handle
    if (getrlimit(RLIMIT_NOFILE, &file_count_limit) != 0) {
        fprintf(stderr, "Error: Call to getrlimit() to determine maximum file descriptor count failed.\n");
        exit(1);
    }

    // Register all known configured GDAL drivers
    GDALAllRegister();

    // Register all basemaps and link them
    for(i = 0; i < basemap_count; i++) {
        // If we're getting too close to the open file count limit, stop
        unsigned int open_files = 0;
        struct stat file_info;
        for (unsigned int file = 0; file < file_count_limit.rlim_cur; file++) {
            if (fstat(file, &file_info) == 0) open_files++;
            else if (errno != EBADF) open_files++;
        }
        if (file_count_limit.rlim_cur - open_files <= 4) {
            fprintf(stderr, "Warning: Too many files given as input; only using those preceding %s.\n         "
                    "To use more, please increase the number of open files allowed.\n", basemap_names[i]);
            basemap_count = i;
            break;
        }

        Basemap * new = register_basemap(basemap_names[i]);
        if (new) {
            // Attempt to obtain GeoTIFF tags
            GDALDatasetH basemap_gdal_data = GDALOpen(new->name, GA_ReadOnly);
            if (basemap_gdal_data) {
                char * projection_reference = (char *)(GDALGetProjectionRef(basemap_gdal_data));
                OGRSpatialReferenceH spatial_reference = OSRNewSpatialReference(NULL);
                OSRImportFromWkt(spatial_reference, &projection_reference);
                char * proj4_string = NULL;
                OSRExportToProj4(spatial_reference, &proj4_string);
                if (*proj4_string) {
                    new->projection_string = strdup(proj4_string);
                    new->projection = pj_init_plus(proj4_string);
                    double transformation[6];
                    if (GDALGetGeoTransform(basemap_gdal_data, transformation) == CE_None) {
                        new->extents.minx = transformation[0];
                        new->extents.maxx = transformation[0] + transformation[1] * GDALGetRasterXSize(basemap_gdal_data);
                        new->extents.maxy = transformation[3];
                        new->extents.miny = transformation[3] + transformation[5] * GDALGetRasterYSize(basemap_gdal_data);
                        new->got_gtiff_tags = true;
                    }
                }
                GDALClose(basemap_gdal_data);
            }

            // Add the currently registered basemap to the list of basemaps
            if (first_basemap == NULL) {
                first_basemap = new;
                last_basemap = new;
            }
            else {
                last_basemap->next = new;
                last_basemap = new;
            }
        }
        else {
            fprintf(stderr, "Warning: file '%s' was not found or could not be open for reading; omitting.\n", basemap_names[i]);
        }
    }

    if(first_basemap == NULL) {
        fprintf(stderr, "Warning: Nothing to do. None of the input files were found/could not be opened for reading.\n");
        project_basemaps(first_basemap, dest_file, projection, bbox, dest_width, dest_height);
        return;
    }

    Basemap *basemap_pointer = first_basemap;
    while (basemap_pointer != NULL) {
            if(0 != load_extents_and_mask(basemap_pointer, projection, bbox, dest_width, dest_height)) {
              exit(1);
            }

            if(0 != load_basemap(basemap_pointer, projection, bbox, dest_width, dest_height)) {
              exit(1);
            }

      basemap_pointer = basemap_pointer->next;
    }

    project_basemaps(first_basemap, dest_file, projection, bbox, dest_width, dest_height);

    // Clean up
    Basemap * cleaner = first_basemap;
    while (cleaner != NULL) {
        Basemap * following = cleaner->next;
        unload_basemap(cleaner);
        unregister_basemap(cleaner);
        cleaner = following;
    }
}

void usage(const char* progname) {
    printf(
            "%s: project an image from a set of basemap tiles.\n\n"
            "Options:\n"
            "  --help, -h: show help\n"
            "  --projection, -p projection: projection to map in (proj4 format)\n"
            "  --minx, -x n;\n"
            "  --maxx, -X n;\n"
            "  --miny, -y n;\n"
            "  --maxy, -Y n: define the bounding box for the output image.\n"
            "  --dest_width, -W size: width of the destination image.\n" 
            "  --dest_height, -H size: height of the destination image.\n"
            "  --output, -o filename: file to render the image into.\n"
            "  --basemap, -b name: filename of a basemap; may be specified multiple times.\n"
            "            If not specified at all, basemap filenames should be passed via stdin, one filename per line.\n", progname);
}

int main(int argc, char **argv) {
    original_tiff_warning_handler = TIFFSetWarningHandler(suppress_geotiff_complaints);
    int c, option_index;
    static struct option long_options[] =
    {
        {"help",        no_argument,       0, 'h'},
        {"projection",  required_argument, 0, 'p'},
        {"minx",        required_argument, 0, 'x'},
        {"maxx",        required_argument, 0, 'X'},
        {"miny",        required_argument, 0, 'y'},
        {"maxy",        required_argument, 0, 'Y'},
        {"dest_width",  required_argument, 0, 'W'},
        {"dest_height", required_argument, 0, 'H'},
        {"output",      required_argument, 0, 'o'},
        {"basemap",     required_argument, 0, 'b'}
    };
    unsigned int basemap_count = 0;
    char ** basemap_names = NULL;

    char *projection = 0, *dest_file = 0;
    int dest_width=0, dest_height=0;
    BBox bbox = {0,0,0,0};
    bool gotx=(false), gotX=(false), goty=(false), gotY=(false);

    while (1) { 
        c = getopt_long(argc, argv, "hp:x:X:y:Y:W:H:o:b:", long_options, &option_index);

        if (c == -1) break;

        switch(c) {
            case 'h':
                usage(argv[0]);
                exit(0);
            case 'p':
                projection = strdup(optarg);
                break;
            case 'x':
                sscanf(optarg, "%lf", &bbox.minx);
                gotx=(true);
                break;
            case 'X':
                sscanf(optarg, "%lf", &bbox.maxx);
                gotX=(true);
                break;
            case 'y':
                sscanf(optarg, "%lf", &bbox.miny);
                goty=(true);
                break;
            case 'Y':
                sscanf(optarg, "%lf", &bbox.maxy);
                gotY=(true);
                break;
            case 'W':
                dest_width = atoi(optarg);
                break;
            case 'H':
                dest_height = atoi(optarg);
                break;
            case 'o':
                dest_file = strdup(optarg);
                break;
            case 'b':
                basemap_names = realloc(basemap_names, ++basemap_count * sizeof(char *));
                basemap_names[basemap_count - 1] = strdup(optarg);
        }
    }
    bool got_required = true;
    if (projection == 0)    {fprintf(stderr, "A projection must be specified.\n");          got_required = false;}
    if (dest_file == 0)     {fprintf(stderr, "An output file is required.\n");              got_required = false;}
    if (gotx == false)      {fprintf(stderr, "BBox value minx must be specified.\n");       got_required = false;}
    if (gotX == false)      {fprintf(stderr, "BBox value maxx must be specified.\n");       got_required = false;}
    if (goty == false)      {fprintf(stderr, "BBox value miny must be specified.\n");       got_required = false;}
    if (gotY == false)      {fprintf(stderr, "BBox value maxy must be specified.\n");       got_required = false;}
    if (dest_width <= 0)    {fprintf(stderr, "The destination image width must be > 0\n");  got_required = false;}
    if (dest_height <= 0)   {fprintf(stderr, "The destination image height must be > 0\n"); got_required = false;}
    if (!got_required) {
        usage(argv[0]);
        exit(1);
    }

    load_and_project_basemaps(basemap_count, basemap_names, dest_file, projection, bbox, dest_width, dest_height);
    if (projection != 0) free(projection);
    if (dest_file != 0) free(dest_file);
    for (unsigned int i = 0; i < basemap_count; i++) {
        free(basemap_names[i]);
    }
    free(basemap_names);

    exit(0);
}
