/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_shapefile_layer.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Mapserver Shapefile layer using proj4
 *
 * 1) Any 3-dimensional object is currently treated the same as a 2-dimensional object, ignoring the Z coordinate
 * 2) Any "measure" object is currently treated the same as a 2-dimensional object, ignoring the M number
 * 3) Different "parts" of objects are rendered simply as different objects entirely; for example, two "parts" of a polygon will appear no differently than two separate polygons defined the same way
 * 4) The "Ring", "First ring", and "Outer ring" sub-classifications of "Multipatch" are all simply rendered as polygons
 * 5) The "Inner ring" is also rendered as a polygon, but with transparent fill (assumed to be a "hole" or some such business)
 *
 */

#include <getopt.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gd.h>
#include <gdal.h>
#include <ogr_srs_api.h>
#include <shapefil.h>

#include "map_shared_code.h"

typedef struct {
	// Shapefiles and associated databases
	int num_shapefiles;
	char ** shapefiles;
	char ** databases;
	char ** projections;

	// Misc
	int pixel_diameter;
	char * label_format;
	char * label_font;
	int label_font_size;
	int label_padding;

	// Stroke color
	int stroke_red;
	int stroke_green;
	int stroke_blue;
	int stroke_opacity;

	// Fill color
	int fill_red;
	int fill_green;
	int fill_blue;
	int fill_opacity;

	// Text background color
	int text_bg_red;
	int text_bg_green;
	int text_bg_blue;
	int text_bg_opacity;
} shapefile_layer_characteristics;

/* This struct keeps track of the geometry of all the labels that have been
 * created */
typedef struct {
	int count;
	int * min_x;
	int * max_x;
	int * min_y;
	int * max_y;
} label_collection;

/* Parse label format string, replacing cues with database field values, and
 * plot the resulting label at a given center */
void parse_and_draw_label(gdImagePtr text_image, int dest_width, int dest_height, shapefile_layer_characteristics traits, int stroke_color, int text_bg_color, int buffer_space, int label_x, int label_y, DBFHandle shapefile_database_file, int shape_index, label_collection * labels) {
	// Prepare the label content
	int field_count = DBFGetFieldCount(shapefile_database_file);
	char field_name[field_count][12];
	for (int field = 0; field < field_count; field++) {
		DBFGetFieldInfo(shapefile_database_file, field, field_name[field], NULL, NULL);
	}
	char * parsed_label = NULL;
	int parsed_label_length = 0;
	char * label_format_reader = traits.label_format;
	while (*label_format_reader) {
		if (*label_format_reader != '#' || *(label_format_reader + 1) != '{') {
			// This is a literal character; simply add it to the parsed label
			parsed_label = realloc(parsed_label, ++parsed_label_length);
			parsed_label[parsed_label_length - 1] = *label_format_reader;
		}
		else {
			// This is a field name; parse it out and substitute
			label_format_reader += 2;
			char * close_brace = strstr(label_format_reader, "}");
			if (close_brace == NULL) {
				fprintf(stderr, "Error: \"#{\" without a matching \"}\" in label format string!\n");
				exit(1);
			}
			// Temporarily null-terminate the field name
			*close_brace = 0;
			// Look for matching name
			int field;
			for (field = 0; field < field_count; field++) {
				if (strcmp(field_name[field], label_format_reader) == 0) {
					const char * field_value = DBFReadStringAttribute(shapefile_database_file, shape_index, field);
					parsed_label_length += strlen(field_value);
					parsed_label = realloc(parsed_label, parsed_label_length);
					memcpy(&parsed_label[parsed_label_length - strlen(field_value)], field_value, strlen(field_value));
					break;
				}
			}
			if (field >= field_count) {
				fprintf(stderr, "Error: Could not find field: \"%s\"!  Field names found:\n", label_format_reader);
				for (field = 0; field < field_count; field++) {
					fprintf(stderr, "  %s\n", field_name[field]);
				}
				exit(1);
			}
			// Restore the close-brace character and fast-forward the parsing pointer
			*close_brace = '}';
			label_format_reader = close_brace;
		}
		label_format_reader++;
	}
	// Null-terminate the parsed label
	parsed_label = realloc(parsed_label, parsed_label_length + 1);
	parsed_label[parsed_label_length] = 0;
	// Get metrics so we can center the label
	int bounding_box[8];
	if (gdImageStringFT(NULL, bounding_box, stroke_color, traits.label_font, traits.label_font_size, 0, 0, 0, parsed_label)) {
		fprintf(stderr, "Error when trying to draw labels -- did you specify a legitimate font path?\n");
		exit(1);
	}
	int text_width = bounding_box[2];
	int text_height = -bounding_box[5];
	bool draw_this_label = true;
	/* Check to see whether this label would be too close to any that we've
	 * already drawn; also make sure we don't go past the edge of the image */
	for (int label = 0; label < labels->count && draw_this_label; label++) {
		if ((label_x - text_width / 2 < labels->max_x[label] + traits.label_padding && label_x + text_width / 2 > labels->min_x[label] - traits.label_padding) && (label_y - text_height / 2 < labels->max_y[label] + traits.label_padding && label_y + text_height / 2 > labels->min_y[label] - traits.label_padding)) draw_this_label = false;
		if (label_x - text_width / 2 < 0 || label_x + text_width / 2 >= dest_width || label_y - text_height / 2 < 0 || label_y + text_height / 2 >= dest_height) draw_this_label = false;
	}
	if (draw_this_label) {
		// Draw the background and the actual label, with adjusted position
		gdImageFilledRectangle(text_image, label_x - text_width / 2 + buffer_space, label_y - text_height / 2 + buffer_space, label_x + text_width / 2 + buffer_space, label_y + text_height / 2 + buffer_space, text_bg_color);
		if (gdImageStringFT(text_image, bounding_box, stroke_color, traits.label_font, traits.label_font_size, 0, label_x - text_width / 2 + buffer_space, label_y + text_height / 2 + buffer_space, parsed_label)) {
			fprintf(stderr, "Error when trying to draw labels!\n");
			exit(1);
		}
		// Note the location and make sure we don't overlap in the future
		labels->count++;
		labels->min_x = realloc(labels->min_x, sizeof(int) * labels->count);
		labels->max_x = realloc(labels->max_x, sizeof(int) * labels->count);
		labels->min_y = realloc(labels->min_y, sizeof(int) * labels->count);
		labels->max_y = realloc(labels->max_y, sizeof(int) * labels->count);
		labels->min_x[labels->count - 1] = label_x - text_width / 2;
		labels->max_x[labels->count - 1] = label_x + text_width / 2;
		labels->min_y[labels->count - 1] = label_y - text_height / 2;
		labels->max_y[labels->count - 1] = label_y + text_height / 2;
	}
	free(parsed_label);
}

int create_shapefile_layer(const char* dest_file, const char* projection, BBox bbox, int dest_width, int dest_height, shapefile_layer_characteristics traits) {
	projPJ pj_dest;
	projPJ pj_latlon;

	pj_latlon = pj_init_plus("+proj=latlon +a=1");
	pj_dest = pj_init_plus(projection);
	if (!pj_dest) {
		fprintf(stderr, "Error: Unable to generate output projection object.\n");
		exit(1);
	}

	validate_bbox(projection, pj_dest, &bbox);

	double dest_scale_x =  (((double) (bbox.maxx-bbox.minx))/dest_width);
	double dest_scale_y =  (((double) (bbox.miny-bbox.maxy))/dest_height);
	/* Add a bit of breathing room to the edge of the image, later to be
	 * cropped, so as to keep libgd from doing something weird with polygon
	 * filling in rare edge/corner cases */
	int buffer_space = (dest_width > dest_height ? dest_width / 20 : dest_height / 20);

	/* Images, colors.  Note: libgd measures opacity from 0-127, and it wants
	 * transparency, not opacity, so we invert our opacity values as well as
	 * scaling them down by half */
	int transparent = gdTrueColorAlpha(255, 0, 0, 127);
	int stroke_color = gdTrueColorAlpha(traits.stroke_red, traits.stroke_green, traits.stroke_blue, (255 - traits.stroke_opacity) / 2);
	int fill_color = gdTrueColorAlpha(traits.fill_red, traits.fill_green, traits.fill_blue, (255 - traits.fill_opacity) / 2);
	int text_bg_color = gdTrueColorAlpha(traits.text_bg_red, traits.text_bg_green, traits.text_bg_blue, (255 - traits.text_bg_opacity) / 2);
	gdImagePtr stroke_image = gdImageCreateTrueColor(dest_width + 2 * buffer_space, dest_height + 2 * buffer_space);
	gdImageSaveAlpha(stroke_image, 1);
	gdImagePtr fill_image = gdImageCreateTrueColor(dest_width + 2 * buffer_space, dest_height + 2 * buffer_space);
	gdImageSaveAlpha(fill_image, 1);
	gdImagePtr text_image = gdImageCreateTrueColor(dest_width + 2 * buffer_space, dest_height + 2 * buffer_space);
	gdImageSaveAlpha(text_image, 1);
	// Leave alpha blending disabled for the fills, because we don't want fills to stack up and get "thick," obscuring their background
	gdImageAlphaBlending(fill_image, 0);
	gdImageFilledRectangle(fill_image, buffer_space, buffer_space, dest_width + buffer_space, dest_height + buffer_space, transparent);
	gdImageAlphaBlending(stroke_image, 0);
	gdImageFilledRectangle(stroke_image, buffer_space, buffer_space, dest_width + buffer_space, dest_height + buffer_space, transparent);
	gdImageAlphaBlending(stroke_image, 1);
	gdImageAlphaBlending(text_image, 0);
	gdImageFilledRectangle(text_image, buffer_space, buffer_space, dest_width + buffer_space, dest_height + buffer_space, transparent);
	gdImageAlphaBlending(text_image, 1);

	// Keep track of text label locations and make sure they are all distanced from each other by at least the "padding" value
	label_collection labels = {0, NULL, NULL, NULL, NULL};

	// Process the shapefiles
	for (int shapefile_index = 0; shapefile_index < traits.num_shapefiles; shapefile_index++) {
		// Open this shapefile and get info
		SHPHandle shapefile_handle = SHPOpen(traits.shapefiles[shapefile_index], "rb");
		if (shapefile_handle == NULL) {
			fprintf(stderr, "Error: Could not open \"%s\" for reading.  Aborting.\n", traits.shapefiles[shapefile_index]);
			exit(1);
		}
		// By default, assume shapefile is in lon/lat
		char * shapefile_projection_str = "+proj=longlat +a=1";
		// Check to see if the shapefile has an accompanying projection file
		FILE * shapefile_projection_file = fopen(traits.projections[shapefile_index], "r");
		if (shapefile_projection_file) {
			// How big is the projection file?
			fseek(shapefile_projection_file, 0, SEEK_END);
			int shapefile_projection_file_length = ftell(shapefile_projection_file);
			fseek(shapefile_projection_file, 0, SEEK_SET);
			if (shapefile_projection_file_length > 0) {
				// Allocate a big enough buffer for the WKT, and read it
				char * WKT = malloc(shapefile_projection_file_length + 1);
				if (WKT == NULL) {
					fprintf(stderr, "Error: Could not allocate memory for well-known text.\n");
					exit(1);
				}
				if (fgets(WKT, shapefile_projection_file_length, shapefile_projection_file) != WKT) {
					fprintf(stderr, "There was an error when attempting to read WKT from file %s\n", traits.projections[shapefile_index]);
					exit(1);
				}
				fclose(shapefile_projection_file);
				/* Special case: this French phrase is confusing to the
				 * conversion routine, so if it's present, change it to
				 * English. */
				const char *FRENCH_PHRASE = "Plate_Carree";
				const char *ENGLISH_REPLACEMENT = "Equirectangular";
				if (strstr(WKT, FRENCH_PHRASE)) {
					char *new_WKT = malloc(strlen(WKT) + 32);
					char *writer = new_WKT;
					char *reader = WKT;
					while (*reader) {
						if (strncmp(reader, FRENCH_PHRASE, strlen(FRENCH_PHRASE)) == 0) {
							strcpy(writer, ENGLISH_REPLACEMENT);
							reader += strlen(FRENCH_PHRASE);
							writer += strlen(ENGLISH_REPLACEMENT);
						}
						else {
							*(writer++) = *(reader++);
						}
					}
					*writer = '\0';
					free(WKT);
					WKT = new_WKT;
				}
				// Translate WKT to proj4
				OGRSpatialReferenceH spatial_reference = OSRNewSpatialReference(NULL);
				/* OSRImportFromWkt modifies the WKT pointer, so we make a copy of
				 * it in order to be able to free its target later... */
				char * WKT_copy = WKT;
				OSRImportFromWkt(spatial_reference, &WKT);
				free(WKT_copy);
				OSRExportToProj4(spatial_reference, &shapefile_projection_str);
			}
		}
		// Initialize shapefile projection object from projection string
		projPJ shapefile_projection = pj_init_plus(shapefile_projection_str);
		if (shapefile_projection == NULL) {
			fprintf(stderr, "Error: Unable to generate source projection object.\n");
			exit(1);
		}
		DBFHandle shapefile_database_file = NULL;
		if (traits.label_format) {
			// Check to see if the shapefile has an accompanying database file
			shapefile_database_file = DBFOpen(traits.databases[shapefile_index], "rb");
			if (!shapefile_database_file) {
				fprintf(stderr, "Error: Label option was specified, but cannot read database file %s.\n", traits.databases[shapefile_index]);
				exit(1);
			}
		}
		int num_shapes, shape_type;
		SHPGetInfo(shapefile_handle, &num_shapes, &shape_type, NULL, NULL);
		// Process all the shapes
		for (int shape_index = 0; shape_index < num_shapes; shape_index++) {
			// Read in this shape
			SHPObject * this_shape = SHPReadObject(shapefile_handle, shape_index);
			// Point
			if (this_shape -> nSHPType == SHPT_POINT ||  this_shape -> nSHPType == SHPT_MULTIPOINT || this_shape -> nSHPType == SHPT_POINTZ || this_shape -> nSHPType == SHPT_MULTIPOINTZ || this_shape -> nSHPType == SHPT_POINTM || this_shape -> nSHPType == SHPT_MULTIPOINTM) {
				for (int point = 0; point < this_shape -> nVertices; point++) {
					double proj_x = this_shape -> padfX[point];
					double proj_y = this_shape -> padfY[point];
					double longitude, latitude;
					if (!project_other_projection_to_latlon(shapefile_projection_str, &longitude, &latitude, pj_latlon, shapefile_projection, proj_x, proj_y)) continue;
					int x, y;
					if (!project_latlon_to_image(projection, &x, &y, pj_latlon, pj_dest, bbox, longitude, latitude, false, dest_scale_x, dest_scale_y)) continue;
					if (x >= (-traits.pixel_diameter / 2) && x < (dest_width + traits.pixel_diameter / 2) && y >= (-traits.pixel_diameter / 2) && y < (dest_height + traits.pixel_diameter / 2)) {
						gdImageFilledEllipse(fill_image, x + buffer_space, y + buffer_space, traits.pixel_diameter, traits.pixel_diameter, stroke_color);
						if (traits.label_format && traits.label_font) {
							parse_and_draw_label(text_image, dest_width, dest_height, traits, stroke_color, text_bg_color, buffer_space, x, y, shapefile_database_file, shape_index, &labels);
						}
					}
				}
			}
			// Polyline / polygon
			if (this_shape -> nSHPType == SHPT_ARC || this_shape -> nSHPType == SHPT_POLYGON || this_shape -> nSHPType == SHPT_ARCZ || this_shape -> nSHPType == SHPT_POLYGONZ || this_shape -> nSHPType == SHPT_ARCM || this_shape -> nSHPType == SHPT_POLYGONM) {
				double biggest_part_area = -1;
				double label_x = -1;
				double label_y = -1;
				for (int part = 0; part < this_shape -> nParts; part++) {
					// Information about this part
					int first_vertex = this_shape -> panPartStart[part];
					int last_vertex = (part == this_shape -> nParts - 1 ? this_shape -> nVertices - 1 : this_shape -> panPartStart[part + 1] - 1);
					int num_vertices = last_vertex - first_vertex + 1;
					point_info points[num_vertices];
					for (int point = first_vertex; point <= last_vertex; point++) {
						double proj_x = this_shape -> padfX[point];
						double proj_y = this_shape -> padfY[point];
						if (!project_other_projection_to_latlon(shapefile_projection_str, &points[point - first_vertex].longitude_degrees, &points[point - first_vertex].latitude_degrees, pj_latlon, shapefile_projection, proj_x, proj_y)) continue;
						points[point - first_vertex].longitude_radians = points[point - first_vertex].longitude_degrees * DEG_TO_RAD;
						points[point - first_vertex].latitude_radians = points[point - first_vertex].latitude_degrees * DEG_TO_RAD;
						// Transform the coordinates of the point and check whether it works
						points[point - first_vertex].in_projection = project_latlon_to_image(projection, &(points[point - first_vertex].x), &(points[point - first_vertex].y), pj_latlon, pj_dest, bbox, points[point - first_vertex].longitude_degrees, points[point - first_vertex].latitude_degrees, false, dest_scale_x, dest_scale_y);
						// See if it's actually within the bounds of the output image
						points[point - first_vertex].in_image = (points[point - first_vertex].in_projection && points[point - first_vertex].x >= 0 && points[point - first_vertex].x < dest_width && points[point - first_vertex].y >= 0 && points[point - first_vertex].y < dest_height);
					}
					// Check legitimacy
					bool line_legit[num_vertices - 1];
					if (polything_is_legitimate(projection, pj_latlon, pj_dest, bbox, dest_scale_x, dest_scale_y, num_vertices, points, line_legit)) {
						// Construct and draw the part
						if (this_shape -> nSHPType == SHPT_POLYGON || this_shape -> nSHPType == SHPT_POLYGONZ || this_shape -> nSHPType == SHPT_POLYGONM) {
							gdPoint image_coords[num_vertices];
							for (int point = 0; point < num_vertices; point++) {
								image_coords[point].x = points[point].x + buffer_space;
								image_coords[point].y = points[point].y + buffer_space;
							}
							gdImageFilledPolygon(fill_image, image_coords, num_vertices, fill_color);
							gdImagePolygon(stroke_image, image_coords, num_vertices, stroke_color);
						}
						else {
							for (int point = 0; point < num_vertices - 1; point++) {
								gdImageLine(stroke_image, points[point].x + buffer_space, points[point].y + buffer_space, points[point + 1].x + buffer_space, points[point + 1].y + buffer_space, stroke_color);
							}
						}
						if (traits.label_format && traits.label_font) {
							double * x = NULL;
							double * y = NULL;
							int points_in_image = 0;
							for (int point = 0; point < num_vertices; point++) {
								if (points[point].in_image) {
									points_in_image++;
									x = realloc(x, sizeof(double) * points_in_image);
									y = realloc(y, sizeof(double) * points_in_image);
									x[points_in_image - 1] = points[point].x;
									y[points_in_image - 1] = points[point].y;
								}
							}
							if (points_in_image > 2) {
								// Check candidacy for label-positioning
								mapped_polygon polygon;
								void init_mapped_polygon(mapped_polygon * polygon, int vertex_count, double * x, double * y, int resolution, bool square_bounding_box);
								init_mapped_polygon(&polygon, points_in_image, x, y, 100, false);
								if (polygon.approx_area > biggest_part_area) {
									biggest_part_area = polygon.approx_area;
									label_position(100, points_in_image, x, y, &label_x, &label_y);
								}
								free_mapped_polygon(&polygon);
							}
							free(x);
							free(y);
						}
					}
					else {
						for (int point = 0; point < num_vertices - 1; point++) {
							if (line_legit[point])
								gdImageLine(stroke_image, points[point].x + buffer_space, points[point].y + buffer_space, points[point + 1].x + buffer_space, points[point + 1].y + buffer_space, stroke_color);
						}
					}
				}
				if (biggest_part_area > 0) parse_and_draw_label(text_image, dest_width, dest_height, traits, stroke_color, text_bg_color, buffer_space, label_x, label_y, shapefile_database_file, shape_index, &labels);
			}
			// Multipatch
			if (this_shape -> nSHPType == SHPT_MULTIPATCH) {
				// Use the biggest section in the multipatch to place the label
				double biggest_section_area = -1;
				double label_x = -1;
				double label_y = -1;
				for (int part = 0; part < this_shape -> nParts; part++) {
					// Information about this part
					int first_vertex = this_shape -> panPartStart[part];
					int last_vertex = (part == this_shape -> nParts - 1 ? this_shape -> nVertices - 1 : this_shape -> panPartStart[part + 1] - 1);
					int num_vertices = last_vertex - first_vertex + 1;
					point_info points[num_vertices];
					for (int point = first_vertex; point <= last_vertex; point++) {
						double proj_x = this_shape -> padfX[point];
						double proj_y = this_shape -> padfY[point];
						if (!project_other_projection_to_latlon(shapefile_projection_str, &points[point - first_vertex].longitude_degrees, &points[point - first_vertex].latitude_degrees, pj_latlon, shapefile_projection, proj_x, proj_y)) continue;
						points[point - first_vertex].longitude_radians = points[point - first_vertex].longitude_degrees * DEG_TO_RAD;
						points[point - first_vertex].latitude_radians = points[point - first_vertex].latitude_degrees * DEG_TO_RAD;
						// Transform the coordinates of the point and check whether it works
						points[point - first_vertex].in_projection = project_latlon_to_image(projection, &(points[point - first_vertex].x), &(points[point - first_vertex].y), pj_latlon, pj_dest, bbox, points[point - first_vertex].longitude_degrees, points[point - first_vertex].latitude_degrees, false, dest_scale_x, dest_scale_y);
						// See if it's actually within the bounds of the output image
						points[point - first_vertex].in_image = (points[point - first_vertex].x >= 0 && points[point - first_vertex].x < dest_width && points[point - first_vertex].y >= 0 && points[point - first_vertex].y < dest_height);
					}
					// Tristrip
					if (this_shape -> panPartType[part] == SHPP_TRISTRIP) {
						// Consider each triangle in the strip
						for (int triangle = 0; triangle < num_vertices - 2; triangle++) {
							// Assemble the triangle as a point_info array
							point_info this_triangle[4];
							memcpy(&(this_triangle[0]), &(points[triangle]), sizeof(point_info));
							memcpy(&(this_triangle[1]), &(points[triangle + 1]), sizeof(point_info));
							memcpy(&(this_triangle[2]), &(points[triangle + 2]), sizeof(point_info));
							memcpy(&(this_triangle[3]), &(points[triangle]), sizeof(point_info));
							bool line_legit[3];
							if (polything_is_legitimate(projection, pj_latlon, pj_dest, bbox, dest_scale_x, dest_scale_y, 4, this_triangle, line_legit)) {
								// Get the triangle into image-coords-only format
								gdPoint triangle[3];
								triangle[0].x = this_triangle[0].x + buffer_space;
								triangle[0].y = this_triangle[0].y + buffer_space;
								triangle[1].x = this_triangle[1].x + buffer_space;
								triangle[1].y = this_triangle[1].y + buffer_space;
								triangle[2].x = this_triangle[2].x + buffer_space;
								triangle[2].y = this_triangle[2].y + buffer_space;
								// Draw it
								gdImageFilledPolygon(fill_image, triangle, 3, fill_color);
								gdImagePolygon(stroke_image, triangle, 3, stroke_color);
								if (traits.label_format && traits.label_font) {
									// Check candidacy of putting the shape's label here
									double this_section_area = abs((this_triangle[0].x * (this_triangle[1].y - this_triangle[2].y) + this_triangle[1].x * (this_triangle[2].y - this_triangle[0].y) + this_triangle[2].x * (this_triangle[0].y - this_triangle[1].y)) / 2);
									if (this_section_area > biggest_section_area) {
										// Use simple formula to find centroid of triangle
										double midpoint_x = ((double)(this_triangle[1].x) + this_triangle[2].x) / 2;
										double midpoint_y = ((double)(this_triangle[1].y) + this_triangle[2].y) / 2;
										double centroid_x = (double)(this_triangle[0].x) + (midpoint_x - (double)(this_triangle[0].x)) * 2 / 3;
										double centroid_y = (double)(this_triangle[0].y) + (midpoint_y - (double)(this_triangle[0].y)) * 2 / 3;
										label_x = centroid_x;
										label_y = centroid_y;
										biggest_section_area = this_section_area;
									}
								}
							}
							else {
								// Draw whatever individual lines we can
								for (int line = 0; line < 3; line++) {
									if (line_legit[line]) gdImageLine(stroke_image, this_triangle[line].x + buffer_space, this_triangle[line].y + buffer_space, this_triangle[line + 1].x + buffer_space, this_triangle[line + 1].y + buffer_space, stroke_color);
								}
							}
						}
					}
					// Trifan
					if (this_shape -> panPartType[part] == SHPP_TRIFAN) {
						// Consider each triangle in the fan
						for (int triangle = 0; triangle < num_vertices - 2; triangle++) {
							// Assemble the triangle as a point_info array
							point_info this_triangle[4];
							memcpy(&(this_triangle[0]), &(points[0]), sizeof(point_info));
							memcpy(&(this_triangle[1]), &(points[triangle + 1]), sizeof(point_info));
							memcpy(&(this_triangle[2]), &(points[triangle + 2]), sizeof(point_info));
							memcpy(&(this_triangle[3]), &(points[0]), sizeof(point_info));
							bool line_legit[3];
							if (polything_is_legitimate(projection, pj_latlon, pj_dest, bbox, dest_scale_x, dest_scale_y, 4, this_triangle, line_legit)) {
								// Get the triangle into image-coords-only format
								gdPoint triangle[3];
								triangle[0].x = this_triangle[0].x + buffer_space;
								triangle[0].y = this_triangle[0].y + buffer_space;
								triangle[1].x = this_triangle[1].x + buffer_space;
								triangle[1].y = this_triangle[1].y + buffer_space;
								triangle[2].x = this_triangle[2].x + buffer_space;
								triangle[2].y = this_triangle[2].y + buffer_space;
								// Draw it
								gdImageFilledPolygon(fill_image, triangle, 3, fill_color);
								gdImagePolygon(stroke_image, triangle, 3, stroke_color);
								if (traits.label_format && traits.label_font) {
									// Check candidacy of putting the shape's label here
									double this_section_area = abs((this_triangle[0].x * (this_triangle[1].y - this_triangle[2].y) + this_triangle[1].x * (this_triangle[2].y - this_triangle[0].y) + this_triangle[2].x * (this_triangle[0].y - this_triangle[1].y)) / 2);
									if (this_section_area > biggest_section_area) {
										// Use simple formula to find centroid of triangle
										double midpoint_x = ((double)(this_triangle[1].x) + this_triangle[2].x) / 2;
										double midpoint_y = ((double)(this_triangle[1].y) + this_triangle[2].y) / 2;
										double centroid_x = (double)(this_triangle[0].x) + (midpoint_x - (double)(this_triangle[0].x)) * 2 / 3;
										double centroid_y = (double)(this_triangle[0].y) + (midpoint_y - (double)(this_triangle[0].y)) * 2 / 3;
										label_x = centroid_x;
										label_y = centroid_y;
										biggest_section_area = this_section_area;
									}
								}
							}
							else {
								// Draw whatever individual lines we can
								for (int line = 0; line < 3; line++) {
									if (line_legit[line]) gdImageLine(stroke_image, this_triangle[line].x + buffer_space, this_triangle[line].y + buffer_space, this_triangle[line + 1].x + buffer_space, this_triangle[line + 1].y + buffer_space, stroke_color);
								}
							}
						}
					}
					// Rings (polygons)
					if (this_shape -> panPartType[part] == SHPP_RING || this_shape -> panPartType[part] == SHPP_OUTERRING || this_shape -> panPartType[part] == SHPP_INNERRING || this_shape -> panPartType[part] == SHPP_FIRSTRING) {
						bool line_legit[num_vertices - 1];
						if (polything_is_legitimate(projection, pj_latlon, pj_dest, bbox, dest_scale_x, dest_scale_y, num_vertices, points, line_legit)) {
							// Treat it the same as a polygon
							gdPoint image_coords[num_vertices];
							for (int point = 0; point < num_vertices; point++) {
								image_coords[point].x = points[point].x + buffer_space;
								image_coords[point].y = points[point].y + buffer_space;
							}
							gdImageFilledPolygon(fill_image, image_coords, num_vertices, fill_color);
							gdImagePolygon(stroke_image, image_coords, num_vertices, stroke_color);
							/*if (this_shape -> panPartType[part] == SHPP_INNERRING) {
								// TODO: Do something special to "inner rings"!
							}*/
							if (traits.label_format && traits.label_font) {
								double * x = NULL;
								double * y = NULL;
								int points_in_image = 0;
								for (int point = 0; point < num_vertices; point++) {
									if (points[point].in_image) {
										points_in_image++;
										x = realloc(x, sizeof(double) * points_in_image);
										y = realloc(y, sizeof(double) * points_in_image);
										x[points_in_image - 1] = points[point].x;
										y[points_in_image - 1] = points[point].y;
									}
								}
								if (points_in_image > 2) {
									if (traits.label_format && traits.label_font) {
										// Check candidacy of putting the shape's label here
										mapped_polygon polygon;
										void init_mapped_polygon(mapped_polygon * polygon, int vertex_count, double * x, double * y, int resolution, bool square_bounding_box);
										init_mapped_polygon(&polygon, points_in_image, x, y, 100, false);
										if (polygon.approx_area > biggest_section_area) {
											biggest_section_area = polygon.approx_area;
											label_position(100, points_in_image, x, y, &label_x, &label_y);
										}
										free_mapped_polygon(&polygon);
									}
								}
								free(x);
								free(y);
							}
						}
						else {
							// As usual, just draw what lines we can
							for (int line = 0; line < num_vertices - 1; line++) {
								if (line_legit[line]) gdImageLine(stroke_image, points[line].x + buffer_space, points[line].y + buffer_space, points[line + 1].x + buffer_space, points[line + 1].y + buffer_space, stroke_color);
							}
						}
					}
				}
				if (traits.label_format && traits.label_font) {
					// Plot the label in the biggest section
					parse_and_draw_label(text_image, dest_width, dest_height, traits, stroke_color, text_bg_color, buffer_space, label_x, label_y, shapefile_database_file, shape_index, &labels);
				}
			}
			SHPDestroyObject(this_shape);
		}
		free(labels.min_x);
		free(labels.max_x);
		free(labels.min_y);
		free(labels.max_y);
		SHPClose(shapefile_handle);
		if (shapefile_database_file) DBFClose(shapefile_database_file);
	}
	// Crop the image down to what we want
	gdImagePtr cropped_image = gdImageCreateTrueColor(dest_width, dest_height);
	gdImageSaveAlpha(cropped_image, 1);
	gdImageAlphaBlending(cropped_image, 0);
	gdImageFilledRectangle(cropped_image, 0, 0, dest_width, dest_height, transparent);
	gdImageAlphaBlending(cropped_image, 1);
	// Drop the fills onto the output first, then the strokes, and finally the text
	gdImageCopy(cropped_image, fill_image, 0, 0, buffer_space, buffer_space, dest_width, dest_height);
	gdImageCopy(cropped_image, stroke_image, 0, 0, buffer_space, buffer_space, dest_width, dest_height);
	gdImageCopy(cropped_image, text_image, 0, 0, buffer_space, buffer_space, dest_width, dest_height);
	FILE * output_file = fopen(dest_file, "wb");
	gdImagePng(cropped_image, output_file);
	fclose(output_file);
	gdImageDestroy(stroke_image);
	gdImageDestroy(fill_image);
	gdImageDestroy(cropped_image);
	pj_free(pj_dest);
	pj_free(pj_latlon);
	return 0;
}

void usage(const char* progname) {
	printf(
			"%s: render a shapefile layer to be superimposed onto a map projection.\n\n"
			"Options:\n"
			"  --help, -h: show help\n"
			"  --projection, -p projection: projection to map in (proj4 format)\n"
			"  --minx, -x n;\n"
			"  --maxx, -X n;\n"
			"  --miny, -y n;\n"
			"  --maxy, -Y n: define the bounding box for the output image.\n"
			"  --dest_width, -W size: width of the destination image.\n" 
			"  --dest_height, -H size: height of the destination image.\n"
			"  --output, -o filename: file to render the image into.\n"
			"  --shapefile, -s filename: path to input shapefile (may be specified multiple times)\n"
			"  --stroke-color, -S s: 8-digit hexadecimal string specifying RGBA values for stroke color of shapefile layer\n"
			"  --fill-color, -F s: 8-digit hexadecimal string specifying RGBA values for fill color of shapefile layer (default: no fill / transparent)\n"
			"  --bg-color, -B s: 8-digit hexadecimal string specifying RGBA values for text background color of shapefile layer (default: no background / transparent)\n"
			"  --diameter, -d n: diameter of plotted single points in pixels\n"
			"  --label, -L s: Formatted string for labeling shapes, with DBF field names inserted as follows:\n"
			"                 \"Example text, including #{DB_FIELD_1} and also #{DB_FIELD_2}.\"\n"
			"  --true-type, -T s: Path to a truetype font; required for label printing\n"
			"                     NOTE: if supplying a path relative to the working directory, it must begin with './'; e.g. 'font.ttf' will not work but './font.ttf' will.\n"
			"  --font-size, -f n: font size; if not specified, defaults to a size derived from the output image size.\n"
			"  --padding, -P n: Minimum padding, in pixels, between text labels.  Default is 10.\n", progname);
}

int main(int argc, char **argv) {
	int c, option_index;
	static struct option long_options[] =
	{
		{"help",                 no_argument,         0, 'h'},
		{"projection",           required_argument,   0, 'p'},
		{"minx",                 required_argument,   0, 'x'},
		{"maxx",                 required_argument,   0, 'X'},
		{"miny",                 required_argument,   0, 'y'},
		{"maxy",                 required_argument,   0, 'Y'},
		{"dest_width",           required_argument,   0, 'W'},
		{"dest_height",          required_argument,   0, 'H'},
		{"output",               required_argument,   0, 'o'},
		{"shapefile",            required_argument,   0, 's'},
		{"stroke-color",         required_argument,   0, 'S'},
		{"fill-color",           required_argument,   0, 'F'},
		{"bg-color",             required_argument,   0, 'B'},
		{"diameter",             required_argument,   0, 'd'},
		{"label",                required_argument,   0, 'L'},
		{"true-type",            required_argument,   0, 'T'},
		{"font-size",            required_argument,   0, 'f'},
		{"padding",              required_argument,   0, 'P'},
	};

	char *projection = 0, *dest_file = 0;
	int dest_width=0, dest_height=0;
	BBox bbox = {0,0,0,0};
	bool gotx=(false), gotX=(false), goty=(false), gotY=(false);
	bool got_stroke = false;
	bool font_size_override = false;

	shapefile_layer_characteristics traits = {0, NULL, NULL, NULL, 8, NULL, NULL, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

	while (1) { 
		c = getopt_long(argc, argv, "hp:x:X:y:Y:W:H:o:s:S:F:B:d:L:T:f:P:", long_options, &option_index);

		if (c == -1) break;

		switch(c) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'p':
				projection = strdup(optarg);
				break;
			case 'x':
				sscanf(optarg, "%lf", &bbox.minx);
				gotx=(true);
				break;
			case 'X':
				sscanf(optarg, "%lf", &bbox.maxx);
				gotX=(true);
				break;
			case 'y':
				sscanf(optarg, "%lf", &bbox.miny);
				goty=(true);
				break;
			case 'Y':
				sscanf(optarg, "%lf", &bbox.maxy);
				gotY=(true);
				break;
			case 'W':
				dest_width = atoi(optarg);
				break;
			case 'H':
				dest_height = atoi(optarg);
				break;
			case 'o':
				dest_file = strdup(optarg);
				break;
			case 's': {
						  traits.shapefiles = realloc(traits.shapefiles, sizeof(char *) * ++traits.num_shapefiles);
						  traits.shapefiles[traits.num_shapefiles - 1] = strdup(optarg);
						  break;
					  }
			case 'S': {
						  // Parse hex
						  if (strlen(optarg) != 8) {fprintf(stderr, "Argument to -S must be an eight-digit hexidecimal number.\n"); exit(1);}
						  int digits[8] = {0,0,0,0,0,0,0,0};
						  for (int i = 0; i < 8; i++) {
							  unsigned char this_char = optarg[i];
							  if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
							  if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
							  else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
							  else {fprintf(stderr, "Argument to -S must be an eight-digit hexidecimal number.\n"); exit(1);}
						  }
						  traits.stroke_red =      16 * digits[0] + digits[1];
						  traits.stroke_green =    16 * digits[2] + digits[3];
						  traits.stroke_blue =     16 * digits[4] + digits[5];
						  traits.stroke_opacity =  16 * digits[6] + digits[7];
						  got_stroke = true;
						  break;
					  }
			case 'F': {
						  // Parse hex
						  if (strlen(optarg) != 8) {fprintf(stderr, "Argument to -F must be an eight-digit hexidecimal number.\n"); exit(1);}
						  int digits[8] = {0,0,0,0,0,0,0,0};
						  for (int i = 0; i < 8; i++) {
							  unsigned char this_char = optarg[i];
							  if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
							  if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
							  else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
							  else {fprintf(stderr, "Argument to -F must be an eight-digit hexidecimal number.\n"); exit(1);}
						  }
						  traits.fill_red =      16 * digits[0] + digits[1];
						  traits.fill_green =    16 * digits[2] + digits[3];
						  traits.fill_blue =     16 * digits[4] + digits[5];
						  traits.fill_opacity =  16 * digits[6] + digits[7];
						  break;
					  }
			case 'B': {
						  // Parse hex
						  if (strlen(optarg) != 8) {fprintf(stderr, "Argument to -B must be an eight-digit hexidecimal number.\n"); exit(1);}
						  int digits[8] = {0,0,0,0,0,0,0,0};
						  for (int i = 0; i < 8; i++) {
							  unsigned char this_char = optarg[i];
							  if (this_char >= 'A' && this_char <= 'F') this_char -= ('A' - 'a'); // Convert to lowercase
							  if (this_char >= '0' && this_char <= '9') digits[i] = this_char - '0';
							  else if (this_char >= 'a' && this_char <= 'f') digits[i] = 10 + this_char - 'a';
							  else {fprintf(stderr, "Argument to -B must be an eight-digit hexidecimal number.\n"); exit(1);}
						  }
						  traits.text_bg_red =      16 * digits[0] + digits[1];
						  traits.text_bg_green =    16 * digits[2] + digits[3];
						  traits.text_bg_blue =     16 * digits[4] + digits[5];
						  traits.text_bg_opacity =  16 * digits[6] + digits[7];
						  break;
					  }
			case 'd':
					  traits.pixel_diameter = atoi(optarg);
					  break;
			case 'L':
					  traits.label_format = strdup(optarg);
					  break;
			case 'T':
					  traits.label_font = strdup(optarg);
					  break;
			case 'f':
					  font_size_override = true;
					  sscanf(optarg, "%d", &traits.label_font_size);
					  break;
			case 'P':
					  sscanf(optarg, "%d", &traits.label_padding);
					  break;
		}
	}

	// Default font size
	if (!font_size_override) traits.label_font_size = (dest_width > dest_height ? dest_width / 50 : dest_height / 50);

	bool got_required = true;
	if (projection == 0)                {fprintf(stderr, "A projection must be specified.\n");                  got_required = false;}
	if (dest_file == 0)                 {fprintf(stderr, "An output file is required.\n");                      got_required = false;}
	if (gotx == false)                  {fprintf(stderr, "BBox value minx must be specified.\n");               got_required = false;}
	if (gotX == false)                  {fprintf(stderr, "BBox value maxx must be specified.\n");               got_required = false;}
	if (goty == false)                  {fprintf(stderr, "BBox value miny must be specified.\n");               got_required = false;}
	if (gotY == false)                  {fprintf(stderr, "BBox value maxy must be specified.\n");               got_required = false;}
	if (dest_width <= 0)                {fprintf(stderr, "The destination image width must be > 0\n");          got_required = false;}
	if (dest_height <= 0)               {fprintf(stderr, "The destination image height must be > 0\n");         got_required = false;}
	if (!got_stroke)                    {fprintf(stderr, "A stroke color is required.\n");                      got_required = false;}
	if (traits.num_shapefiles == 0)     {fprintf(stderr, "At least one shapefile must be supplied.\n");         got_required = false;}
	if (traits.pixel_diameter < 1)      {fprintf(stderr, "The pixel diameter must be a positive integer\n");    got_required = false;}
	if (traits.label_font_size < 1)     {fprintf(stderr, "The font size must be a positive integer\n");         got_required = false;}
	if (traits.label_padding < 1)       {fprintf(stderr, "Padding must be a positive integer\n");               got_required = false;}

	if (!got_required) {
		usage(argv[0]);
		exit(1);
	}

	/* Derive paths to possible DB files and projection files associated with
	 * the input shapefiles */
	traits.databases   = malloc(sizeof(char *) * traits.num_shapefiles);
	traits.projections = malloc(sizeof(char *) * traits.num_shapefiles);
	for (int shapefile = 0; shapefile < traits.num_shapefiles; shapefile++) {
		char * peek_shapefile_name = traits.shapefiles[shapefile];
		while (*peek_shapefile_name) peek_shapefile_name++;
		peek_shapefile_name -= 3;
		if (strcmp(peek_shapefile_name, "shp") == 0) {
			traits.databases[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.databases[shapefile][strlen(traits.databases[shapefile]) - 3]), "dbf");
			traits.projections[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.projections[shapefile][strlen(traits.projections[shapefile]) - 3]), "prj");
		}
		else if (strcmp(peek_shapefile_name, "SHP") == 0) {
			traits.databases[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.databases[shapefile][strlen(traits.databases[shapefile]) - 3]), "DBF");
			traits.projections[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.projections[shapefile][strlen(traits.projections[shapefile]) - 3]), "PRJ");
		}
		else if (strcmp(peek_shapefile_name, "Shp") == 0) {
			traits.databases[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.databases[shapefile][strlen(traits.databases[shapefile]) - 3]), "Dbf");
			traits.projections[shapefile] = strdup(traits.shapefiles[shapefile]);
			strcpy(&(traits.projections[shapefile][strlen(traits.projections[shapefile]) - 3]), "Prj");
		}
		else {
			traits.databases[shapefile] = NULL;
			traits.projections[shapefile] = NULL;
		}
	}

	create_shapefile_layer(dest_file, projection, bbox, dest_width, dest_height, traits);

	if (projection != 0) free(projection);
	if (dest_file != 0) free(dest_file);

	exit(0);
}
