/*
 * $URL: https://repos.ser.asu.edu/svn/lroc/map_utils/tags/release_20160315/src/map_shared_code.c $
 * 
 * Copyright (C) 2012 by Arizona State University and Mark Robinson.
 * All rights reserved.
 *
 * This file has been released under the modified BSD license.
 * See COPYING in the distribution package for details.
 *
 * Author: Caleb Hanger <changer@ser.asu.edu>
 *
 * $Author: nme $
 *   $Date: 2015-10-29 08:18:29 -0700 (Thu, 29 Oct 2015) $
 *    $Rev: 16916 $
 *
 * Some procedures that are shared between various renderers
 */

#include <ctype.h>
#include <inttypes.h>
#include <limits.h>
#include <string.h>

#include "map_shared_code.h"

// Tolerances for legitimacy checking in each direction
#define LON_TOLERANCE 90
#define LAT_TOLERANCE 45

// Iniitialize basemap tile cache
cached_tile * first_cached_tile = NULL;
cached_tile * final_cached_tile = NULL;
unsigned int cached_tile_count = 0;

// Find a destination given a starting point and an angular distance
void find_destination(double start_longitude, double start_latitude, double bearing, double distance, double * end_longitude, double * end_latitude) {
	start_longitude *= DEG_TO_RAD;
	start_latitude  *= DEG_TO_RAD;
	bearing         *= DEG_TO_RAD;
	distance        *= DEG_TO_RAD;

	*end_latitude  = asin(sin(start_latitude) * cos(distance) + cos(start_latitude) * sin(distance) * cos(bearing));
	*end_longitude = start_longitude + atan2(sin(bearing) * sin(distance) * cos(start_latitude), cos(distance) - sin(start_latitude) * sin(*end_latitude));

	*end_latitude  *= RAD_TO_DEG;
	*end_longitude *= RAD_TO_DEG;
}

// Find the midpoint of two locations along the appropriate "great circle"
void find_midpoint(double * midpoint, double lon1, double lat1, double lon2, double lat2) {
	lon1 *= DEG_TO_RAD;
	lon2 *= DEG_TO_RAD;
	lat1 *= DEG_TO_RAD;
	lat2 *= DEG_TO_RAD;
	double b_x = cos(lat2) * cos(lon2 - lon1);
	double b_y = cos(lat2) * sin(lon2 - lon1);
	midpoint[0] = (lon1 + atan2(b_y, cos(lat1) + b_x)) * RAD_TO_DEG;
	midpoint[1] = (atan2(sin(lat1) + sin(lat2), sqrt(pow(cos(lat1) + b_x, 2) + pow(b_y, 2)))) * RAD_TO_DEG;
}

// Shortest-distance arc length between two lat/lon points (Haversine formula)
double short_arc_length(double lon1, double lat1, double lon2, double lat2) {
	lon1 *= DEG_TO_RAD;
	lon2 *= DEG_TO_RAD;
	lat1 *= DEG_TO_RAD;
	lat2 *= DEG_TO_RAD;
	double a = pow(sin(((lat2 - lat1) / 2)), 2) + cos(lat1) * cos(lat2) * pow(sin((lon2 - lon1) / 2), 2);
	return (2 * atan2(sqrt(a), sqrt(1 - a)) * RAD_TO_DEG);
}

// Check whether a point lies in a polygon in Cartesian space
bool point_in_poly(double x, double y, int num_vertices, double * poly_x, double * poly_y) {
	int lines_to_the_left = 0;
	for (int line = 0; line < num_vertices; line++) {
		// Get the endpoints of the line in polygon land
		double polygon_space_x1 = poly_x[line];
		double polygon_space_y1 = poly_y[line];
		double polygon_space_x2 = poly_x[line == num_vertices - 1 ? 0 : line + 1];
		double polygon_space_y2 = poly_y[line == num_vertices - 1 ? 0 : line + 1];
		// Figure out which is the upper endpoint and which is the lower
		double upper_endpoint_x = (polygon_space_y1 > polygon_space_y2 ? polygon_space_x1 : polygon_space_x2);
		double upper_endpoint_y = (polygon_space_y1 > polygon_space_y2 ? polygon_space_y1 : polygon_space_y2);
		double lower_endpoint_x = (polygon_space_y1 > polygon_space_y2 ? polygon_space_x2 : polygon_space_x1);
		double lower_endpoint_y = (polygon_space_y1 > polygon_space_y2 ? polygon_space_y2 : polygon_space_y1);
		if (y > lower_endpoint_y && y <= upper_endpoint_y) {
			double line_projection_x = lower_endpoint_x + (upper_endpoint_x - lower_endpoint_x) * ((y - lower_endpoint_y) / (upper_endpoint_y - lower_endpoint_y));
			if (x > line_projection_x) lines_to_the_left++;
		}
	}
	return (lines_to_the_left % 2 == 1);
}

// Initialize a mapped_polygon structure
void init_mapped_polygon(mapped_polygon * polygon, int vertex_count, double * x, double * y, int resolution, bool square_bounding_box) {
	// Find bounding box
	polygon->min_x = x[0];
	polygon->max_x = x[0];
	polygon->min_y = y[0];
	polygon->max_y = y[0];
	for (int i = 1; i < vertex_count; i++) {
		if (x[i] < polygon->min_x) polygon->min_x = x[i];
		if (x[i] > polygon->max_x) polygon->max_x = x[i];
		if (y[i] < polygon->min_y) polygon->min_y = y[i];
		if (y[i] > polygon->max_y) polygon->max_y = y[i];
	}
	if (square_bounding_box) {
		// Stretch the bounding box to make it square, if necessary
		if (polygon->max_x - polygon->min_x > polygon->max_y - polygon->min_y) {
			polygon->max_y = polygon->min_y + polygon->max_x - polygon->min_x;
		}
		if (polygon->max_y - polygon->min_y > polygon->max_x - polygon->min_x) {
			polygon->max_x = polygon->min_x + polygon->max_y - polygon->min_y;
		}
	}
	polygon->resolution = resolution;
	// Build map of polygon area
	polygon->map = calloc(1, resolution * resolution);
	polygon->x = malloc(sizeof(double) * resolution);
	polygon->y = malloc(sizeof(double) * resolution);
	double unit_area = ((polygon->max_x - polygon->min_x) / resolution) * ((polygon->max_y - polygon->min_y) / resolution);
	polygon->approx_area = 0;
	// Store map-space to polygon-space conversion results
	for (int map_x = 0; map_x < resolution; map_x++) {
		polygon->x[map_x] = polygon->min_x + ((double)(map_x) / (resolution - 1)) * (polygon->max_x - polygon->min_x);
	}
	for (int map_y = 0; map_y < resolution; map_y++) {
		polygon->y[map_y] = polygon->max_y + ((double)(map_y) / (resolution - 1)) * (polygon->min_y - polygon->max_y);
	}
	for (int line = 0; line < vertex_count; line++) {
		// Get the endpoints of the line in polygon land
		double polygon_space_x1 = x[line];
		double polygon_space_y1 = y[line];
		double polygon_space_x2 = x[line == vertex_count - 1 ? 0 : line + 1];
		double polygon_space_y2 = y[line == vertex_count - 1 ? 0 : line + 1];
		// Figure out which is the upper endpoint and which is the lower
		double upper_endpoint_x = (polygon_space_y1 > polygon_space_y2 ? polygon_space_x1 : polygon_space_x2);
		double upper_endpoint_y = (polygon_space_y1 > polygon_space_y2 ? polygon_space_y1 : polygon_space_y2);
		double lower_endpoint_x = (polygon_space_y1 > polygon_space_y2 ? polygon_space_x2 : polygon_space_x1);
		double lower_endpoint_y = (polygon_space_y1 > polygon_space_y2 ? polygon_space_y2 : polygon_space_y1);
		// Scan every row of the map, looking for rows affected by this line, and keeping track of the polygon area
		for (int map_y = 0; map_y < resolution; map_y++) {
			if (polygon->y[map_y] > lower_endpoint_y && polygon->y[map_y] <= upper_endpoint_y) {
				// This line is affected; find the individual points to the right of the line
				double line_polygon_space_x = lower_endpoint_x + ((polygon->y[map_y] - lower_endpoint_y) / (upper_endpoint_y - lower_endpoint_y)) * (upper_endpoint_x - lower_endpoint_x);
				for (int map_x = 0; map_x < resolution; map_x++) {
					if (polygon->x[map_x] > line_polygon_space_x) {
						// This point is affected; toggle this byte in the map
						polygon->map[map_y * resolution + map_x] ^= 1;
						// Update the polygon's total area appropriately
						polygon->approx_area += (polygon->map[map_y * resolution + map_x] ? unit_area : -unit_area);
					}
				}
			}
		}
	}
}

// Clean up a mapped_polygon
void free_mapped_polygon(mapped_polygon * polygon) {
	free(polygon->map);
	free(polygon->x);
	free(polygon->y);
}

// Find a good place in a polygon to center a horizontal text label
void label_position(unsigned int map_resolution, int vertex_count, double * x, double * y, double * x_center, double * y_center) {
	// Figure out center of non-square bounding box
	double min_x = x[0];
	double max_x = x[0];
	double min_y = y[0];
	double max_y = y[0];
	for (int vertex = 0; vertex < vertex_count; vertex++) {
		if (x[vertex] < min_x) min_x = x[vertex];
		if (x[vertex] > max_x) max_x = x[vertex];
		if (y[vertex] < min_y) min_y = y[vertex];
		if (y[vertex] > max_y) max_y = y[vertex];
	}
	double bbox_x_center = (min_x + max_x) / 2;
	double bbox_y_center = (min_y + max_y) / 2;
	// Map the polygon in a square bounding box
	mapped_polygon polygon;
	init_mapped_polygon(&polygon, vertex_count, x, y, map_resolution, true);
	// Choose a label location by shrinking the polygon.
	uint8_t * shrunken_map = malloc(map_resolution * map_resolution);
	memcpy(shrunken_map, polygon.map, map_resolution * map_resolution);
	while (true) {
		uint8_t * new_shrunken_map = malloc(map_resolution * map_resolution);
		memcpy(new_shrunken_map, shrunken_map, map_resolution * map_resolution);
		for (int map_y = 0; map_y < map_resolution; map_y++) {
			for (int map_x = 0; map_x < map_resolution; map_x++) {
				/* On each iteration, wear the polygon down from all sides, and
				 * corners, with more weight on the left/right since the label
				 * will be horizontal */
				if (map_x == 0 ||
					map_x == 1 ||
					map_x == map_resolution - 1 ||
					map_x == map_resolution - 2 ||
					map_y == 0 ||
					map_y == map_resolution - 1 ||
					!shrunken_map[map_y * map_resolution + map_x - 1] ||
					!shrunken_map[map_y * map_resolution + map_x - 2] ||
					!shrunken_map[map_y * map_resolution + map_x + 1] ||
					!shrunken_map[map_y * map_resolution + map_x + 2] ||
					!shrunken_map[map_y * map_resolution + map_x - map_resolution] ||
					!shrunken_map[map_y * map_resolution + map_x + map_resolution] ||
					!shrunken_map[map_y * map_resolution + map_x - 1 - map_resolution] ||
					!shrunken_map[map_y * map_resolution + map_x + 1 - map_resolution] ||
					!shrunken_map[map_y * map_resolution + map_x - 1 + map_resolution] ||
					!shrunken_map[map_y * map_resolution + map_x + 1 + map_resolution])
					new_shrunken_map[map_y * map_resolution + map_x] = 0;
			}
		}
		/* If this iteration would destroy all area, then we are done shrinking
		 * the polygon */
		bool done_shrinking = true;
		for (int map_y = 0; map_y < map_resolution; map_y++) {
			for (int map_x = 0; map_x < map_resolution; map_x++) {
				if (new_shrunken_map[map_y * map_resolution + map_x]) {
					done_shrinking = false;
					break;
				}
			}
		}
		if (done_shrinking) {
			free(new_shrunken_map);
			break;
		}
		free(shrunken_map);
		shrunken_map = new_shrunken_map;
	}
	/* Use the shrunken map to select a label center -- Out of all the points
	 * still remaining in the last iteration, we choose the one closest to the
	 * center of the polygon's original bounding box */
	int label_x = -1, label_y = -1;
	double map_center_x = (bbox_x_center - polygon.min_x) * (map_resolution - 1) / (polygon.max_x - polygon.min_x);
	double map_center_y = (bbox_y_center - polygon.max_y) * (map_resolution - 1) / (polygon.min_y - polygon.max_y);
	int distance_from_center = map_resolution * map_resolution;
	for (int map_y = 0; map_y < map_resolution; map_y++) {
		for (int map_x = 0; map_x < map_resolution; map_x++) {
			if (shrunken_map[map_y * map_resolution + map_x]) {
				if (map_x == -1 || (abs(map_x - map_center_x) * abs(map_x - map_center_x) + abs(map_y - map_center_y) * abs(map_y - map_center_y)) < distance_from_center) {
					label_x = map_x;
					label_y = map_y;
					distance_from_center = abs(map_x - map_center_x) * abs(map_x - map_center_x) + abs(map_y - map_center_y) * abs(map_y - map_center_y);
				}
			}
		}
	}
	free(shrunken_map);
	free_mapped_polygon(&polygon);
	// Convert the label location back to polygon-space coordinates
	*x_center = polygon.min_x + ((double)(label_x) / (map_resolution - 1)) * (polygon.max_x - polygon.min_x);
	*y_center = polygon.max_y + ((double)(label_y) / (map_resolution - 1)) * (polygon.min_y - polygon.max_y);
}

/*
 * Check to see whether a polyline/polygon is legitimate.  This check also
 * populates an array of bools with indicators as to whether the individual
 * lines comprising the polything are legitimate.  The return value can never
 * be true unless all the lines were deemed legitimate, and the return value
 * can still be false in that case.  If the return value is true, the polything
 * as a whole looks good, so you can fill it etc. if you want.  If not, you are
 * still okay to draw the individual lines, but only the ones whose slots were
 * given "true" values.  This function expects that closed polygons are
 * explicity described as such, by way of the last point being equal to the
 * first point.  Thus, the bool array for the inidividual lines should be one
 * element smaller than the array of points, in every case.
 */

bool polything_is_legitimate(const char * projection, projPJ pj_latlon, projPJ pj_dest, BBox bbox, double dest_scale_x, double dest_scale_y, int num_points, point_info * points, bool * line_legit) {

	/*
     * This line is black magic.  If the caller sent us a shape that, in its
	 * entirety, only has two points, then we will not recurse in the following
	 * block.  Therefore, the caller may get a "false" return value but if the
	 * bool *line_legit has junk memory making it look "true," the caller will
	 * think that the line is okay to draw.  Thus we initialize that bool to
	 * false.
     */

	if (num_points == 2 && line_legit != NULL) *line_legit = false;

	// Check the individual lines first, by recursively investigating them as 2-point polythings
	if (num_points > 2) {
		bool has_invalid_lines = false;
		for (int i = 0; i < num_points - 1; i++) {
			line_legit[i] = polything_is_legitimate(projection, pj_latlon, pj_dest, bbox, dest_scale_x, dest_scale_y, 2, &(points[i]), NULL);
			if (!line_legit[i]) has_invalid_lines = true;
		}
		if (has_invalid_lines) return false;
	}

	// If this is just one line, and either point is not in the projection, we know immediately that it isn't a good point
	if (num_points == 2) {
		if (!points[0].in_projection || !points[1].in_projection) return false;
	}

	/*
     * If we're looking at an orthographic projection and every vertex is in the projection, the entire polygon is definitely legitimate
     * since there are no wrapping edges in an orthographic projection -- you are looking at only one side of the planetary body).
     */

	if (strstr(projection, "ortho")) return true;

	int ret = 0;
	char ortho_string[200];
	// Use an ortho centered around the polything to compare real vs. perceived
	sprintf(ortho_string, "+proj=ortho +lat_0=%lf +lon_0=%lf +ellps=WGS84", points[0].latitude_degrees, points[0].longitude_degrees);
	projPJ pj_ortho = pj_init_plus(ortho_string);
	if (!pj_ortho) {
		fprintf(stderr, "Error: Unable to generate orthographic projection object.\n");
		exit(1);
	}
	// Find the midpoints, perceived and real
	double real_min_x = 0, real_min_y = 0, real_mid_x = 0, real_mid_y = 0, real_max_x = 0, real_max_y = 0;
	double perceived_min_x = 0, perceived_min_y = 0, perceived_mid_x = 0, perceived_mid_y = 0, perceived_max_x = 0, perceived_max_y = 0;
	for (int i = 0; i < num_points; i++) {
		double this_point_x = points[i].longitude_radians;
		double this_point_y = points[i].latitude_radians;
		ret = pj_transform(pj_latlon, pj_ortho, 1, 1, &this_point_x, &this_point_y, NULL);
		if (ret != 0) return false;
		if (i == 0) {
			real_min_x = this_point_x;
			real_max_x = this_point_x;
			real_min_y = this_point_y;
			real_max_y = this_point_y;
		}
		else {
			if (this_point_x < real_min_x) real_min_x = this_point_x;
			if (this_point_x > real_max_x) real_max_x = this_point_x;
			if (this_point_y < real_min_y) real_min_y = this_point_y;
			if (this_point_y > real_max_y) real_max_y = this_point_y;
		}
	}
	real_mid_x = (real_min_x + real_max_x) / 2;
	real_mid_y = (real_min_y + real_max_y) / 2;
	// Find the needed statistics of the perceived points in the orthographic projection
	for (int i = 0; i < num_points; i++) {
		double this_point_x = points[i].x * dest_scale_x + bbox.minx;
		double this_point_y = points[i].y * dest_scale_y + bbox.maxy;
		if (i == 0) {
			perceived_min_x = this_point_x;
			perceived_max_x = this_point_x;
			perceived_min_y = this_point_y;
			perceived_max_y = this_point_y;
		}
		else {
			if (this_point_x < perceived_min_x) perceived_min_x = this_point_x;
			if (this_point_x > perceived_max_x) perceived_max_x = this_point_x;
			if (this_point_y < perceived_min_y) perceived_min_y = this_point_y;
			if (this_point_y > perceived_max_y) perceived_max_y = this_point_y;
		}
	}
	perceived_mid_x = (perceived_min_x + perceived_max_x) / 2;
	perceived_mid_y = (perceived_min_y + perceived_max_y) / 2;
	// Need to convert to radians here if we're projecting from latlon
	if (pj_is_latlong(pj_dest)) {
		real_mid_x *= DEG_TO_RAD;
		real_mid_y *= DEG_TO_RAD;
		perceived_mid_x *= DEG_TO_RAD;
		perceived_mid_y *= DEG_TO_RAD;
	}
	ret = pj_transform(pj_dest, pj_ortho, 1, 1, &(perceived_mid_x), &(perceived_mid_y), NULL);
	if (ret != 0) return false;
	// Now, project back into latlon space....unless this is an oblique transformation, in which case we want to project onto the destination lat/lon space
	if (strstr(projection, "ob_tran")) {
		ret = pj_transform(pj_ortho, pj_dest, 1, 1, &(perceived_mid_x), &(perceived_mid_y), NULL);
		if (ret != 0) return false;
		ret = pj_transform(pj_ortho, pj_dest, 1, 1, &(real_mid_x), &(real_mid_y), NULL);
		if (ret != 0) return false;
	}
	else {
		ret = pj_transform(pj_ortho, pj_latlon, 1, 1, &(perceived_mid_x), &(perceived_mid_y), NULL);
		if (ret != 0) return false;
		ret = pj_transform(pj_ortho, pj_latlon, 1, 1, &(real_mid_x), &(real_mid_y), NULL);
		if (ret != 0) return false;
	}
	// Proj4 gave us back radians
	real_mid_x *= RAD_TO_DEG;
	real_mid_y *= RAD_TO_DEG;
	perceived_mid_x *= RAD_TO_DEG;
	perceived_mid_y *= RAD_TO_DEG;
	// Need to fix some nonsense in the case of an oblique transformation with a radius that isn't 1
	if (strstr(projection, "ob_tran")) {
		double radius;
		// Get the bogus radius from the projection string so we can fix this nonsense
		char * position = strstr(projection, "+r=");
		if (position != 0) sscanf(&(position[3]), "%lf", &radius);
		else {
			position = strstr(projection, "+R=");
			if (position != 0) sscanf(&(position[3]), "%lf", &radius);
			else radius = 1;
		}
		real_mid_x /= radius;
		real_mid_y /= radius;
		perceived_mid_x /= radius;
		perceived_mid_y /= radius;
	}

    pj_free(pj_ortho);

	if (short_arc_length(real_mid_x, real_mid_y, perceived_mid_x, perceived_mid_y) < 0.25) return true;
	if (fabs(real_mid_x - perceived_mid_x) > LON_TOLERANCE) return false;
	if (fabs(real_mid_y - perceived_mid_y) > LAT_TOLERANCE) return false;
	return true;
}

// Translate a point from lon/lat space into another projection
bool project_latlon_to_other_projection (const char * projection, double * x, double * y, projPJ pj_latlon, projPJ pj_dest, BBox bbox, double longitude, double latitude, bool radians, double dest_scale_x, double dest_scale_y) {
	// Proj4 wants degrees if the projection is cylindrical...
	if (radians && pj_is_latlong(pj_dest)) {
		latitude *= RAD_TO_DEG;
		longitude *= RAD_TO_DEG;
	}
	// ...but it wants radians if the projection is something else
	if (!radians && !pj_is_latlong(pj_dest)) {
		latitude *= DEG_TO_RAD;
		longitude *= DEG_TO_RAD;
	}
	// Transform the coordinates of the point
	int ret = pj_transform(pj_latlon, pj_dest, 1, 1, &longitude, &latitude, NULL);
	if (strstr(projection, "+proj=eqc")) {
		// EQC can be a little ornery...look for a latitude of true scale and figure out the appropriate "radius"
		double true_scale_latitude = 0;
		char * lat_ts_scan = strstr(projection, "lat_ts");
		if (lat_ts_scan) sscanf(lat_ts_scan, "lat_ts=%lf", &true_scale_latitude);
		true_scale_latitude *= DEG_TO_RAD;
		// Find the radius at the equator
		double radius = 1;
		char * radius_scan = strstr(projection, "+a=");
		if (radius_scan) sscanf(radius_scan, "+a=%lf", &radius);
		radius_scan = strstr(projection, "+A=");
		if (radius_scan) sscanf(radius_scan, "+A=%lf", &radius);
		radius_scan = strstr(projection, "+b=");
		if (radius_scan) sscanf(radius_scan, "+b=%lf", &radius);
		radius_scan = strstr(projection, "+B=");
		if (radius_scan) sscanf(radius_scan, "+B=%lf", &radius);
		// Find the radius+circumference along the latitude line of true scale
		double radius_at_true_scale = radius * cos(true_scale_latitude);
		double circumference_at_true_scale = radius_at_true_scale * 2 * PI;
		// In case we got a mis-aligned ("wrapped") EQC X-coordinate, shift it up and down by an interval of this pseudo-circumference to check whether it lands within the extents
		while (longitude < bbox.minx) longitude += circumference_at_true_scale;
		while (longitude > bbox.maxx) longitude -= circumference_at_true_scale;
	}
	// If the transformation didn't go okay, skip this point
	if (ret != 0) return false;
	// If the destination is a projection measured in latitudes / longitudes, we need to get the longitude as close as possible to the bbox to avoid edge problems
	if (pj_is_latlong(pj_dest) || strstr(projection, "ob_tran")) {
		// Decide which longitude domain to go with based on how far the bbox is from 360 / -180
		bool longitude_type_is_360 = (360 - bbox.maxx < bbox.minx - (-180));
		// Get the longitude into 180-type
		while (longitude < -180) longitude += 360;
		while (longitude >  180) longitude -= 360;
		// If the longitude type is actually 360, switch over if needed
		if (longitude_type_is_360 && longitude < 0) longitude += 360;
	}
	*x = longitude;
	*y = latitude;
	return true;
}

// Translate a point from a non-latlon projection space into image space
void project_to_image(projPJ pj_dest, int * x, int * y, BBox bbox, double proj_x, double proj_y, double dest_scale_x, double dest_scale_y) {
	if (pj_is_latlong(pj_dest)) {
		double image_middle = (bbox.minx + bbox.maxx) / 2;
		while (proj_x > image_middle + 180) {
			proj_x -= 360;
		}
		while (proj_x < image_middle - 180) {
			proj_x += 360;
		}
	}
	*x = floor((proj_x - bbox.minx) / dest_scale_x);
	*y = floor((proj_y - bbox.maxy) / dest_scale_y);
}

// Project a point from lon/lat space into image space
bool project_latlon_to_image(const char * projection, int * x, int * y, projPJ pj_latlon, projPJ pj_dest, BBox bbox, double longitude, double latitude, bool radians, double dest_scale_x, double dest_scale_y) {
	double proj_x, proj_y;
	if (!project_latlon_to_other_projection(projection, &proj_x, &proj_y, pj_latlon, pj_dest, bbox, longitude, latitude, radians, dest_scale_x, dest_scale_y)) return false;
	project_to_image(pj_dest, x, y, bbox, proj_x, proj_y, dest_scale_x, dest_scale_y);
	return true;
}

// Project a point from image space to arbitrary projection space
void project_from_image(double * proj_x, double * proj_y, BBox bbox, int x, int y, double dest_scale_x, double dest_scale_y) {
	*proj_x = dest_scale_x * ((double)x + 0.5) + bbox.minx;
	*proj_y = dest_scale_y * ((double)y + 0.5) + bbox.maxy;
}

// Project a point from a non-latlon projection space into latlon space
bool project_other_projection_to_latlon(const char * projection, double * longitude, double * latitude, projPJ pj_latlon, projPJ pj_dest, double proj_x, double proj_y) {
	bool point_is_good = true;
	// Prevent against unwanted repetition for latlon projections
	if (pj_is_latlong(pj_dest) || strstr(projection, "ob_tran")) {
		// Decide whether or not the pixel falls within the appropriate boundaries
		point_is_good = point_is_good && proj_y <= 90 && proj_y >= -90;
	}
	int ret = pj_transform(pj_dest, pj_latlon, 1, 1, &proj_x, &proj_y, NULL); // After this line we have a good lat/lon
    if(proj_x == INFINITY) {
        printf("%d\n", __LINE__);
        exit(1);
    }

	if (ret != 0) {
        // This point did not project, return false to inform the caller that this data is no good since the point doesn't even belong on the Moon
        return false;
    }

	*longitude = proj_x;
	*latitude = proj_y;
	// Proj4 gave us radians iff pj_dest is latlon.  So convert to degrees if that's the case
	if(!pj_is_latlong(pj_dest)){
		*longitude *= RAD_TO_DEG;
		*latitude *= RAD_TO_DEG;
	}
	// Tell caller whether or not the point is good
	return point_is_good;
}

// Project a point from image space to lon/lat space
bool project_image_to_latlon(const char * projection, double * longitude, double * latitude, projPJ pj_latlon, projPJ pj_dest, BBox bbox, int x, int y, double dest_scale_x, double dest_scale_y) {
	bool point_is_good = true;
	double proj_x, proj_y;
	project_from_image(&proj_x, &proj_y, bbox, x, y, dest_scale_x, dest_scale_y);
	point_is_good = point_is_good && project_other_projection_to_latlon(projection, longitude, latitude, pj_latlon, pj_dest, proj_x, proj_y);
	// Prevent against unwanted repetition for sinusoidal projections, as well
	if (strstr(projection, "sinu") || strstr(projection, "moll")) {
		// Reverse-project the basemap location back into the output image space and check whether it's the original image or a repetition of it
		double lon_rad = *longitude * DEG_TO_RAD;
		double lat_rad = *latitude * DEG_TO_RAD;
		point_is_good = point_is_good && (pj_transform(pj_latlon, pj_dest, 1, 1, &lon_rad, &lat_rad, NULL) == 0);
		int test_x = floor((lon_rad - bbox.minx) / dest_scale_x);
		int test_y = floor((lat_rad - bbox.maxy) / dest_scale_y);
		// If we didn't land in the same spot, this looks like a repetition
		if (test_x != x || test_y != y)
				point_is_good = false;
	}
	return point_is_good;
}

/* If output measurement is lat/lon, conform the bbox to [0, 360] to avoid
 * imprecision issues between 180-type and 360-type */
void validate_bbox(const char * projection, projPJ pj_dest, BBox * bbox) {
	if (pj_is_latlong(pj_dest) || strstr(projection, "ob_tran")) {
		while (bbox->minx < 0) {
			bbox->minx += 360;
			bbox->maxx += 360;
		}
		while (bbox->maxx > 360) {
			bbox->minx -= 360;
			bbox->maxx -= 360;
		}
	}
}

// These functions load tiles
void read_single_bit_tile(Basemap * current_tile, tdata_t basebuf, int base_x, int base_y) {
	TIFFReadTile(current_tile->ptif, basebuf, base_x, base_y, 0, 0);
}

void read_mono_tile(Basemap * current_tile, tdata_t basebuf, int base_x, int base_y) {
	TIFFReadTile(current_tile->ptif, basebuf, base_x, base_y, 0, 0);
}
void read_color_tile(Basemap * current_tile, tdata_t basebuf, int base_x, int base_y) {
	/* Unlike TIFFReadTile, TIFFReadRGBATile expects upper left pixel
	 * coordinates to be passed in, instead of calculating them itself */
	int tile_UL_X = (base_x / current_tile->tile_width) * current_tile->tile_width;
	int tile_UL_Y = (base_y / current_tile->tile_height) * current_tile->tile_height;
	TIFFReadRGBATile(current_tile->ptif, tile_UL_X, tile_UL_Y, basebuf);
}

// These functions pull individual values from loaded tiles
DATA32 compute_single_bit_value(tdata_t * basebuf, Basemap * current_tile, int base_x, int base_y) {
	uint64_t pixel_offset = (base_y % current_tile -> tile_height * current_tile -> tile_width) + base_x % current_tile -> tile_width;
	unsigned char containing_byte = ((uint8_t *)basebuf)[pixel_offset / 8];
	/* The bits within a byte are arranged last-to-first; that is, the
	 * rightmost pixel is the most significant bit, and the leftmost pixel is
	 * the least significant bit.  So for a bitmask we can't just use
	 * pixel_offset % 8, it has to be mirrored. */
	unsigned char bitmask = (1 << 7) >> (pixel_offset % 8);
	if (containing_byte & bitmask) return 1;
	else return 0;
}

DATA32 compute_mono_value(tdata_t *basebuf, Basemap *current_tile, int base_x, int base_y){
	DATA32 value;
	value = ((uint8_t*)basebuf)[((base_y % current_tile->tile_height * current_tile->tile_width) + base_x % current_tile->tile_width)];
	value = value | (value << 8) | (value << 16) | (0xff000000);
	return value;
}

DATA32 compute_color_value(tdata_t *basebuf, Basemap *current_tile, int base_x, int base_y){
	DATA32 value;
	value = ((DATA32*)basebuf)[(((current_tile->tile_height - (base_y % current_tile->tile_height)-1) * current_tile->tile_width) + base_x % current_tile->tile_width)];
	// Libtiff and Imlib2 like to have the R and B channels in flip-flopped positions from each other, so we swap them
	char * color = (char *)(&value);
	char swap = color[0];
	color[0] = color[2];
	color[2] = swap;
	return value;
}

// Wrapper for above functions; Get the value at (x, y) in (basemap)
DATA32 get_value(Basemap * basemap, int x, int y) {
	value_func compute_value;
	read_func read_tile;
	// Set our function pointers based off of whether we're dealing with 1-bit, grayscale, or RGBA
	if (basemap->BPS == 1) {
		compute_value = &compute_single_bit_value;
		read_tile = &read_single_bit_tile;
	}
	else if (basemap->SPP == 1) {
		compute_value = &compute_mono_value;
		read_tile = &read_mono_tile;
	}
	else {
		compute_value = &compute_color_value;
		read_tile = &read_color_tile;
	}
	// Check which tiff_tile of the basemap we should be in
	int tile_num = y / basemap->tile_height * basemap->tile_height + x / basemap->tile_width;
	// Look for the tile in memory
	cached_tile * look_cache = first_cached_tile;
	while (look_cache != NULL) {
		if (look_cache->containing_basemap == basemap && look_cache->tile_number == tile_num) {
			/* We've already got the tile loaded.  Move it to the head if it
			 * isn't already there (so that we'll find it more quickly next
			 * time)... */
			if (look_cache->previous) {
				look_cache->previous->next = look_cache->next;
				if (look_cache->next) {
					look_cache->next->previous = look_cache->previous;
				}
				else {
					/* This is the final node that we're moving, so we'd better
					 * remember to update the final_cached_tile pointer */
					final_cached_tile = look_cache->previous;
				}
				look_cache->previous = NULL;
				first_cached_tile->previous = look_cache;
				look_cache->next = first_cached_tile;
				first_cached_tile = look_cache;
			}
			// ...and return the desired pixel value.
			return compute_value(look_cache->tile_data, basemap, x, y);
		}
		look_cache = look_cache->next;
	}
	// We didn't find the tile we need already-loaded; load it and cache it at the head
	cached_tile * this_tile = malloc(sizeof(cached_tile));
	this_tile->containing_basemap = basemap;
	this_tile->tile_number = tile_num;
	this_tile->tile_data = _TIFFmalloc(TIFFTileSize(basemap->ptif) * (basemap->SPP == 3 ? 4 : 1));
	read_tile(basemap, this_tile->tile_data, x, y);
	if (cached_tile_count == 0) {
		// This is the first tile we've loaded
		first_cached_tile = this_tile;
		final_cached_tile = this_tile;
		this_tile->next = NULL;
		this_tile->previous = NULL;
	}
	else {
		// We've got at least one tile we're attaching onto
		this_tile->next = first_cached_tile;
		this_tile->previous = NULL;
		first_cached_tile->previous = this_tile;
		first_cached_tile = this_tile;
	}
	cached_tile_count++;
	if (cached_tile_count > CACHE_SIZE) {
		// We've exceeded the maximum number of cached tiles, so free one up
		cached_tile * new_final_tile = final_cached_tile->previous;
		final_cached_tile->previous->next = NULL;
		_TIFFfree(final_cached_tile->tile_data);
		free(final_cached_tile);
		final_cached_tile = new_final_tile;
		cached_tile_count--;
	}
	// Now fetch and return the pixel value
	return compute_value(this_tile->tile_data, basemap, x, y);
}

// Register a basemap
Basemap * register_basemap(const char* name) {
	// Make sure the basemap file exists!
	FILE * test_access = fopen(name, "r");
	if (!test_access) {
		return NULL;
	}
	fclose(test_access);
	Basemap* new = malloc(sizeof(Basemap));
	new->name = strdup(name);
	new->projection = 0;
	new->projection_string = 0;
	new->ptif = 0;
	new->next = 0;
	new->mask = 0;
	return new;
}

// This guy should not be called until extents and projection type are loaded (caller is responsible for that, since extents filename schema is dependent on what's being done)
int load_basemap(Basemap* basemap, const char* projection, BBox bbox, int dest_width, int dest_height) {
	// Load the TIFF and choose a resolution page
	basemap->ptif = pick_a_page(basemap -> name, basemap, projection, bbox, dest_width, dest_height);

        if(basemap->ptif == NULL) {
          fprintf(stderr, "No PTIF loaded.\n");
          return 1;
        }

	// Read the TIFF header tags that we care about
	TIFFGetField(basemap->ptif, TIFFTAG_IMAGEWIDTH, &basemap->width);
	TIFFGetField(basemap->ptif, TIFFTAG_IMAGELENGTH, &basemap->height);
	TIFFGetField(basemap->ptif, TIFFTAG_TILEWIDTH, &basemap->tile_width);
	TIFFGetField(basemap->ptif, TIFFTAG_TILELENGTH, &basemap->tile_height);
	TIFFGetField(basemap->ptif, TIFFTAG_SAMPLESPERPIXEL, &basemap->SPP);
	TIFFGetField(basemap->ptif, TIFFTAG_BITSPERSAMPLE, &basemap->BPS);

	// Calculate scales
	basemap->scalex =  (((double) (basemap->extents.maxx-basemap->extents.minx))/basemap->width); 
	basemap->scaley =  (((double) (basemap->extents.miny-basemap->extents.maxy))/basemap->height);

    // Success
    return 0;
}

// De-allocate TIFF
void unload_basemap(Basemap* basemap) {
	if (basemap->ptif != 0) TIFFClose(basemap->ptif);
	if (basemap->mask != 0 && basemap->mask->ptif != 0) TIFFClose(basemap->mask->ptif);
}

// Unregister a basemap
void unregister_basemap(Basemap * basemap) {
	if (basemap->name != 0) free(basemap->name);
	if (basemap->projection_string != 0) free(basemap->projection_string);
	if (basemap->projection != 0) pj_free(basemap->projection);
	if (basemap->mask != 0) {
		unregister_basemap(basemap->mask);
	}
	free(basemap);
}

// Returns an integer giving a pixel-hit-count in the source resolution
int resolution_quality(projPJ pj_latlon, const char * output_projection_str, projPJ output_projection, const char * source_projection_str, projPJ source_projection, BBox output_bbox, BBox source_bbox, int output_width, int output_height, int source_width, int source_height) {
	// Compute scales
	double source_scale_x = (source_bbox.maxx - source_bbox.minx) / source_width;
	double source_scale_y = (source_bbox.miny - source_bbox.maxy) / source_height;
	double output_scale_x = (output_bbox.maxx - output_bbox.minx) / output_width;
	double output_scale_y = (output_bbox.miny - output_bbox.maxy) / output_height;
	// We will project a 20x20 lattice of 5x5 "blobs" into the basemap, and see how many pixels we hit on the page
	const int lattice_size = 20;
	const int blob_radius  = 2;
	const int first_y_center = blob_radius;
	const int first_x_center = blob_radius;
	const int last_x_center = output_width  - 1 - blob_radius;
	const int last_y_center = output_height - 1 - blob_radius;
	uint64_t * source_x_hit = NULL;
	uint64_t * source_y_hit = NULL;
	int source_pixel_hit_count = 0;
	/* If this is a latlon or ob_tran projection, don't worry about the
	 * resolution at the 1/3 of the image with highest absolute latitude
	 * values, where we expect to look really stretched out */
	double minimal_absolute_latitude = 0;
	if (pj_is_latlong(output_projection) || strstr(output_projection_str, "ob_tran")) {
		if (output_bbox.miny <= 0 && output_bbox.maxy >= 0) {
			minimal_absolute_latitude = 0;
		}
		else {
			minimal_absolute_latitude = (fabs(output_bbox.miny) < fabs(output_bbox.maxy) ? fabs(output_bbox.miny) : fabs(output_bbox.maxy));
		}
	}
	for (int output_y_index = 0; output_y_index < lattice_size; output_y_index++) {
		for (int output_x_index = 0; output_x_index < lattice_size; output_x_index++) {
			int output_y_center = first_y_center + (last_y_center - first_y_center) * output_y_index / (lattice_size - 1);
			int output_x_center = first_x_center + (last_x_center - first_x_center) * output_x_index / (lattice_size - 1);
			if (pj_is_latlong(output_projection) || strstr(output_projection_str, "ob_tran")) {
				double projection_y = output_bbox.maxy + output_scale_y * output_y_center;
				if (fabs(projection_y) > minimal_absolute_latitude + (90 - minimal_absolute_latitude) * 2 / 3) continue;
			}
			for (int output_y = output_y_center - blob_radius; output_y <= output_y_center + blob_radius; output_y++) {
				for (int output_x = output_x_center - blob_radius; output_x <= output_x_center + blob_radius; output_x++) {
					// First, project to latlon
					double longitude, latitude;
					if (project_image_to_latlon(output_projection_str, &longitude, &latitude, pj_latlon, output_projection, output_bbox, output_x, output_y, output_scale_x, output_scale_y)) {
						// If that worked, continue into input space
						int source_x, source_y;
						if (project_latlon_to_image(source_projection_str, &source_x, &source_y, pj_latlon, source_projection, source_bbox, longitude, latitude, false, source_scale_x, source_scale_y)) {
							if (source_x >= 0 && source_x < source_width && source_y >= 0 && source_y < source_height) {
								// Have we hit this pixel before?
								bool already_hit_this_pixel = false;
								for (int check_source_pixel = 0; check_source_pixel < source_pixel_hit_count; check_source_pixel++) {
									if (source_x_hit[check_source_pixel] == source_x && source_y_hit[check_source_pixel] == source_y) {
										already_hit_this_pixel = true;
										break;
									}
								}
								if (!already_hit_this_pixel) {
									// We hit a pixel!
									source_pixel_hit_count++;
									source_x_hit = realloc(source_x_hit, source_pixel_hit_count * sizeof(uint64_t));
									source_y_hit = realloc(source_y_hit, source_pixel_hit_count * sizeof(uint64_t));
									source_x_hit[source_pixel_hit_count - 1] = source_x;
									source_y_hit[source_pixel_hit_count - 1] = source_y;
								}
							}
						}
					}
				}
			}
		}
	}
	// Clean up
	free(source_x_hit);
	free(source_y_hit);
	return source_pixel_hit_count;
}

// Cube-picking wrapper for resolution_quality()
char * pick_a_cube(int num_cubes, char ** cubes, const char* projection, BBox bbox, int dest_width, int dest_height) {
	if (num_cubes == 1) {
		// There's not much "picking" to do!
		return cubes[0];
	}

	projPJ pj_dest, pj_latlon;
    // We need a latlon projection middle-man, here
	pj_latlon = pj_init_plus("+proj=latlong +a=1");
	if (pj_latlon == NULL){
		fprintf(stderr, "Fatal error: Failed to init latlon projection.\n");
		exit(1);
	}
	// We also need the output projection object
	pj_dest = pj_init_plus(projection);
	if(pj_dest == NULL) {
		fprintf(stderr, "Fatal error: Failed to init dest projection.\n");
		exit(1);
	}
	int cube_quality[num_cubes];
	int cube_size[num_cubes];
	int maximum_cube_quality = 0;
	for (int try_cube = 0; try_cube < num_cubes; try_cube++) {
		// Temporary cube object for this one -- radius and cache size don't matter as we're only looking at the header
		cube_properties cube_info;
		init_cube(&cube_info, cubes[try_cube], 0, 0, false, false);
		BBox cube_bbox;
		cube_bbox.minx = cube_info.longitude_minimum;
		cube_bbox.maxx = cube_info.longitude_maximum;
		cube_bbox.miny = cube_info.latitude_minimum;
		cube_bbox.maxy = cube_info.latitude_maximum;
		cube_quality[try_cube] = resolution_quality(pj_latlon, projection, pj_dest, "+proj=latlong", pj_latlon, bbox, cube_bbox, dest_width, dest_height, cube_info.width, cube_info.height);
		// Keep track of the maximum cube quality
		if (cube_quality[try_cube] > maximum_cube_quality) maximum_cube_quality = cube_quality[try_cube];
		// Keep track of the total sizes of the cubes as well; we'll need them in a bit
		cube_size[try_cube] = cube_info.width * cube_info.height;
		// Don't need this cube anymore
		free_cube(&cube_info);
	}
	pj_free(pj_dest);
	pj_free(pj_latlon);
	// Take the lowest-res cube which still gives us 95% of the pixels we could possibly be hitting.
	// If we get *too* demanding, we can run into big slowness problems.
	uint64_t desired_cube_quality = maximum_cube_quality * 95 / 100;
	int chosen_cube = -1;
	int chosen_cube_size = INT_MAX;
	for (int i = 0; i < num_cubes; i++) {
		if (cube_quality[i] >= desired_cube_quality) {
			if (chosen_cube == -1 || cube_size[i] < chosen_cube_size) {
				chosen_cube = i;
				chosen_cube_size = cube_size[i];
			}
		}
	}
	return cubes[chosen_cube];
}

// Page-picking wrapper for resolution_quality()
TIFF* pick_a_page(const char * ptif_filename, Basemap * tile, const char* projection, BBox bbox, int dest_width, int dest_height) {
	// We need a latlon projection middle-man, here
	projPJ pj_latlon = pj_init_plus("+proj=latlong +a=1");
	if (pj_latlon == NULL){
	  fprintf(stderr, "Fatal error: Failed to init latlon projection.\n");
		return NULL;
	}
	// We also need the output projection object
	projPJ pj_dest = pj_init_plus(projection);
	if(pj_dest == NULL) {
		fprintf(stderr,"Fatal error: Failed to init dest projection.\n");
		return NULL;
	}
	// Open the PTIF and count the number of pages
	TIFF * ptif = TIFFOpen(ptif_filename, "r");
	if (ptif == NULL) {
        fprintf(stderr, "Error: Unable to open %s for reading.\n", ptif_filename);
		return NULL;
	}
	int page_count = 1;
	while (TIFFReadDirectory(ptif) == 1) page_count++;
	// Investigate each page
	int base_pixel_hit_count[page_count];
	for (int i = 0; i < page_count; i++) {
		base_pixel_hit_count[i] = 0;
	}
	int try_page;
	uint64_t maximum_pixel_hit_count = 0;
	bool basemap_is_in_output = true;
	for (try_page = page_count - 1; try_page >= 0; try_page--) {
		TIFFSetDirectory(ptif, try_page);
		// See how big this page is
		uint32 page_width, page_height;
		TIFFGetField(ptif, TIFFTAG_IMAGEWIDTH, &page_width);
		TIFFGetField(ptif, TIFFTAG_IMAGELENGTH, &page_height);
		// Get the quality for this page
		base_pixel_hit_count[try_page] = resolution_quality(pj_latlon, projection, pj_dest, tile->projection_string, tile->projection, bbox, tile->extents, dest_width, dest_height, page_width, page_height);
		// Keep track of the highest quantity of pixels hit on a page
		if (base_pixel_hit_count[try_page] > maximum_pixel_hit_count) {
			maximum_pixel_hit_count = base_pixel_hit_count[try_page];
		}
		if (base_pixel_hit_count[try_page] == 0) {
			basemap_is_in_output = false;
			break;
		}
	}

	// Clean up
	pj_free(pj_dest);
	pj_free(pj_latlon);
	/* If the basemap is not in the output at all, use the highest resolution
	 * available.  We don't have to worry about it slowing us down since this
	 * basemap doesn't show up in the output viewport, and we want to make the
	 * pixels as "small" as possible in order to avoid a weird problem where
	 * "big pixels" "bleed" into adjacent basemaps.  --changer@ser.asu.edu */
	if (!basemap_is_in_output) {
		TIFFSetDirectory(ptif, 0);
		return ptif;
	}
	// Take the lowest-res page which still gives us 95% of the pixels we could possibly be hitting.
	// If we get too demanding, we run into big slowness problems, especially with the JMARS projection.
	uint64_t desired_pixel_hit_count = maximum_pixel_hit_count * 95 / 100;
	for (int i = page_count - 1; i < page_count; i--) {
		if (base_pixel_hit_count[i] >= desired_pixel_hit_count) {
			TIFFSetDirectory(ptif, i);
			return ptif;
		}
	}
	/* This should never be reached, because a page *should* be selected above.
	 * But in case it happens somehow, we will choose the lowest resolution
	 * page, for maximum speed. */
	TIFFSetDirectory(ptif, page_count - 1);
	return ptif;
}

// Standardize a longitude value to the input domain
void validate_longitude(double * longitude, cube_properties * cube_info) {
	while (*longitude > cube_info->longitude_maximum) *longitude -= 360;
	while (*longitude < cube_info->longitude_minimum) *longitude += 360;
}

// This returns the pixel value at a specific row/column in a cube, with no respect to the meaning of the data
float get_value_at_image_point(cube_properties * cube_info, uint64_t row, uint64_t column) {
	float pixel_value = 0;
	// Calculate tile index
	uint64_t tile_number = (row / cube_info->tile_height) * (cube_info->width / cube_info->tile_width + (cube_info->width % cube_info->tile_width == 0 ? 0 : 1)) + (column / cube_info->tile_width);
	// These are the coordinates *within* a cube tile
	uint64_t pixel_row = row % cube_info->tile_height;
	uint64_t pixel_column = column % cube_info->tile_width;
	if (cube_info->complete_data) {
		// We have got the entire cube preloaded, so fetch the value from memory rather than from disk
		if (cube_info->bytes_per_pixel == 1) {
			uint8_t stored_value;
			memcpy(&stored_value, &(cube_info->complete_data[tile_number][cube_info->tile_width * pixel_row + pixel_column]), 1);
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 2) {
			int16_t stored_value;
			memcpy(&stored_value, &(cube_info->complete_data[tile_number][2 * cube_info->tile_width * pixel_row + 2 * pixel_column]), 2);
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 4) {
			float stored_value;
			memcpy(&stored_value, &(cube_info->complete_data[tile_number][4 * cube_info->tile_width * pixel_row + 4 * pixel_column]), 4);
			pixel_value = stored_value;
		}
		return pixel_value;
	}
	if (cube_info->tile_cache) {
		struct cached_cube_tile *this_tile = NULL;
		// Check to see if we've already loaded this tile
		uint32_t cached_tile_index;
		for (cached_tile_index = 0; cached_tile_index < cube_info->tile_cache_current_size; cached_tile_index++) {
			if (cube_info->tile_cache[cached_tile_index].tile_number == tile_number) {
				this_tile = &cube_info->tile_cache[cached_tile_index];
				break;
			}
		}
		if (this_tile) {
			// Yep -- send it to the head of the list, if it isn't already there
			if (cached_tile_index > 0) {
				struct cached_cube_tile transfer = *this_tile;
				memmove(&cube_info->tile_cache[1], &cube_info->tile_cache[0], cached_tile_index * sizeof(struct cached_cube_tile));
				cube_info->tile_cache[0] = transfer;
				this_tile = &cube_info->tile_cache[0];
			}
		}
		else {
			// Nope -- load it
			// If we've filled the cache to capacity, remove the least recently used tile
			if (cube_info->tile_cache_current_size == cube_info->tile_cache_capacity) {
				free(cube_info->tile_cache[cube_info->tile_cache_capacity - 1].data);
				memmove(&cube_info->tile_cache[1], &cube_info->tile_cache[0], (cube_info->tile_cache_capacity - 1) * sizeof(struct cached_cube_tile));
			}
			// Otherwise, just make room at the beginning
			else {
				memmove(&cube_info->tile_cache[1], &cube_info->tile_cache[0], cube_info->tile_cache_current_size * sizeof(struct cached_cube_tile));
				cube_info->tile_cache_current_size++;
			}
			// Make a node for the tile, load the tile, and add it at the head of the cache
			this_tile = &cube_info->tile_cache[0];
			this_tile->tile_number = tile_number;
			this_tile->data = malloc(cube_info->bytes_per_pixel * cube_info->tile_width * cube_info->tile_height);
			if (this_tile->data == NULL) {
				fprintf(stderr, "Failed to allocate memory to cache an ISIS cube tile.  Aborting.\n");
				exit(1);
			}
			unsigned char * byte_loader = this_tile->data;
			fseek(cube_info->file, cube_info->data_start_byte - 1 + tile_number * cube_info->tile_width * cube_info->tile_height * cube_info->bytes_per_pixel, SEEK_SET);
			for (uint64_t byte = 0; byte < cube_info->tile_width * cube_info->tile_height * cube_info->bytes_per_pixel; byte++) {
				*byte_loader = fgetc(cube_info->file);
				byte_loader++;
			}
		}
		// Whether the tile was already there or we've just loaded it, it's now loaded.  Get the pixel!
		if (cube_info->bytes_per_pixel == 1) {
			uint8_t stored_value;
			memcpy(&stored_value, &(this_tile->data[cube_info->tile_width * pixel_row + pixel_column]), 1);
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 2) {
			int16_t stored_value;
			memcpy(&stored_value, &(this_tile->data[2 * cube_info->tile_width * pixel_row + 2 * pixel_column]), 2);
			if (!cube_info->byte_order_matches) {
				// Endianness doesn't match; swap the bytes
				unsigned char * swapper = (unsigned char *)(&stored_value);
				swapper[0] ^= swapper[1];
				swapper[1] ^= swapper[0];
				swapper[0] ^= swapper[1];
			}
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 4) {
			float stored_value;
			memcpy(&stored_value, &(this_tile->data[4 * cube_info->tile_width * pixel_row + 4 * pixel_column]), 4);
			if (!cube_info->byte_order_matches) {
				// Endianness doesn't match; swap the bytes
				unsigned char * swapper = (unsigned char *)(&stored_value);
				swapper[0] ^= swapper[3];
				swapper[3] ^= swapper[0];
				swapper[0] ^= swapper[3];
				swapper[1] ^= swapper[2];
				swapper[2] ^= swapper[1];
				swapper[1] ^= swapper[2];
			}
			pixel_value = stored_value;
		}
	}
	else {
		// We're not using a tile cache for this cube, so just pull from the file
		if (cube_info->bytes_per_pixel == 1) {
			uint8_t stored_value;
			fseek(cube_info->file, cube_info->data_start_byte - 1 + tile_number * cube_info->tile_width * cube_info->tile_height + cube_info->tile_width * pixel_row + pixel_column, SEEK_SET);
			stored_value = fgetc(cube_info->file);
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 2) {
			int16_t stored_value;
			unsigned char * reader = (unsigned char *)(&stored_value);
			fseek(cube_info->file, cube_info->data_start_byte - 1 + 2 * tile_number * cube_info->tile_width * cube_info->tile_height + 2 * cube_info->tile_width * pixel_row + 2 * pixel_column, SEEK_SET);
			reader[0] = fgetc(cube_info->file);
			reader[1] = fgetc(cube_info->file);
			if (!cube_info->byte_order_matches) {
				// Endianness doesn't match; swap the bytes
				reader[0] ^= reader[1];
				reader[1] ^= reader[0];
				reader[0] ^= reader[1];
			}
			pixel_value = stored_value;
		}
		else if (cube_info->bytes_per_pixel == 4) {
			float stored_value;
			unsigned char * reader = (unsigned char *)(&stored_value);
			fseek(cube_info->file, cube_info->data_start_byte - 1 + 4 * tile_number * cube_info->tile_width * cube_info->tile_height + 4 * cube_info->tile_width * pixel_row + 4 * pixel_column, SEEK_SET);
			reader[0] = fgetc(cube_info->file);
			reader[1] = fgetc(cube_info->file);
			reader[2] = fgetc(cube_info->file);
			reader[3] = fgetc(cube_info->file);
			if (!cube_info->byte_order_matches) {
				// Endianness doesn't match; swap the bytes
				reader[0] ^= reader[3];
				reader[3] ^= reader[0];
				reader[0] ^= reader[3];
				reader[1] ^= reader[2];
				reader[2] ^= reader[1];
				reader[1] ^= reader[2];
			}
			pixel_value = stored_value;
		}
	}
	// Apply base and multiplier, if applicable
	pixel_value *= cube_info->multiplier;
	pixel_value += cube_info->base;
	return pixel_value;
}

// Get value given geographic coordinates
float get_value_at_geographic_point(cube_properties * cube_info, double longitude, double latitude) {
	validate_longitude(&longitude, cube_info);
	double row_fraction = 1 - ((latitude - cube_info->latitude_minimum) / (cube_info->latitude_maximum - cube_info->latitude_minimum));
	if (row_fraction < 0) row_fraction = 0;
	if (row_fraction > 1) row_fraction = 1;
	uint64_t row = (uint64_t)(row_fraction * (cube_info->height - 1));
	double column_fraction = (longitude - cube_info->longitude_minimum) / (cube_info->longitude_maximum - cube_info->longitude_minimum);
	if (column_fraction < 0) column_fraction = 0;
	if (column_fraction > 1) column_fraction = 1;
	uint64_t column = (uint64_t)(column_fraction * (cube_info->width - 1));
	return get_value_at_image_point(cube_info, row, column);
}

/* Initialize an ISIS cube interfacing structure.
 *
 * cube_info:   Pointer to the structure in which to store information
 * cube_file:   Path to the ISIS cube file
 * radius:      Overridden radius of the planetary body in meters.  This may be
 *              non-positive, in which case the radius must be supplied by the
 *              ISIS cube header.
 * cache_MB:    An allowable size in megabytes for the tile cache.  This may be
 *              0, in which case no caching will be performed.
 * preload:     If true, the entire content of the cube will be sequentially
 *              loaded into memory immediately, and kept there.  This is a lot
 *              of time and work at initialization time, but no disk I/O is
 *              required afterward to access the cube content, and in some
 *              cases, the tradeoff is well worth it.
 */
void init_cube(cube_properties * cube_info, char * cube_file, double radius, unsigned int cache_MB, bool preload, bool verbose) {
	if (radius > 0) {
		// Radius was supplied; use it and ignore any radius in the header
		cube_info->radius = radius;
	}
	// Open the file
	cube_info->file_path = strdup(cube_file);
	cube_info->file = fopen(cube_file, "r");
	if (cube_info->file == NULL) {
		fprintf(stderr, "Error: could not open cube file %s!  Aborting...\n", cube_file);
		exit(1);
	}
	// Things we require from the header
	bool found_data_start_byte = false;
	bool found_width = false, found_height = false;
	bool found_longitude_domain = false;
	bool found_longitude_center = false, found_latitude_center = false;
	bool found_upper_left_x = false, found_upper_left_y = false;
	bool found_radius = false;
	bool found_meters_per_pixel = false;
	bool found_pixels_per_degree = false;
	bool found_bytes_per_pixel = false, found_byte_order = false;
	// Defaults for things we don't require
	cube_info->tile_width = 1;
	cube_info->tile_height = 1;
	cube_info->base = 0;
	cube_info->multiplier = 1;
	// Parse the header
	char line[2048];
	bool finished_reading_header = false;
	bool looking_at_core_object = false;
	bool looking_at_pixel_group = false;
	while (!finished_reading_header) {
		// Get a line
		char * reader = line;
		int byte;
		while ((byte = fgetc(cube_info->file)) != '\n') {
			if (byte == EOF) {
				fprintf(stderr, "Error: premature EOF mid-header!\n");
				exit(1);
			}
			*reader = byte;
			reader++;
		}
		*reader = 0;
		// Lowercase the line so we don't have to worry about any ambiguity in capitalization
		for (int i = 0; i < strlen(line); i++) line[i] = tolower(line[i]);
		// Interpret the line
		if (strstr(line, "end") != NULL && strstr(line, "group") == NULL && strstr(line, "object") == NULL) finished_reading_header = true;
		if (strstr(line, "object") != NULL && strstr(line, "core") != NULL) looking_at_core_object = true;
		if (strstr(line, "end") != NULL && strstr(line, "object") != NULL) looking_at_core_object = false;
		if (strstr(line, "group") != NULL && strstr(line, "pixels") != NULL) looking_at_pixel_group = true;
		if (strstr(line, "end") != NULL && strstr(line, "group") != NULL) looking_at_pixel_group = false;
		if (strstr(line, "projectionname") != NULL) {
			if (strstr(line, "simplecylindrical") != NULL) {
				// The "Simple cylindrical" projection implies center latitude of 0
				cube_info->latitude_center = 0;
				found_latitude_center = true;
			}
			else if (strstr(line, "equirectangular") == NULL) {
				fprintf(stderr, "Error: Cube %s's projection does not appear to be SimpleCylindrical / Equirectangular --  other projections are not supported.\n", cube_file);
				exit(1);
			}
		}
		if (looking_at_core_object) {
			if (sscanf(line, "%*[ =\t]startbyte%*[ =\t]%"PRIu64"", &cube_info->data_start_byte) == 1) found_data_start_byte = true;
		}
		if (sscanf(line, "%*[ =\t]samples%*[ =\t]%"PRIu64"", &cube_info->width) == 1) found_width = true;
		if (sscanf(line, "%*[ =\t]lines%*[ =\t]%"PRIu64"", &cube_info->height) == 1) found_height = true;
		if (sscanf(line, "%*[ =\t]longitudedomain%*[ =\t]%"PRId32"", &cube_info->longitude_domain) == 1) found_longitude_domain = true;
		if (sscanf(line, "%*[ =\t]centerlongitude%*[ =\t]%lf", &cube_info->longitude_center) == 1) found_longitude_center = true;
		if (sscanf(line, "%*[ =\t]centerlatitude%*[ =\t]%lf", &cube_info->latitude_center) == 1) found_latitude_center = true;
		if (sscanf(line, "%*[ =\t]upperleftcornerx%*[ =\t]%lf", &cube_info->upper_left_x) == 1) found_upper_left_x = true;
		if (sscanf(line, "%*[ =\t]upperleftcornery%*[ =\t]%lf", &cube_info->upper_left_y) == 1) found_upper_left_y = true;
		// Only check the header for a radius if the caller didn't supply one manually
		if (radius <= 0) {
			if (sscanf(line, "%*[ =\t]polarradius%*[ =\t]%lf", &cube_info->radius) == 1) found_radius = true;
			if (sscanf(line, "%*[ =\t]equatorialradius%*[ =\t]%lf", &cube_info->radius) == 1) found_radius = true;
		}
		if (sscanf(line, "%*[ =\t]pixelresolution%*[ =\t]%lf", &cube_info->meters_per_pixel) == 1) found_meters_per_pixel = true;
		if (sscanf(line, "%*[ =\t]scale%*[ =\t]%lf", &cube_info->pixels_per_degree) == 1) found_pixels_per_degree = true;
		char data_type[64];
		if (looking_at_pixel_group && sscanf(line, "%*[ =\t]type%*[ =\t]%s", data_type) == 1) {
			if (strcmp(data_type, "unsignedbyte") == 0) {
				cube_info->bytes_per_pixel = 1;
			}
			else if (strcmp(data_type, "signedword") == 0) {
				cube_info->bytes_per_pixel = 2;
			}
			else if (strcmp(data_type, "real") == 0) {
				cube_info->bytes_per_pixel = 4;
			}
			else {
				fprintf(stderr, "Unrecognized data type for %s.\nHeader content: \"%s\".\nAborting.\n", cube_file, line);
				exit(1);
			}
			found_bytes_per_pixel = true;
		}
		char byte_order[64];
		if (looking_at_pixel_group && sscanf(line, "%*[ =\t]byteorder%*[ =\t]%s", byte_order) == 1) {
			// Determine the byte ordering of this machine
			bool this_machine_MSB;
			uint16_t test = 1;
			unsigned char * peek = (unsigned char *)(&test);
			this_machine_MSB = (*peek == 0);
			if (strcmp(byte_order, "lsb") == 0) {
				cube_info->byte_order_matches = !this_machine_MSB;
			}
			else if (strcmp(byte_order, "msb") == 0) {
				cube_info->byte_order_matches = this_machine_MSB;
			}
			else {
				fprintf(stderr, "Unrecognized byte ordering for %s.\nHeader content: \"%s\".\nAborting.\n", cube_file, line);
				exit(1);
			}
			found_byte_order = true;
		}
		sscanf(line, "%*[ =\t]tilesamples%*[ =\t]%"PRIu64"", &cube_info->tile_width);
		sscanf(line, "%*[ =\t]tilelines%*[ =\t]%"PRIu64"", &cube_info->tile_height);
		sscanf(line, "%*[ =\t]base%*[ =\t]%lf", &cube_info->base);
		sscanf(line, "%*[ =\t]multiplier%*[ =\t]%lf", &cube_info->multiplier);
	}
	// Did we get the mandatory things?
	if (!found_data_start_byte) {
		fprintf(stderr, "Error: Data start byte not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_width) {
		fprintf(stderr, "Error: Image width not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_height) {
		fprintf(stderr, "Error: Image height not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_longitude_domain) {
		fprintf(stderr, "Error: LongitudeDomain not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_longitude_center || !found_latitude_center) {
		fprintf(stderr, "Error: CenterLongitude and/or CenterLatitude not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_upper_left_x || !found_upper_left_y) {
		fprintf(stderr, "Error: UpperLeftCornerX and/or UpperLeftCornerY not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_radius && radius <= 0) {
		fprintf(stderr, "Error: Radius (either polar or equatorial) not found in header for %s, and was not supplied by caller.\n", cube_file);
		exit(1);
	}
	if (!found_meters_per_pixel) {
		fprintf(stderr, "Error: PixelResolution not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_pixels_per_degree) {
		fprintf(stderr, "Error: Scale not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_bytes_per_pixel) {
		fprintf(stderr, "Error: Type (image pixel value format) not found in header for %s.\n", cube_file);
		exit(1);
	}
	if (!found_byte_order) {
		fprintf(stderr, "Error: ByteOrder not found in header for %s.\n", cube_file);
		exit(1);
	}
	// Compute tile count
	cube_info->tiles_wide = cube_info->width  / cube_info->tile_width  + (cube_info->width  % cube_info->tile_width  == 0 ? 0 : 1);
	cube_info->tiles_high = cube_info->height / cube_info->tile_height + (cube_info->height % cube_info->tile_height == 0 ? 0 : 1);
	cube_info->tile_count = cube_info->tiles_wide * cube_info->tiles_high;
	// Compute input longitude and latitude boundaries
	double min_x = (1 - 0.5) * cube_info->meters_per_pixel + cube_info->upper_left_x;
	double max_x = ((double)cube_info->width - 0.5) * cube_info->meters_per_pixel + cube_info->upper_left_x;
	double min_y = cube_info->upper_left_y - ((double)cube_info->height - 0.5) * cube_info->meters_per_pixel;
	double max_y = cube_info->upper_left_y - (1 - 0.5) * cube_info->meters_per_pixel;
	cube_info->longitude_minimum = cube_info->longitude_center + min_x * (360.0 / (cube_info->radius * cos(cube_info->latitude_center * DEG_TO_RAD) * 2 * PI));
	cube_info->longitude_maximum = cube_info->longitude_center + max_x * (360.0 / (cube_info->radius * cos(cube_info->latitude_center * DEG_TO_RAD) * 2 * PI));
	cube_info->latitude_minimum  = min_y * (360.0 / (cube_info->radius * 2 * PI));
	cube_info->latitude_maximum  = max_y * (360.0 / (cube_info->radius * 2 * PI));
	// Initialize the cache if the caller wants one *and* the cube is actually
	// tiled; otherwise remember that we're not using one!
	if (cache_MB > 0 && (cube_info->tile_width > 1 || cube_info->tile_height > 1)) {
		// Decide how many tiles we can allow
		uint32_t cache_tile_count = cache_MB * 1024 * 1024 / cube_info->tile_width / cube_info->tile_height / cube_info->bytes_per_pixel;
		cube_info->tile_cache = malloc(cache_tile_count * sizeof(struct cached_cube_tile));
		if (!cube_info->tile_cache) {
			fprintf(stderr, "Error: could not allocate memory for cube tile cache!\n");
			exit(1);
		}
		cube_info->tile_cache_capacity = cache_tile_count;
		cube_info->tile_cache_current_size = 0;
	}
	else {
		cube_info->tile_cache = NULL;
	}
	// Load the entire cube into memory if desired, otherwise ensure the pointer is null
	if (preload) {
		SAY("Reading entire cube %s into memory...\n", cube_info->file_path);
		cube_info->complete_data = malloc(sizeof(unsigned char *) * cube_info->tile_count);
		if (cube_info->complete_data == NULL) {
			fprintf(stderr, "Error: Could not allocate enough tile pointers to load cube %s.\n", cube_info->file_path);
			exit(1);
		}
		for (uint64_t tile = 0; tile < cube_info->tile_count; tile++) {
			cube_info->complete_data[tile] = malloc(cube_info->tile_width * cube_info->tile_height * cube_info->bytes_per_pixel);
			if (cube_info->complete_data[tile] == NULL) {
				fprintf(stderr, "Error: Could not allocate enough memory for cube %s.\n", cube_info->file_path);
				exit(1);
			}
		}
		// Seek to cube_info->complete_data beginning
		fseek(cube_info->file, cube_info->data_start_byte - 1, SEEK_SET);
		// Read in the cube cube_info->complete_data, with progress report if we are being verbose
		uint64_t total_bytes = cube_info->tile_width * cube_info->tile_height * cube_info->tile_count * cube_info->bytes_per_pixel;
		uint64_t bytes_read = 0;
		int previous_percent = -1;
		for (uint64_t tile = 0; tile < cube_info->tile_count; tile++) {
			unsigned char * reader = (unsigned char *)(cube_info->complete_data[tile]);
			for (uint64_t byte = 0; byte < cube_info->tile_width * cube_info->tile_height * cube_info->bytes_per_pixel; byte++) {
				*reader = fgetc(cube_info->file);
				bytes_read++;
				int percent_finished = (int)(bytes_read * 100 / total_bytes);
				if (percent_finished > previous_percent) {
					SAY("\r%d%% complete...", percent_finished);
					fflush(stdout);
					previous_percent = percent_finished;
				}
				reader++;
			}
		}
		SAY("\n");
	}
	else {
		cube_info->complete_data = NULL;
	}
}

// Cleanly destroy an ISIS cube interface structure
void free_cube(cube_properties * cube_info) {
	// Close the file
	fclose(cube_info->file);
	// Free up the pathname string
	free(cube_info->file_path);
	if (cube_info->tile_cache) {
		// We used a tile cache; free it up
		for (int i = 0; i < cube_info->tile_cache_current_size; i++) {
			free(cube_info->tile_cache[i].data);
		}
		free(cube_info->tile_cache);
	}
	if (cube_info->complete_data) {
		// We preloaded the entire cube; clean up
		for (uint64_t tile = 0; tile < cube_info->tile_count; tile++) {
			free(cube_info->complete_data[tile]);
		}
		free(cube_info->complete_data);
	}
}
